/*
 *  Copyright Quantum Software Systems Ltd. 1984 UNPUBLISHED
 *
 */
#ifndef	_diskio_included
#define	_diskio_included

/*

	usual include files for disk drivers.

*/
#include	<conio.h>
#include	<i86.h>
#include	<sys/types.h>
#include	<sys/fsys.h>
#include	<sys/_dskdrvr.h>
#include	<sys/disk.h>
#include	<sys/sched.h>
#include	<sys/sys_msg.h>
#include	<sys/qioctl.h>
#include	"pragma.h"


#define	FIXED_DISK_PREFIX		"hd"
#define	FLOPPY_DISK_PREFIX		"fd"


typedef struct  _disk  DISK;
typedef struct  _block   BLOCK;

/*	physical address */
typedef long paddr_t;	

/*
 *	Structure Definitions
 *
 * A block relative to a QNX partition is decomposed into the
 * following structure.
 */

typedef struct _bdata_entry {
	long block;				/* Relative to start of disk 1...n		*/
	unsigned pdrive;		/* Physical drive 0 or 1				*/
	unsigned track;			
	unsigned head;
	unsigned sector;
	} BDATA;

int	fsys_mount_driver(pid_t fsys,short ndevs,struct _disk *devtab,
						int (**jmptab)(), short maxio, short maxpar,
						void *sptr, struct _sysmsg_version_reply *vers);

/* convenient type for mapping logical to physical addresses */
typedef	struct	dma_buffer {
	void	__far *bufp;	/*	logical address 	*/
	paddr_t phys_addr;		/*	physical address 	*/
	} dma_t;

/*	for timeouts -- prevents compiler from eating loops! */
typedef volatile int timeout_t;


/*	iorequest types */
#define	_DNAME_PREFIX	8
#define	_OPEN_SYNC		001
enum {
	READ_BLK = _READ_BLK,		
	WRITE_BLK = _WRITE_BLK,
	GET_STATUS,
	FMT_TRACK
	};




#define	HDPT_1	0x41 		/*	Hard Disk Parameter Table for physical drive 1 */
#define	HDPT_2	0x46 		/*	Hard Disk Parameter Table for physical drive 2 */


void map_thn(BDATA *,DISK *);
long unmap_thn(BDATA *, DISK *);

int get_cmos(int cmos_addr);	/*	read a byte from setup memory */

int get_disk_parms(struct _disk *dp,int disknum,char ps2);


/*	some changes are required here:
	change the output_flag to a set of flags.
	combine the setup_dma/floppy_setup_dma into 1
*/
		
#define	DMA_OUTPUT	001
/*	ps2 / at	*/
#define	DMA_PS2		002
/*	word / byte	*/
#define	DMA_WORD	004
/* don't arm dma controller on setup */
#define	DMA_WAIT	010

int	alloc_dma_buf(struct dma_buffer *bufp, pid_t process, long nbytes);
int	setup_dma(unsigned dma_flags, dma_t *dbuf, unsigned iosize, int channel);
int	kick_dma(unsigned dma_flags, int channel);
int	clear_dma(unsigned dma_flags, int channel);


paddr_t phys_addr(void __far *ptr);

void cache_to_buf(char __far *bufp, BLOCK  __far *cachep, unsigned nblocks);

void buf_to_cache(BLOCK __far *cachep, char __far *bufp, unsigned nblocks);

/*	managing trace buffers */
void		init_trace(void __far *bufp, int bufsiz);
int			trace_enable(void);
int			trace_disable(void);
void __far *trace_head(void);
void __far *alloc_trace(int nbytes);


int	get_cpu_speed(void);
void usec_delay(int usecs);


/* qnx2-style io functions. */
unsigned io_in(unsigned ioaddr);
unsigned io_win(unsigned ioaddr);
void io_out(unsigned ioaddr, unsigned val);
void io_wout(unsigned ioaddr, unsigned val);



/*	paranioa disk verification: */

int		init_watch(unsigned long nblocks);
int		check_block(unsigned long blockno, const unsigned char far *bufp, 
				int write_flag);



/*	for debugging: */
#define	COLOUR_DISP	0x30
#define	MONO_DISP	0x28

#ifdef DEBUGGING

int	set_display(void __far *scr_buf, int nlines, int ncols);
int	set_attr(int attr);
int	disp_char(int c);
int disp_str(char *s);
int disp_hex(int value);
int	disp_lhex(long value);

#else

#define	set_display(x,y,z)
#define	set_attr(x)
#define	disp_char(x)
#define	disp_str(x)
#define	disp_hex(x)
#define	disp_lhex(x)

#endif




/*	enforce no stack checking */
#pragma off (check_stack);

#endif


