
#include <diskio.h>
#include <i86.h>


/*	8237 mode bits */

#define	DMA_M_OUT		0x08
#define	DMA_M_IN		0x04

#define	DMA_M_AUTO_INIT	0x10

#define	DMA_M_AUTO_DECR	0x20
#define	DMA_M_AUTO_INCR	0x00

#define	DMA_M_DEMAND	0x00
#define	DMA_M_SINGLE	0x40
#define	DMA_M_BLOCK		0x80
#define	DMA_M_CASCADE	0xc0

/*	PS/2 dma function registers */
#define	PS2_EFUNC_REG	0x18
#define	PS2_EFUNC_EXE	0x1a


struct	regs_8237 {
		int		maddr, tcount, pagereg;
		int		stat,  mask, mode;
		int		dma_clear, dma_reset, clear_mask, wmask;
	};

/*	at dma registers	*/
static	const	struct	regs_8237	regtable[] = {

		{	0x00, 0x01, 0x87, 0x08, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f },
		{	0x02, 0x03, 0x83, 0x08, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f },
		{	0x04, 0x05, 0x81, 0x08, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f },
		{	0x06, 0x07, 0x82, 0x08, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f },

		{	0xc0, 0xc1, 0x8f, 0xd0, 0xd4, 0xd6, 0xd8, 0xda, 0xdc, 0xde },
		{	0xc4, 0xc5, 0x8b, 0xd0, 0xd4, 0xd6, 0xd8, 0xda, 0xdc, 0xde },
		{	0xc8, 0xc9, 0x89, 0xd0, 0xd4, 0xd6, 0xd8, 0xda, 0xdc, 0xde },
		{	0xcc, 0xcd, 0x8a, 0xd0, 0xd4, 0xd6, 0xd8, 0xda, 0xdc, 0xde }
	};
                                                   

setup_dma(unsigned dflags, dma_t *iobufp, unsigned iosize, int channel)
{
int	psw;
struct	regs_8237	*dp = regtable + channel;
union {
	paddr_t	addr;
	unsigned char bytes[3];
} physaddr;


	if (channel > 7)
		return -1;		/*	only 0-7 are supported */

	physaddr.addr = iobufp->phys_addr;

	psw = disable();

	if (dflags & DMA_PS2) {
/* Set channel mask bit		*/
		io_out(PS2_EFUNC_REG,  0x90 | channel);			
		io_out(PS2_EFUNC_REG,  0x20 | channel);			
/* Setup to write mem addr	*/
		io_out(PS2_EFUNC_EXE, physaddr.bytes[0]);
		io_out(PS2_EFUNC_EXE, physaddr.bytes[1]);
		io_out(PS2_EFUNC_EXE, physaddr.bytes[2]);
/* Setup to write count		*/
		io_out(PS2_EFUNC_REG,  0x40 | channel);
		if ((dflags & DMA_WORD) == 0) {
			iosize /= sizeof(int)/sizeof(char);
		}
		--iosize;
		io_out(PS2_EFUNC_EXE, iosize);
		io_out(PS2_EFUNC_EXE, (iosize >> 8) & 0xff);
/* Setup to write ext mode	*/
		io_out(PS2_EFUNC_REG,  0x70 | channel);
		io_out(PS2_EFUNC_EXE, (dflags & DMA_OUTPUT) ? 0x44 : 0x4c);
/* Clear channel mask bit	*/
		io_out(PS2_EFUNC_REG,  0xa0 | channel);
    } else {

		if (dflags & DMA_WORD)
			iosize *= sizeof(int)/sizeof(char);
		iosize--;
		io_out(dp->dma_clear,0);			/*	clear first/last FF */
		io_out(dp->mode, 	  (channel & 0x03) | ((dflags & DMA_OUTPUT) ? 
							  (DMA_M_OUT | DMA_M_SINGLE) 
							: (DMA_M_IN  | DMA_M_SINGLE)));
		io_out(dp->maddr,   physaddr.bytes[0]);
		io_out(dp->maddr,   physaddr.bytes[1]);
		io_out(dp->pagereg, physaddr.bytes[2]);
		io_out(dp->tcount,  iosize);
		io_out(dp->tcount,  (iosize >> 8) & 0xff);
		io_out(dp->mask,    channel);
    }
	restore(psw);
	return 0;
}
kick_dma(unsigned dflags, int channel)
{
	if (dflags & DMA_PS2) {
		io_out(PS2_EFUNC_REG,  0xa0 | channel);
	} else {
		io_out((regtable+channel)->mask,channel);
	}

	return 0;
}
clear_dma(unsigned dflags, int channel)
{
	if (dflags & DMA_PS2) {
		io_out(PS2_EFUNC_REG, 0x090 | channel);
	} else {
		io_out(regtable[channel].dma_clear, 0);
	}
	return 0;
}


