
#include	<stdio.h>
#include	<string.h>
#include	<errno.h>
#include	"diskio.h"
#include	<i86.h>
#include	<sys/seginfo.h>

struct disk_param_table {
	unsigned		disk_param_num_tracks;	
	unsigned char	disk_param_num_heads;
	unsigned		not_used1;
	unsigned		disk_param_precomp;
	unsigned char	not_used2;
	unsigned char	disk_param_control;
	unsigned char	not_used3[3];
	unsigned		disk_param_landing_zone;
	unsigned char	disk_param_num_sectors;
	unsigned char	not_used4;
	};

#define	HDPT_1	0x41 	/*	Hard Disk Parameter Table for physical drive 1 */
#define	HDPT_2	0x46 	/*	Hard Disk Parameter Table for physical drive 2 */

/*
 * The structure of the fixed disk parameter table in the AT BIOS
 */

/*
 *	Get the disk parameters from the BIOS
 */
int
get_disk_parms( struct _disk *dp, int disknum, char ps2) 
{
struct disk_param_table __far *tp;
long			offset;
unsigned __far *far_off;
unsigned		baseseg, seg, off;
unsigned char	cmos;

	/*
	 * Read Drive type from CMOS RAM
	 *    | c c c c | d d d d |
	 */
	if (ps2) {
		if (get_cmos(0x11 + (disknum != 0)) == 0)
			return -1;
	} else {
		cmos = get_cmos(0x12);

		if (((disknum ? cmos : cmos >> 4) & 0x0f) == 0)
			return( -1 );			/* No drive set */
		}

	if ((baseseg = qnx_segment_overlay(0L, 1024L)) == -1)
		return( -1 );

	/*	Get offset of the hard disk parameter table in the BIOS */
	far_off = (unsigned __far *) 
			MK_FP(baseseg,(disknum ? HDPT_2:HDPT_1)*sizeof(void __far *));
	
	seg = *(far_off + 1);
	off = *far_off;
	qnx_segment_free( baseseg );
	
	offset = (long) seg << 4;
	if ((seg = qnx_segment_overlay((long)seg << 4, 48L+off)) == -1) {
		return -1;
	}
	/*
	 * Extract the hard disk parameter info from the table entry pointed
	 * to by the interrupt vector
	 */
	tp = (struct disk_param_table __far *) MK_FP(seg, off);

	dp->d_cyls		= tp->disk_param_num_tracks;
	dp->d_heads		= tp->disk_param_num_heads;
	dp->d_sctr_trk	= tp->disk_param_num_sectors;
	dp->d_precomp	= tp->disk_param_precomp;
	qnx_segment_free( seg );

	return( 0 );
}
