
/*
 *	Header File for the Adaptec AHA-5240A/1542A SCSI Host Adapter driver
 *
 */


#define ON			1
#define OFF			0
#define TRUE		1
#define FALSE		0





/*
 *	CDB Flags
 */
#define CF_LINK		0x01
#define CF_FLAG		0x02
#define CF_VENDOR0	0x40
#define CF_VENDOR1	0x80

#define CF_FUA		0x08
#define CF_DPO		0x10

/*
 *	Port Address defines
 */
#define CONTROL		0		/*	Control port (write) */
#define STATUS		0		/*	Status port (read) */
#define COMMAND		1		/*	Command port (write) */
#define DATA		1		/*	Data I/O port (read/write) */
#define INT_FLAGS	2		/*	Interrupt Flags port (read only) */	

/*
 *  AHA-154xA Control Register defines
 *  (write base port + 0)
 */
#define HRST	0x80				/*	Hard reset (like powerup) */
#define SRST	0x40				/*	Soft reset */
#define IRST	0x20				/*	Interrupt reset */
#define SCRST	0x10				/*	SCSI bus reset */

/*
 *	AHA-154xA Status Register defines
 *  (read base port + 0) 
 */
#define STST	0x80				/*	Self test in progress */
#define DIAGF	0x40				/*	Internal diagnostic failure */
#define INIT	0x20				/*	Mailbox init not done */
#define IDLE	0x10				/*	Ready for next command */
#define CDF		0x08				/*	Command/data out port full */
#define DF		0x04				/*	Data in port full */
#define INVDCMD	0x01				/*	Illegal command */
#define RESET	0x00				/*	Reset in progress */

/*
 *	AHA-154xA Interrupt Flags Register defines
 *	(read base port + 2)
 */
#define AINT	0x80				/*	Any interrupt */
#define SCRD	0x08				/*	SCSI Reset Detected */
#define HACC	0x04				/*	Host Adaptor Command Complete */
#define MBOA	0x02				/*	MBO Available */
#define MBIF	0x01				/*	MBI Full */

/*
 *	Return codes from aha_cmd()
 */
#define NOT_IDLE		0xef
#define INVD_OPCODE		0xfe

/*
 *	Adaptec AHA-154xA Host Adapter command codes
 */
#define AHA_NOP			0x00		/*	Null Operation */
#define AHA_MBINIT		0x01		/*	Mailbox init */
#define AHA_SCSI_CMD	0x02		/*	Start SCSI command */
#define AHA_BIOS_CMD	0x03		/*	Start BIOS command */
#define AHA_INQUIRE		0x04 		/*	Adapter inquiry */
#define AHA_MBO_INT		0x05		/*	Set mailbox available interrupt */
#define AHA_SEL_TIME	0x06		/*	Set SCSI selection timeout */
#define AHA_BUSON_TIME	0x07		/*	Set bus on time */
#define AHA_BUSOFF_TIME	0x08		/*	Set bus off time */
#define AHA_XFER_SPEED	0x09		/*	Set transfer speed */
#define AHA_DEVICES		0x0A		/*	Return installed devices */
#define AHA_CONFIG		0x0B		/*	Return config data */
#define AHA_TARGET_MODE	0x0C		/*	Enable target mode */
#define AHA_SETUP		0x0D		/*	Return setup data */
#define AHA_WRITE_C2BUF	0x1A		/*	Write adapter channel 2 buffer */
#define AHA_READ_C2BUF	0x1B		/*	Read adapter channel 2 buffer */
#define AHA_WRITE_FIFO	0x1C		/*	Write adapter FIFO buffer */
#define AHA_READ_FIFO	0x1D		/*	Read adapter FIFO buffer */
#define AHA_ECHO_CMD	0x1F		/*	ECHO command data */

#define MBOX_MAX	1

/*
 *	MBO Command codes
 */
#define MBO_FREE		0x00		/*	MBO is free */
#define MBO_START		0x01		/*	Start SCSI command */
#define MBO_ABORT		0x02		/*	Abort SCSI command */

/*
 *	MBI Status codes
 */
#define MBI_FREE		0x00		/*	MBI is free */
#define MBI_COMPLETE	0x01		/*	CCB completed without error */
#define MBI_ABORT		0x02		/*	CCB aborted by host */
#define MBI_ABORTLOST	0x03		/*	Aborted CCB not found */
#define MBI_ERROR		0x04		/*	CCB completed with error */
#define MBI_TARGET		0x10		/*	Target cmd, no CCB available */

/*
 *	CCB Command Codes
 */
#define CCB_INITR		0x00		/*	Initiator command */
#define CCB_TARGET		0x01		/*	Target mode command */
#define CCB_INITR_SG	0x02		/*	Initiator command w. scatter/gather */
#define CCB_BDR			0x81		/*	Bus Device Reset */

/*
 *	CCB mask definitions
 */
#define CCB_ID			0xe0		/*	Target/Initiator ID */
#define CCB_OUT			0x10		/*	Outbound data transfer */
#define CCB_IN			0x08		/*	Inbound data transfer */
#define CCB_LUN			0x07		/*	Logical Unit Number */


/*
 *	Definition of a mailbox entry
 */
#if 0
typedef struct _mbox {
	uchar		mb_cmd;
	scsi_adr	mb_ccb;
} mbox;


/*
 *	Definition of driver's Target Descriptor Block
 */
typedef struct _tdb {
	unsigned	td_id;			/*	Target ID */
	unsigned	td_luns;		/*	LUNs */
	unsigned	td_flags;		/*	Flags */
	uchar		td_version;		/*	SCSI Version */
	uchar		td_devtype;		/*	Device Type */
	uchar		td_fblk_dblk;	/*	Number of fsys blocks per device block */
	uchar		td_shift;		/*	Shift factor for blocks > 512 uchars */
	uchar		td_mask;		/*	Mask used for blocks > 512 uchars */
	uchar		td_spare;
	long		td_numblocks;	/*	Number of blocks */
	unsigned	td_blksize;		/*	Block size (in uchars) */
	unsigned	td_blks_cyl;	/*	Blocks per cylinder */
	unsigned	td_err_sense;	/*	Error Sense Key */
	unsigned	td_err_asc;		/*	Additional Sense Code and Qualifier (hi/lo) */
	long		td_err_blk;		/*	Block number of error (if applicable) */
	unsigned	td_spare2[2];
} tdb;


/*
 *	Definition of a AHA-154xA SCSI Command Control Block
 *
 *	N.B.
 *		- SCSI addresses are three uchar big-endian (High-Low) addresses
 *		- The SCSI CDB has a length of ccb_cmdlen
 *		- The Request Sense Data uchars follow the SCSI CCB and
 *		  the length is given in ccb_rsal_das
 *		- The CDB has been given 12 uchars which is the maximum length
 *		  of a CDB defined in the ANSI Standard X3.131 SCSI-2 Rev10.
 *		  However, the CDB length can vary from 6 uchars to more than 
 *		  12 uchars for a vendor specific CDB.  Likewise the Request
 *		  Sense Data can be up to 256 uchars although X3.131 only defines
 *		  the first 18.  Therefore, with a Group 5 (12 uchar) CDB a maximum
 *		  of 34 uchars of Sense Data can be retrieved without enlarging the
 *		  the overflow field of this structure from 16 uchars
 *
 *		- The current size of this CCB definition is 64 uchars
 *
 */
typedef struct _scsi_ccb {
	uchar	ccb_op;				/*	Operation code */
	uchar	ccb_adr_dc;			/*	Address and Direction Control */
	uchar	ccb_cmdlen;			/*	SCSI Command Length */
	uchar	ccb_rsal_das;		/*	Request Sense Allocation Length or
									Disable Auto Sense */
	scsi_adr	ccb_datalen;	/*	Data Length */
	scsi_adr	ccb_dataptr;	/*	Data Pointer */
	scsi_adr	ccb_linkptr;	/*	Link Pointer */
	uchar	ccb_cli;			/*	Command Linking Identifier */
	uchar	ccb_hastat;			/*	Host Adapter Status */
	uchar	ccb_tarstat;		/*	Target device Status */
	uchar	ccb_reserved1;
	uchar	ccb_reserved2;
	uchar	ccb_cdb[12];		/*	SCSI Command Descriptor Block */
	uchar	ccb_sense[18];		/*	Request Sense Data */
	uchar	ccb_overflow[16];	/*	Request Sense Overflow */
} ccb;

#endif

struct	aha_regs {
	unsigned	control, status;
	unsigned	command, data;
	unsigned	int_status;
	};




#define	MB_STAT		0
#define	MB_CCBPTR	1
#define	MB_SIZE		4


#define	CCB_OPCODE	0
#define	CCB_ADDR	1
#define	CCB_CMDLEN	2
#define	CCB_SENSE	3
#define	CCB_DATALEN	4
#define	CCB_DATAPTR	7
#define	CCB_LINKPTR	10
#define	CCB_LINKID	13
#define	CCB_HASTAT	14
#define	CCB_TARSTAT	15
#define	CCB_SCSICMD	18
#define	CCB_SCSICMD_MAX	12
#define	CCB_REQSENSE(x) (CCB_SCSICMD+x)
#define	CCB_REQSENSE_MAX	34
#define	CCB_MAXBUF	(CCB_SCSICMD + CCB_SCSICMD_MAX + CCB_REQSENSE_MAX)

struct	scsi_device {
/*	init time only */
		int		is_valid;
		int		target;
		int		lun;
		dma_t	iobuf;
		pid_t	proxyid;
		pid_t (*event_hdlr)(struct scsi_device *dev,int event);
		uchar	*scsicmd;
		long	aha_phys;		/*	address of aha ccb */
		int		mbox;
		int		retry_count;
		
/*	updated by io request */
		int		nblocks;
		int		function;
		long	blockno;
		BLOCK	__far *cache_list;

/*	updated at interrupt time, initialed at init time */
		int		state;
		DISK	*reqptr;
/*	updated by aha */
		uchar	aha_out[CCB_MAXBUF];
	};

typedef struct scsi_device	SCSI;


enum event_type {
	IORequest = 0,
	Interrupt
	};

