#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<string.h>
#include	<sys/kernel.h>
#include	<sys/irqinfo.h>

#include	<diskio.h>

#include	"endian.h"
#include	"genscsi.h"
#include	"aha1540.h"
#include	"ahaextrn.h"

#define	LOCAL	


enum driver_state {
	ST_IDLE = 0,
	ST_WAIT_INQ,
	ST_WAIT_START,
	ST_WAIT_CAPAC,
	ST_WAIT_IO,
	};

enum {
	START_IO,                                       
	ISSUE_TEST,
	ISSUE_CAP,
	GOT_PARMS,
	IO_DONE,
	TX_ERROR
	};

LOCAL const int	action[][2] = {
						/*	IORequest */	/*	Interrupt	*/
/*	ST_IDLE */		{   	START_IO,			TX_ERROR	},
/*	ST_WAIT_INQ*/	{   	TX_ERROR,			ISSUE_TEST	},
/*	ST_WAIT_START */{   	TX_ERROR,			ISSUE_CAP	},
/*	ST_WAIT_CAP*/	{   	TX_ERROR,			GOT_PARMS	},
/*	ST_WAIT_IO */	{   	TX_ERROR,			IO_DONE		}
			};
			
pid_t	fsys_event(SCSI *dp, int input);
LOCAL	pid_t	fsys_proxy;


int __far drvinit(pid_t driver_proxy, short tabsize, unsigned __far *cachetab)
{
	if (tabsize || cachetab)
		;
	fsys_proxy = driver_proxy;

	if (adaptec_init() == -1) {
disp_str("adaptec_init: failed\n");
		return -1;
	}
	return 0;
}

int __far drvopen(int disknum, int flags, int rsvd1, int rsvd2)
{
SCSI	*p;
	if (rsvd1 || rsvd2)
		;
		
	if (disknum >= MAX_NDRIVES) {
		return -1;
	}
disp_str("open");
disp_hex(disknum);
	p = devtab + disknum;

	if (devmap[disknum] == 0) {
		disktab[disknum].d_retval = -1;
		return -1;
	}
	if (alloc_dma_buf(&p->iobuf,FsysPid,MaxIO*_BLOCK_SIZE) == -1) {
disp_str("dma failed");
		return -1;
	}
disp_str("iobuf addr =");
disp_lhex(p->iobuf.phys_addr);
	strcpy(disktab[disknum].d_prefix,driver_name);
	p->target       = disknum;
	p->lun          = 0;
	p->proxyid      = fsys_proxy;
	p->event_hdlr   = fsys_event;
	p->scsicmd      = p->aha_out + CCB_SCSICMD;
	p->reqptr       = disktab+disknum;
	p->nblocks      = 0;
	p->cache_list   = NULL;
	p->state        = ST_IDLE;

	memset(p->aha_out,0,sizeof(p->aha_out));
/*	init aha buffer */

	p->reqptr->d_busy = 1;
	aha_init(p);

	p->function = GET_STATUS;
	(*(p->event_hdlr))(p,IORequest);

disp_str("wait for status");
	if (flags & 01)
		Receive(p->proxyid,0,0);
	return	p->reqptr->d_retval;
}





volatile	long	io_started = 0;
volatile	long	io_finished = 0;

volatile	int		npar = 0;
volatile	int		max_par = 0;


int	io_req_start()
{
int	psw;
	psw = disable();
	++io_started;
	if (++npar > 1) {
		if (npar > max_par)
			max_par = npar;
	}
	restore(psw);
}

int	io_req_end()
{
int	psw;
	psw = disable();
	++io_finished;
	--npar;
	restore(psw);
	return 0;
}

int __far drvio(unsigned disknum, unsigned mode,
                unsigned nblocks, unsigned rsvd,
                struct _block __far *bp)
{
SCSI	*p;

	if (rsvd)
		;
	io_req_start();

	if (disknum >= MAX_NDRIVES || nblocks > MaxIO)
		return -1;
/*
 *	Note: you don't need C++ or smalltalk to get virtual functions.
 */
	p = devtab + disknum;
	p->reqptr->d_busy = 1;
	p->nblocks = nblocks;
	p->blockno = bp->b_block - 1;
	p->function = mode;
	p->cache_list = bp;
	(*(p->event_hdlr))(p,IORequest);

	return 0;
}


/*
 * These routines are not currently used, but later functionality may be 
 * added.
 */




/*
 *	Control the disk
 *
 *	Prospective control functions:  ???
 *		format
 *		driver_info - irq, port, statistics, errors detected, etc.
 */
int __far drvctl(unsigned disknum, unsigned request,
				 unsigned rsvd1, unsigned rsvd2,
				 void __far *cdata)
{
	if (disknum || request || rsvd1 || rsvd2 || cdata)
		;
	return -1;
}

int __far drvctmask(unsigned disknum, unsigned request,
				unsigned rsvd1, unsigned rsvd2)
{
	if (disknum || request || rsvd1 || rsvd2)
		;
	return 0;
}

/*
 *	Terminate the driver - disconnect the interrupt handler.
 */
 
int __far drvterm(unsigned flags, int rsvd1, int rsvd2, int rsvd3)
{
	if (flags || rsvd1 || rsvd2 || rsvd3)
		;
	qnx_hint_detach(ctl_interrupt);
	return 0;
}

/*
 * driver_disk_close
 * 
 *  Mark stored driver information as invalid and continue.
 *
*/

int __far drvclose(int disknum,int rsvd1, int rsvd2, int rsvd3)
{
	if (disknum || rsvd1 || rsvd2 || rsvd3)
		;
	return 0;
}




pid_t	fsys_event(SCSI *dp, int input)
{
pid_t	rvalue=0;

disp_str("event:");
disp_hex(dp->function);
disp_hex(dp->state);
disp_hex(input);
	switch (action[dp->state][input]) {
	case	START_IO:
		dp->mbox = alloc_mbox(dp->target);
		switch (dp->function) {
		case	GET_STATUS:
			dp->state = ST_WAIT_INQ;
			dp->aha_out[CCB_CMDLEN] =
					scsi_inquiry(dp->lun, dp->scsicmd);
			aha_setup(dp,sizeof(inquiry));
			break;
		case	READ_BLK:
			dp->state = ST_WAIT_IO;
			dp->aha_out[CCB_CMDLEN]=
				scsi_read(dp->lun, dp->scsicmd, dp->blockno, dp->nblocks);
			aha_setup(dp,dp->nblocks*_BLOCK_SIZE);
			break;
		case	WRITE_BLK:
			dp->state = ST_WAIT_IO;
			/*	copy data to dma buffer */
			cache_to_buf(dp->iobuf.bufp, dp->cache_list, dp->nblocks);
			dp->aha_out[CCB_CMDLEN]= 
				scsi_write(dp->lun, dp->scsicmd,dp->blockno, dp->nblocks);
			aha_setup(dp,dp->nblocks * _BLOCK_SIZE);
			break;
		default:
			ll_debug();
			break;
		}
		aha_cmd(AHA_SCSI_CMD,dp->aha_out);
		break;
	case	ISSUE_TEST:
		free_mbox(dp->mbox);
		dp->retry_count = 0;
		dp->state = ST_WAIT_START;
		dp->aha_out[CCB_CMDLEN] = scsi_test_rdy(dp->lun,dp->scsicmd);
		dp->mbox = alloc_mbox(dp->target);
		aha_setup(dp,0);
		aha_cmd(AHA_SCSI_CMD, dp->aha_out);
		break;

	case	ISSUE_CAP:
		dp->state = ST_WAIT_CAPAC;
		free_mbox(dp->mbox);
		dp->aha_out[CCB_CMDLEN] = scsi_capacity(dp->lun, dp->scsicmd);
		dp->mbox = alloc_mbox(dp->target);
		aha_setup(dp,8);
		aha_cmd(AHA_SCSI_CMD, dp->aha_out);
		break;

	case	GOT_PARMS:
		free_mbox(dp->mbox);
		dp->state = ST_IDLE;
		rvalue = dp->proxyid;
		dp->reqptr->d_type = _HARD;
		swab_dword((uchar *) &(dp->reqptr->d_num_blks),
						  (uchar __far *)dp->iobuf.bufp);
disp_str("num blocks=");
disp_lhex(dp->reqptr->d_num_blks);
		dp->reqptr->d_sctr_trk = 32;
		dp->reqptr->d_sctr_cyl = 32*64;
		dp->reqptr->d_cyls = dp->reqptr->d_num_blks / dp->reqptr->d_sctr_cyl;
		dp->reqptr->d_heads = 64;
		dp->reqptr->d_sctr_base = 1;

		dp->reqptr->d_nblks = 0;
		dp->reqptr->d_retval = 0;
		dp->reqptr->d_busy = 0;
		break;

	case	IO_DONE:
		dp->state = ST_IDLE;
		free_mbox(dp->mbox);
#ifndef TEST_SPEED
		if (dp->function == READ_BLK){
			buf_to_cache(dp->cache_list, dp->iobuf.bufp, dp->nblocks);
		}
#endif
		rvalue = dp->proxyid;
		dp->reqptr->d_retval = 0;
		dp->reqptr->d_nblks = dp->nblocks;
		dp->reqptr->d_busy =0;
		io_req_end();
		break;

	case	TX_ERROR:
		ll_debug();
		break;

	default:
		ll_debug();
		break;
	}
	return rvalue;
}

/*	
	for now, a target|lun is one-to-one with mailbox entries.
	This will likely change in the future, and some of these
	fields will have to be updated at commit time.
*/

aha_init(SCSI *p)
{
	p->aha_phys = phys_addr(MK_FP(my_ds(), p->aha_out));
}

aha_setup(SCSI *dp, int len)
{
int	mboxind = dp->mbox + outbox_base;
uchar	*p = dp->aha_out;

disp_str("len=");
disp_hex(len);
disp_str("mbox=");
disp_hex(mboxind);

	memset(p+CCB_DATALEN, 0, 3);
	memset(p+CCB_LINKPTR, 0, 3);
	swab_24(p+CCB_DATAPTR, (uchar *)&(dp->iobuf.phys_addr));

	p[CCB_OPCODE] = CCB_INITR;	/* initiator */
	p[CCB_ADDR] = (dp->target << 5 ) | dp->lun;
	p[CCB_SENSE] = CCB_REQSENSE_MAX;
	p[CCB_DATALEN] = 0;
	p[CCB_DATALEN+1] = len >> 8;
	p[CCB_DATALEN+2] = len;

	swab_24(mboxes[mboxind]+MB_CCBPTR,(uchar *)& dp->aha_phys);
	mboxes[mboxind][0] = MBO_START;

disp_str("dlen=");
disp_hex(p[CCB_DATALEN]);
disp_hex(p[CCB_DATALEN+1]);
disp_hex(p[CCB_DATALEN+2]);
}
