

#ifdef __USAGE
%C	[-n name] [-i interrupt] [-a address] [-m maxio]
where:
    -n device name   name of device (hd)
    -i interrupt     interrupt to attach (6)
    -a controller    address of controller (1502)
    -m MaxIO         set maximum io size
    -x on[,off]      set bus on/off duty cycle (5/9)
    -t transfer      set transfer speed
#endif
/*
	Qnx 4.0 SCSI Driver.
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <i86.h>
#include <sys/name.h>
#include <sys/irqinfo.h>
#include <sys/kernel.h>
#include <sys/proxy.h>
#include <sys/sys_msg.h>

#include	<diskio.h>

#include	"endian.h"
#include	"genscsi.h"
#include	"aha1540.h"
#include	"ahaextrn.h"


#ifdef DEBUGGING
#include <sys/osinfo.h>
struct _osinfo _osdata;
static	int	debug_sel = -1;
#endif

#ifndef LOCAL
#define	LOCAL	static
#endif

/* 
	Stack used by fsys when it calls the driver, 
	initial stack pointer is set to be 4-byte aligned.
*/

static	char	stack[STACK_SIZE + sizeof(long)];
static	char	*init_sptr = stack + (sizeof(stack) & ~ 003);	



LOCAL	char	*fsys_name=_FSYS_NAME;
LOCAL	struct	_sysmsg_version_reply	version_stamp = {
	"AHA SCSI",
	__DATE__,
	400,
	' ',
	0
	};

int __far drvinit(pid_t driver_proxy, short tabsize, unsigned __far *cachetab);
int __far drvterm(unsigned flags, int rsvd1, int rsvd2, int rsvd3);
int __far drvopen(int disknum, int flags, int rsvd1, int rsvd2);
int __far drvclose(int disknum,int rsvd1, int rsvd2, int rsvd3);
int __far drvio(unsigned __disknum, unsigned mode,
                unsigned num_blocks, unsigned rsvd,
                struct _block __far *bp);
int __far drvctl(unsigned disknum, unsigned request,
				 unsigned rsvd1, unsigned rsvd2,
				 void __far *cdata);
int __far drvctmask(unsigned disknum, unsigned request,
				unsigned rsvd1, unsigned rsvd2);

void (* (etable[_DRIVER_TBL_SIZE]))() = {
		drvinit,
		drvterm,
		drvopen,
		drvclose,
		drvio,
		drvctl,
		drvctmask,
	};




pid_t	FsysPid;


process_args(int argc, char **argv) 
{
extern	int	optind;
extern	char	*optarg;
char	*p;
int	c;
int	rows = 15, cols = 80;

	while ((c=getopt(argc,argv,"a:b:i:n:m:r:c:x:t:")) != EOF) {
		switch (c) {
		case	'a':
			base_address = strtol(optarg,NULL,16);
			break;
		case	'i':
			ctl_interrupt  = strtol(optarg,NULL,0);
			break;
		case	'n':
			if (strlen(optarg) > sizeof(driver_name)-1) {
				strcpy(driver_name,optarg);
			}
			break;
		case	'm':
			MaxIO = strtol(optarg, NULL,0);
			break;
		case	'b':
			num_boxes = strtol(optarg,NULL,0);
			inbox_base = num_boxes;
			break;
		case	'r':
			rows = strtol(optarg,NULL,0);
			break;
		case	'c':
			cols = strtol(optarg,NULL,0);
			break;
		case	'x':
			bus_on_time = strtol(optarg,&p,0);
			if (*p == ',') {
				bus_off_time = strtol(p+1,&p,0);
			}
			break;
		case	't':
			set_transfer_speed = strtol(optarg,NULL,0);
			break;
		default:
			fprintf(stderr,"unknown argument ignored\n");
		}
	}
#ifdef DEBUGGING
	if (debug_sel == -1) {	/*	not a valid selector! */
		qnx_osinfo(0, &_osdata);
		debug_sel = (_osdata.primary_monitor & 001) ? MONO_DISP : COLOUR_DISP;
	}
	printf("enabling debug at selector %x\n",_osdata.machinesel);
	set_display(MK_FP(debug_sel,0),rows,cols);
#endif
	return 0;
}


main(int argc, char **argv) 
{


	process_args(argc, argv);

	if ((FsysPid = qnx_name_locate(0,fsys_name,0,0)) == -1) {
		perror(_FSYS_NAME);
		fputs("Cannot locate filesystem\n",stderr);
		return EXIT_FAILURE;
	}

	if (fsys_mount_driver(FsysPid, MAX_NDRIVES, disktab, etable,
						MaxIO, MAX_PARALLEL,init_sptr,&version_stamp) == -1) {
		perror("fsys mount");
		fputs("cannot mount onto filesystem\n",stderr);
		return EXIT_FAILURE;
	}	
	return EXIT_SUCCESS;
}


#ifdef TEST_SPEED
int	test_speed(unsigned disknum)
{
SCSI	*p;
long	x;
long	last_io;

	if (disknum >= MAX_NDRIVES)
		return -1;

/*
 *	Note: you don't need C++ or smalltalk to get virtual functions.
 */
	p = devtab + disknum;
	p->proxyid = qnx_proxy_attach(0,NULL,0,-1);

	start_clock();

	for (x=0; x + MaxIO < p->reqptr->d_num_blks -1 ; x += last_io) {
		p->reqptr->d_busy = 1;
		p->nblocks = MaxIO;
		p->blockno = x;
		p->function = 0;
		(*(p->event_hdlr))(p,IORequest);
		Receive(p->proxyid,0,0);
		last_io = p->reqptr->d_nblks;
	}
	end_clock();
	put_delta();
	return 0;
}




#include <sys/types.h>
#include <sys/timers.h>
#include <time.h>

#define ONE_MILLION		1000000L
#define	ONE_BILLION		1000000000L

struct	timespec	tp1, tp2;

#define secs1	tp1.tv_sec
#define secs2	tp2.tv_sec
#define nsecs1	tp1.tv_nsec
#define nsecs2	tp2.tv_nsec

start_clock()
{
	getclock( TIMEOFDAY, &tp1 );
}


end_clock()
{
	getclock( TIMEOFDAY, &tp2 );
}

#define	abs_diff(_x,_y)	(((_x) < (_y)) ? ((_y) - (_x)) : ((_x) - (_y)))

put_delta()
{
long	diff_s, diff_ns;
long	diff_ms;

	diff_s =  secs2 - secs1;
	if (nsecs2 < nsecs1) {
		diff_s--;
		diff_ns = nsecs2 + ONE_BILLION - nsecs1;
	} else {
		diff_ns = nsecs2 - nsecs1;
	}
	diff_ms = diff_ns/ONE_MILLION + (diff_ns % ONE_MILLION != 0);
	if (diff_ms >= 1000) {
		diff_s ++;
		diff_ms -= 1000;
	}
    printf("%ld.%03ld\n",diff_s,diff_ms);
}


#endif






