
/*
 *	Manifests and structures for floppy driver.
 */


#include <stddef.h>
#include <sys/qioctl.h>

/*
 *   Startup parameters.
 */

#define DISK_TYPE		_FLOPPY /* 0-floppy, 1-fixed, 2-ramdisk, ...*/
#define MAX_NDRIVE		2		/*	Maximum number of physical devices per controller */
#define MAX_LOGICAL		1		/*	Maximum number of logical devices per device */
#define	MAX_IO			36		/*	Largest single request I/O amount		*/
#define MAX_PARALLEL	1		/* 	Maximum pending operations				*/
#define	STACK_SIZE		64		/* 	Size of locally allocated stack			*/
#define CTL_ADDRESS		0x3f0	/* 	Control address for disk controller		*/
#define CTL_INTERRUPT	6		/* 	Interrupt line for disk controller		*/


#define	MAX_RETRY	5

/*	
	these are timer values, in 100 msec quanti.	
*/

#define	READ_SPINUP		7
#define	WRITE_SPINUP	11
#define	WAIT_SPINDOWN	25
#define	SPINDOWN		10
#define	XFER_TIMEOUT	20

/*
 *	Definition of a diskette information entry
 */

struct _floppy_info {
	int		fd_drive_type;		/*	Floppy drive type */
	int		fd_sctr_trk;		/*	Sectors per track */
	int		fd_cyls;			/*	Number of cylinders */
	int		fd_last_cyl;		/*	Last cylinder accessed */
	int		fd_flags;			/*	Flags */
	int		fd_clock;			/*	Clock rate */
	int		fd_rd_gap;			/*	Read gap */
	int		fd_fmt_gap;			/*	Format gap */
	int		fd_srt;				/*	Step Rate Time */
	int		fd_hlt;				/*	Head Load Time */
	int		fd_hut;				/*	Head Unload Time */
	int		fd__zero1;			/*	spare */
};

typedef struct _floppy_info FLOPPY_INFO;




/*
 *	Driver state definitions
 */

enum {

	ST_START=0,
	ST_RESET,
	ST_IDLE, 	
	ST_SPIN,
	ST_SEEK, 	
	ST_XFER,	
	ST_READY,
	ST_SPECIFY,
	ST_SPECIFY2,
	ST_FMT_TRK,
	ST_RTRAP,
	ST_ITRAP,
	ST_TTRAP,
	MAX_STATE,
};

enum {			/*	Inputs	*/

	IORequest = 0,
	TimeOut,
	Interrupt,
	Event_Reserved,	/*	brings it to 4 -- even multiple */
	MAX_EVENT

	};



extern void fdc_init(int base_address);
extern int fdc_direction(void);
extern long fdc_last_block(int sect_cyl,int sect_trk);
extern void fdc_setclock(int clock_speed);
extern int fdc_motor_on(int driveno);
extern int fdc_motor_off(int driveno);
extern int fdc_status(void);
extern int fdc_cmd(int no_wait);
extern int fdc_reset(void);
extern int setup_recalibrate(int driveno);
extern int setup_seek(int drive,int cyl, int shift);
extern int seek_ok(void);
extern int setup_xfer(int mode,int drive,int head,int track,int cyl,
						int sect_trk,int read_gap);
extern int setup_sense_intr(void);
extern void setup_sense_drive(int drive,int head);
extern int fdc_specify(int step,int unload,int load);

extern int setup_fmt_trk(int, struct fd_fmt_trk __far *,
						short, unsigned char __far *);
extern int fdc_fmt_stat(struct fd_fmt_trk __far *);

extern int nec_setres(struct fd_last_err __far *);

extern int fdc_fmt_ok(void);
extern int fdc_xfer_ok(void);





#ifdef	DEBUGGING
static	char	*ntab[] = {			/*	states */
			"ST_VIRGIN",
			"ST_IDLE",	
			"ST_SPIN",	
			"ST_SEEK",	
			"ST_XFER",	
			"ST_READY",
			"ST_SPECIFY",
			"ST_SPECIFY2",
			"ST_FMT_TRACK",
			"ST_RTRAP",
			"ST_ITRAP",
			"ST_TTRAP",
			NULL
	};

static	char	*etab[] = {			/*	inputs	*/
			"IORequest",
			"TimeOut",
			"Interrupt"
	};


#endif


