

/*
 *	This header file contains generic scsi information.
 *  The first cut will probably contain *too* much adaptec
 *  peculiarities, but later ones will be more general.
 *
 */
#define	READ_BLK	0
#define	WRITE_BLK	1
#define	GET_STATUS	2


#define	STACK_SIZE	256
#define	MAX_PARALLEL 1
#define	MAX_NDRIVES	 8
#define	MAX_MBOX	 16


/*
 *	SCSI Command Code definitions
 *
 *	The first uchar of the CDB is the operation code.  The top 3 bits
 *	are the group code field and the bottom 5 bits are the command code
 *	field.
 *
 *	Group 0 - six-uchar commands
 *	Group 1 - ten-uchar commands
 *	Group 2 - ten-uchar commands
 *	Group 3 - reserved
 *	Group 4 - reserved
 *	Group 5 - twelve-uchar commands
 *	Group 6 - vendor specific
 *	Group 7 - vendor specific
 */

#define FBLK_SIZE	512		/*	File system block size */
#define DBLK_SIZE	4096	/*	Max device block size */
#define TARG_MAX	8		/*	Max number of SCSI targets */
#define LUN_MAX		8		/*	Max number of LUNs per target */

#define M_GROUP(cmd)	((cmd)>>5)

/*
 *	Group 0 Command Codes 
 */
#define SC_UNIT_RDY	0x00	/*	Test Unit Ready */
#define SC_REZERO	0x01	/*	Rezero Unit */
#define SC_RQ_SENSE	0x03	/*	Request Sense */
#define SC_FORMAT	0x04	/*	Format unit */
#define SC_REASSIGN	0x07	/*	Reassign Blocks */
#define SC_READ		0x08	/*	Read */
#define SC_WRITE	0x0a	/*	Write */
#define SC_SEEK		0x0b	/*	Seek */
#define SC_INQUIRY	0x12	/*	Inquiry */

struct _group0 {
	uchar	g0_opcode;		/*	Operation Code */
	be_24	g0_block;		/*	Block address (top 3 bits contain LUN) */
	uchar	g0_datalen;		/*	Data length (xfer|params|alloc) */
	uchar	g0_control;		/*	Control uchar */
};

#define LEN_G0	sizeof(struct _group0)

#define SC_MSELECT	0x15	/*	Mode Select */
#define SC_COPY		0x18	/*	Copy */
#define SC_MSENSE	0x1a	/*	Mode Sense */
#define SC_SPINDLE	0x1b	/*	Start/Stop Unit */
#define SC_RCV_DIAG	0x1c	/*	Receive Diagnostic Results */
#define SC_SND_DIAG	0x1d	/*	Send Diagnostic */
#define SC_REMOVAL	0x1e	/*	Prevent/Allow Medium Removal */		

/*
 *	Group 1 Command Codes
 */
#define SC1_RD_CAP	0x25	/*	Read Capacity */
#define SC1_READ	0x28	/*	Read */
#define SC1_WRITE	0x2a	/*	Write */
#define SC1_SEEK	0x2b	/*	Seek */
#define SC1_WR_VER	0x2e	/*	Write with Verify */
#define SC1_VERIFY	0x2f	/*	Verify */
#define SC1_SRCH_HI	0x30	/*	Search Data High */
#define SC1_SRCH_EQ	0x31	/*	Search Data Equal */
#define SC1_SRCH_LO	0x32	/*	Search Data Low */
#define SC1_LIMITS	0x33	/*	Set Limits */
#define SC1_COMPARE	0x39	/*	Compare */
#define SC1_CPY_VER	0x3a	/*	Copy and Verify */
#define SC1_WR_BUF	0x3b	/*	Write Buffer */
#define SC1_RD_BUF	0x3c	/*	Read Buffer */

/*
 *	Group 1 and 2 - ten-uchar
 */
struct _group1 {
	uchar	g1_opcode;		/*	Operation Code */
	uchar	g1_lun;			/*	Logical Unit Number */
	be_32	g1_block;		/*	Block address */
	uchar	g1_reserved;
	be_16	g1_datalen;		/*	Data length (xfer|params|alloc) */
	uchar	g1_control;		/*	Control uchar */
};
#define LEN_G1	sizeof(struct _group1)

/*
 *	Group 5 - twelve-uchar
 */
struct _group5 {
	uchar	g5_opcode;		/*	Operation Code */
	uchar	g5_lun;			/*	Logical Unit Number */
	be_32	g5_block;		/*	Block address */
	be_32	g5_datalen;		/*	Data length (xfer|params|alloc) */
	uchar	g5_reserved;
	uchar	g5_control;		/*	Control uchar */
};
#define LEN_G5	sizeof(struct _group5)
/*
 *	Vendor Specific Command Codes (Groups 6 & 7)
 *	(Sean's WORMs)
 */
#define SCW_STANDBY	0xce	/*	Standby Check */
#define SCW_LOAD	0xcf	/*	Load/Unload Unit */
#define SCW_RWD_DCB	0xd2	/*	Read DCB */
#define SCW_DREAD	0xee	/*	Disk Read */
#define SCW_DWRITE	0xef	/*	Disk Write */

/*
 *	CDB Flags
 */
#define CF_LINK		0x01
#define CF_FLAG		0x02
#define CF_VENDOR0	0x40
#define CF_VENDOR1	0x80

#define CF_FUA		0x08
#define CF_DPO		0x10


/*
 *	Generic Group
 */
struct _groupg {
	uchar	gg_uchar0;
	uchar	gg_uchar1;
	uchar	gg_uchar2;
	uchar	gg_uchar3;
	uchar	gg_uchar4;
	uchar	gg_uchar5;
	uchar	gg_uchar6;
	uchar	gg_uchar7;
	uchar	gg_uchar8;
	uchar	gg_uchar9;
	uchar	gg_uchar10;
	uchar	gg_uchar11;
};


/*
 *	INQUIRY Data flags
 */
#define ID_RMB		0x80	/*	Removable Medium Bit */
#define ID_AENC		0x80	/*	Asynchronous Event Notificaton Capability */
#define ID_TRM_IOP	0x40	/*	Terminate I/O Process bit */
#define ID_RDF		0x0f	/*	Response Data Format mask */
#define ID_REL_ADR	0x80	/*	Relative Addressing bit */
#define ID_WBUS32	0x40	/*	Wide Bus 32 (32 bit bus) support */
#define ID_WBUS16	0x20	/*	Wide Bus 16 (16 bit bus) support */
#define ID_SYNC		0x10	/*	Synchronous Data Transfers supported */
#define ID_LINKED	0x08	/*	Linked commands supported */
#define ID_RES		0x04	/*	(reserved) */
#define ID_CMD_QUE	0x02	/*	Tagged Command Queuing supported */
#define ID_SFT_RST	0x01	/*	Soft RESET alternative supported */

/*
 *  Definition of the SCSI-2 INQUIRY data format
 *	(only the first 3 uchars are valid for SCSI-1)
 */
typedef struct _scsi_inquiry {
	uchar	id_peripheral;	/*	Peripheral Qualifier/Device Type */
	uchar	id_rmb;			/*	Removable Medium Bit */
	uchar	id_ver;			/*	Versions (ISO, ECMA, ANSI) */
	uchar	id_rdf;			/*	Flags and Response Data Format */
	uchar	id_adlen;		/*	Additional Data Length */
	uchar	id_res1;
	uchar	id_res2;
	uchar	id_flags;		/*	Flags */
	char	id_vend_id[8];	/*	Vendor identification */
	char	id_prod_id[16];	/*	Product identification */
	char	id_prod_rev[4];	/*	Product revision level */
	char	id_vend_spc[20];/*	Vendor Specific */
	char	id_res3[40];
} inquiry;

/*
 *	Sense Keys
 */
#define SK_NO_SENSE		0		/*	No sense data (no error) */
#define SK_RECOVERED	1		/*	Recovered error */
#define SK_NOT_RDY		2		/*	Device not ready */
#define SK_MEDIUM		3		/*	Medium error */
#define SK_HARDWARE		4		/*	Hardware error */
#define SK_ILLEGAL		5		/*	Illegal Request (bad command) */
#define SK_UNIT_ATN		6		/*	Unit Attention */
#define SK_DATA_PROT	7		/*	Data Protect */
#define SK_BLNK_CHK		8		/*	Blank Check */
#define SK_VENDOR		9		/*	Vendor Specific */
#define SK_CPY_ABORT	10		/*	Copy Aborted */
#define SK_CMD_ABORT	11		/*	Aborted Command */
#define SK_EQUAL		12		/*	Equal */
#define SK_VOL_OFL		13		/*	Volume Overflow */
#define SK_MISCMP		14		/*	Miscompare */
#define SK_RESERVED		15		/*	Reserved */

/*
 *	Definition of the SCSI-2 Sense Data Structure
 *	(similar to SCSI-1 extended sense data)
 */
typedef struct _scsi_sense {
	uchar	sd_err;			/*	Error code and Valid flag */
	uchar	sd_seg;			/*	Segment number */
	uchar	sd_sense;		/*	Sense key & Flags */
	be_32	sd_info;		/*	Information (32bit big-endian value) */
	uchar	sd_asl;			/*	Additional Sense Length */
	be_32	sd_csinfo;		/*	Command-Specific Information */
	uchar	sd_asc;			/*	Additional Sense Code */
	uchar	sd_ascq;		/*	Additional Sense Code Qualifier */
	uchar	sd_fruc;		/*	Field Replaceable Unit Code */
	uchar	sd_sks_flags;	/*	Sense Key Specific Flags */
	be_16	sd_sks_data;	/*	Sense Key Specific Data (16bit big-endian) */
	uchar	sd_asb[2];		/*	Additional Sense uchars (Max 256-18) */
} sense;

/*
 *	SCSI Device Type definitions
 *	(types 0Ah - 1Eh are reserved)
 */
#define D_DIR_ACC	0		/*	Direct Access device */
#define D_SEQ_ACC	1		/*	Sequential Access device */
#define D_PRINTER	2		/*	Printer device */
#define D_PROCESSOR	3		/*	Processor device */
#define D_WORM		4		/*	Write Once Read Multiple device */
#define D_READ_ONLY	5		/*	Read Only (CD-ROM) device */
#define D_SCANNER	6		/*	Scanner device */
#define D_OPTICAL	7		/*	Optical Memory device */
#define D_MEDIA_CHG	8		/*	Media Changer device */
#define D_COMM		9		/*	Communication device */
#define D_UNKNOWN	0x1f	/*	Unknown device type */


/*
 *	Definition of driver's Target Descriptor Block
 */
typedef struct _tdb {
	unsigned	td_id;			//	Target ID 
	unsigned	td_luns;		//	LUNs 
	unsigned	td_flags;		//	Flags 
	uchar		td_version;		//	SCSI Version 
	uchar		td_devtype;		//	Device Type 
	uchar		td_fblk_dblk;	//	Number of fsys blocks per device block 
	uchar		td_shift;		//	Shift factor for blocks > 512 uchars 
	uchar		td_mask;		//	Mask used for blocks > 512 uchars 
	uchar		td_spare;
	long		td_numblocks;	//	Number of blocks 
	unsigned	td_blksize;		//	Block size (in bytes) 
	unsigned	td_blks_cyl;	//	Blocks per cylinder 
	unsigned	td_err_sense;	//	Error Sense Key 
	unsigned	td_err_asc;		//	Additional Sense Code and Qualifier (hi/lo) 
	long		td_err_blk;		//	Block number of error (if applicable) 
	unsigned	td_spare2[2];
} tdb;




/*
	

*/


#define	SCSI_OP(grp_code,cmd_code)	(((grp_code)<<5)|(cmd_code))

