/*
	Qnx 4.0 Floppy Disk Driver.
	
	ToDo:
		check out format.
	

*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <i86.h>
#include <sys/name.h>
#include <sys/irqinfo.h>
#include <sys/kernel.h>
#include <sys/proxy.h>

#include <diskio.h>

#include "floppy.h"
#include "nec765.h"


#ifndef DEBUGGING
#define DEBUGGING 0
#endif

#if DEBUGGING
#define	LOCAL	 static
#else
#define	LOCAL
#endif

extern vputc(int ch);

#define	DELAY_STATUS_REG	usec_delay(12)


LOCAL	unsigned char	Command[32];
LOCAL	unsigned char	Result[32];
LOCAL	int		ResultLen = 0;

LOCAL	int		MotorStatus = 0;
LOCAL	struct fdc_regs 	Fdc;

LOCAL const int fdc_command_length[] =	{
				0,	0,	9,	3,	2,	9,	9,	2,
				1,	9,	2,	0,	9,	6,	0,	3,
				0,	9,	0,	0,	0,	0,	0,	0,
				0,	9,	0,	0,	0,	9,	0,	0
	};

/*
	-1 invalid command.
	0 = no status.
	1 = st0,1,2;C,H,R,N.
	2 = st3
	3 = st0,PCN

*/

LOCAL const int fdc_cmd_result[] = {
				-1,  -1,   1,   0,   2,   1,   1,   0,
        		 3,   1,   1,  -1,   1,   1,  -1,   0,
        		-1,   1,  -1,  -1,  -1,  -1,  -1,  -1,
        		-1,   1,  -1,  -1,  -1,   9,  -1,  -1
	};


static	nec_st1(unsigned short st1)
{
	if (st1 & ST1_EN) 		return FD_NOT_FOUND;
	if (st1 & ST1_DE) 		return FD_DATA_ERROR;
	if (st1 & ST1_OR) 		return FD_DMA_FAIL;
	if (st1 & ST1_ND) 		return FD_NOT_FOUND;
	if (st1 & ST1_NW)		return FD_WRITE_PROTECT; 
	if (st1 & ST1_MA)		return FD_NOT_FOUND;
	return FD_UNKNOWN;
}

static	nec_st3(unsigned short st3)
{
	if (st3 & ST3_WP)		return FD_WRITE_PROTECT;
	return FD_UNKNOWN;
}

nec_setres(struct fd_last_err __far * fderr)
{
	fderr->head = fderr->track = fderr->sector = 0;
	switch (fdc_cmd_result[Command[0] & CMD_MASK]) {
	case	-1:
		return fderr->result = FD_UNKNOWN;
	case	0:
		fderr->result = FD_OK;
		break;
	case	1:
		fderr->head   = Result[4];
		fderr->track  = Result[3];
		fderr->sector = Result[5];
		if ((Result[0] & ST0_IC_MSK) == 0) { 
			return	fderr->result = FD_OK;
		}
		return fderr->result = nec_st1(Result[1]);
		break;
	case	2:
		return fderr->result = nec_st3(Result[0]);
	case	3:
		return fderr->result = (Result[0] & ST0_IC_MSK) ? FD_UNKNOWN : FD_OK;    	
	}
}


#if DEBUGGING
void	disp_Command()
{
int	r;
int	i;
	disp_char('<');
	r = fdc_command_length[Command[0] & CMD_MASK];
	for (i=0; i < r; i++)
		disp_hex(Command[i]);
	disp_char('>');
}

void disp_Result()
{
int	i;
	disp_char('!');
	for (i=0; i < 8; i++) {
		disp_hex(Result[i]);
	}
	disp_char('!');
}

#endif



void fdc_init(int base_addr)
{
	Fdc.dout   = base_addr + FD_DOR;
	Fdc.status = base_addr + FD_STATUS;
	Fdc.dioreg = base_addr + FD_DIOR;
	Fdc.fdisk  = base_addr + FD_FDR;
	Fdc.din    = base_addr + FD_DIR;
	Fdc.fdclock= base_addr + FD_DCR;
}


/*
	check the No Data flag.
*/

fdc_fmt_ok()
{
	if (Result[0] & (ST0_IC_MSK|ST0_EC|ST0_NR)) {
		return 1;
	}
	if (Result[1] & ST1_NW) {
#if DEBUGGING 
		disp_str("write protect, no error!\n");
        ll_debug();
#endif
	}
	return 0;
}

fdc_xfer_ok()
{
		/*	transferred nothing! */
	if (Result[0] && ST0_IC_IC)
		return -1;
	return (Result[1] && 0x04) == 0 ? 0 : -1;
}


fdc_fmt_stat(struct fd_fmt_trk __far *fp)
{
	return (Result[0] && ST0_IC_MSK) == 0 ?	0 : -1;
}

long	 fdc_last_block(int sectr_cyl, int sectr_trk)
{
	return Result[3]*sectr_cyl + Result[4]*sectr_trk + Result[5];
}


/*
 *	Transfer the specified number of blocks
 */
setup_xfer(int mode, int unit, 
			int head, int track, int sect, int nsect, int rgap)
{

	Command[0] = FD_MULTI_TRACK | FD_MFM |
			((mode == WRITE_BLK) ? FD_WRITE : (FD_READ|FD_SK));
	Command[1] = (head << 2) | unit;
	Command[2] = track;
	Command[3] = head;
	Command[4] = sect;
	Command[5] = _BLOCK_SIZE >> 8;	/*	Bytes per sector / 256 */
	Command[6] = nsect;				/*	Last sector on track */
	Command[7] = rgap;				/*	Read gap */
	Command[8] = 0xff;				/*	Data Length */
	return 0;
}


/*
 *	Recalibrate the drive
 */
setup_recalibrate(int drive)
{
	Command[0] = FD_RECAL;
	Command[1] = drive;
	return	0;
}
/*
 *	Seek to the specified cylinder
 */
setup_seek(int drive, int cyl, int shift)
{

	Command[0] = FD_SEEK;
	Command[1] = drive;
	Command[2] = cyl << shift;
	return 0;
}





setup_sense_intr()
{
	*Command = FD_INTSTAT;
	return	 0;
}


void setup_sense_drive(int drive, int head)
{
	Command[0] = FD_DRVSTAT;
	Command[1] = (head ? 0x04 : 0) | (1 << drive);
}

fdc_specify(int step, int unload, int load)
{

	Command[0] = FD_SPECIFY;
	Command[1] = (step << 4) | unload;
	Command[2] = load << 1;
	if (fdc_cmd(1) < 0) {
		vputc('&');
#if DEBUGGING
		ll_debug();
#endif
		return -1;
	}
	return 0;
}



setup_fmt_trk(int	driveno, struct fd_fmt_trk __far *p,  
				short fmt_gap,	unsigned char __far *fmtptr)
{
int	i;

	for (i=0; i < p->nsector; i++) {
		*fmtptr++ = p->trackno;
		*fmtptr++ = p->head;
		*fmtptr++ = p->xtab[i];	
		*fmtptr++ = p->sectsize >> 8;
	}
	Command[0] = FD_FORMAT | FD_MFM;		
	Command[1] = (p->head << 2) | driveno;
	Command[2] = p->sectsize >> 8;
	Command[3] = p->nsector;
	Command[4] = fmt_gap;
	Command[5] = p->filler;	
	return 0;
}


seek_ok()
{
	setup_sense_intr();
	if (fdc_cmd(0) < 0) {
		vputc('*');
		ll_debug();
	}

	if (ResultLen != 2)  {
#if DEBUGGING
		ll_debug();
#endif
		return -1;
	}
	if (*Result & ST0_IC_MSK) {
#if DEBUGGING
		ll_debug();
#endif
		return -1;
	}
	if ((*Result & ST0_SE) == 0) {
#if DEBUGGING
		ll_debug();
#endif
		return -1;
	}
	return 0;
}




void fdc_setclock(int clock)
{
	io_out(Fdc.fdclock, clock);
}

fdc_motor_on(int drive)
{
int	psw;

	psw = disable();

	io_out(Fdc.dout, 
			MotorStatus = (0x10 << drive) | drive | FD_EID | FD_RST);

	restore(psw);

	return 0;
}

fdc_motor_off(int driveno)
{
int  	psw;
	psw = disable();
	io_out(Fdc.dout,MotorStatus = FD_EID | FD_RST);
	restore(psw);
	return 0;
}


fdc_direction()
{
volatile int tickit = 16000;
	while (tickit--) {
		switch (io_in(Fdc.status) & (FD_RQM|FD_DIO)) {
		case	0: 				continue;	/*	not ready */
		case	FD_DIO|FD_RQM:	return 1;	/*	incomming */
		case	FD_RQM:			return 2;	/*  outgoing */
		default:
			continue;
		}
	}
	return 0;
}

fdc_status()
{
int	i = 0;
int	x;

	ResultLen = 0;
	while (1) {
		if (fdc_direction() != 1) {
#if DEBUGGING
			ll_debug();
#endif
			return -1;
		} 
		x = io_in(Fdc.dioreg);
		if (i < 7)
			Result[i++] = x;
		DELAY_STATUS_REG;
		if ((io_in(Fdc.status) & FD_CB) == 0) {
			break;
		}
	}
	if ((ResultLen = i) == 0)
		return 0;
    /*	have give a 'quick' return code for any & all */
	return (Result[0] & (ST0_IC_MSK)) == 0 ? 0 : 1;
}


fdc_cmd(int no_wait)
{
unsigned char	*p = Command;
int		nbytes;

	nbytes = fdc_command_length[ *p & 0x1f ];

	if (fdc_direction() != 2) {
#if DEBUGGING
		ll_debug();
#endif
		return -1;
	}	
	while (nbytes-- && fdc_direction() == 2) {
		io_out(Fdc.dioreg, *p++);
	}
	DELAY_STATUS_REG;
	if (no_wait)
		return 0;
	return fdc_status();
}



fdc_reset()
{
	io_out(Fdc.dout, FD_EID);
	io_out(Fdc.dout, (FD_EID|FD_RST));
	MotorStatus = FD_EID | FD_RST;
	return 0;
}
