
/*
 *	This header defines registers, flag bits, timeouts, for the nec 765
 *	floppy disk controller.
 */




#define	RQM_PERIOD		12

/*
 *	765 Commands and bits
 */

#define	FD_RD_TRACK	2
#define FD_SPECIFY	3			/*	Specify step rate */
#define FD_DRVSTAT	4			/*	Sense Drive Status */
#define FD_WRITE	5			/*	Write Data */
#define FD_READ		6			/*	Read Data */
#define FD_RECAL	7			/*	Recalibrate Drive */
#define FD_INTSTAT	8			/*	Sense Interrupt Status */
#define	FD_WRITEDEL	9			/*	write deleted data */
#define	FD_READID	10			/*  read first correct id on cylinder */
#define	FD_READDEL	12			/*	read deleted data */
#define FD_FORMAT	13			/*	Format */
#define FD_SEEK		15			/*	Seek */
#define	FD_SCANEQ	17			/*	scan equal */
#define	FD_SCANLE	25			/*	scan low or equal	*/
#define	FD_SCNANHE	29			/*	scan high or equal */


#define	FD_MULTI_TRACK	0200	/*	Go ahead and swap head if nec. */
#define FD_MFM			0100	/*	MFM flag */
#define FD_SK			0040	/*	Skip Deleted Data Address Mark flag */

/*
	diskette command info.
*/
#define	FD_NORESULT	0x00000008L	/*	does not generate a result code	*/
#define	FD_IMRESULT 0x00000110L	/*	generates an immediate result	*/
#define	FD_OVERLAP  0x00008080L	/*	can be overlapped on 'n' disks	*/
#define	FD_DISKNUM  0x2202b6f4L	/*	command requires disk number	*/
#define	FD_NOTSURE  0x2202b7e4L	/*	is an error code possible?		*/
#define	FD_VALID	0x2202b7fcL	/*	are valid commands				*/
#define FD_STAT3	0x00000010L /*  generates a st-3 rather than st-0,1,2 */

/*	tests whether a command x has attribute attr */
#define	CMD_MASK	0037

#define	CMD_ATTR(x,attr)	((attr) & (1 << (long)(x & CMD_MASK)))

/*
 *	FDC Control registers and bits
 */
#define FD_DOR		2			/*	Digitial Output Register */
#define FD_STATUS	4			/*	Status Register */
#define FD_DIOR		5			/*	Data Input/Output Register */
#define FD_FDR		6			/*	Fixed Disk Register */
#define FD_DIR		7			/*	Digital Input Register */
#define FD_DCR		7			/*	Diskette Control Register */

struct	fdc_regs {
	int	dout,
		status,
		dioreg,
		fdisk,					/*	not used */
		din,					/*	not used */
		fdclock;
	};


/*
 *	DOR bits
 */
#define FD_D2E		0040		/*	Drive 2 motor Enable */
#define FD_D1E		0020		/*	Drive 1 motor Enable */
#define FD_EID		0010		/*	Enable Interrupts and DMA */
#define FD_RST		0004		/*	Reset FDC */
#define FD_SEL		0001		/*	Drive Select */

#define FD_DMEM		0360		/*	Drive motor enable mask */
#define FD_DSM		0003		/*	Drive select mask */

/*
 *	Status bits
 */
#define FD_RQM		0200		/*	Request for Master */
#define FD_DIO		0100		/*	Data Input/Ouput (1=FDC to CPU) */
#define FD_NDM		0040		/*	Non-DMA Mode */
#define FD_CB		0020		/*	Controller Busy */
#define FD_D2B		0002		/*	Drive 2 Busy (seeking) */
#define FD_D1B		0001		/*	Drive 1 Busy (seeking) */

#define FD_DCM		0003		/*	Diskette Control Mask */

/*
 *	Status Register 0 bits
 */
#define ST0_IC_MSK	0300		/*	Interrupt Code Mask */
#define ST0_IC_AT	0100		/*	IC Abnormal Termination */
#define ST0_IC_IC	0200		/*	IC Invalid Command issue */

#define ST0_SE		0040		/*	Seek End */
#define ST0_EC		0020		/*	Equipment Check */
#define ST0_NR		0010		/*	Not Ready */
#define ST0_HD		0004		/*	Head Address */
#define ST0_US_MSK	0003		/*	Unit Select Mask */
 

#define	ST1_EN		0200		/*	End of cyl */
#define	ST1_DE		0040		/*	Data Error */
#define	ST1_OR		0020		/*	dma overrun */
#define	ST1_ND		0004		/*	no data */
#define	ST1_NW		0002		/*	not writable -- write protected */
#define	ST1_MA		0001		/*	missing address mark */


#define	ST3_FAULT	0200		/* fault signal from fdd */
#define	ST3_WP		0100		/* write protected */
#define	ST3_RDY		0040		/* fd is ready */
#define	ST3_T0		0020		/* fd is on track 0 */
#define	ST3_TS		0010		/* fd is two sided */
#define	ST3_HD		0004		/* head select */
#define	ST3_US_MSK	0003		/* unit select, 0-3 */


