
#ifdef __USAGE

%C	[-a address] [-d dma channel] [-i interrupt] [-m num blocks] [-n name]",
where:",
   -a address       controller address ( 3510)
   -d dma channel   dma channel (5)
   -i interrupt     interrupt to attach (14)
   -m num blocks    maximum blocks per operation (36)
   -n device name   name of device (hd)
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <i86.h>
#include <conio.h>
#include <sys/name.h>
#include <sys/irqinfo.h>
#include <sys/kernel.h>
#include <sys/seginfo.h>
#include "diskio.h"
#include "ps2esdi.h"



#define DRIVER			FIXED_DISK_PREFIX

char	driver_name[_DNAME_PREFIX+1]= DRIVER;

#define	MAKE_VERS(x,y)	((x<<8)|y)

struct	_sysmsg_version_reply	version_stamp = {
		"PS/2 ESDI",
		__DATE__,
		400,
		' ',
		0
	};


/*
	data shared between interrupt handler & call level.
	Parameters, etc.. from call level are remembered here.

*/
static	volatile	IntStatus;		/* bumped when interrupt transaction done */
static	unsigned	Function;		/* Function interrupt handler is doing	*/
static	unsigned	NumBlocks;		/* Number of blocks yet to transfer		*/
static	unsigned	DriveNo = 0;	/* Disk number */
static	DISK		*StatusPtr;		/* Blocks actually transfered so far	*/
static	BLOCK __far *CachePtr;		/* Linked list of blks read/written		*/


/*
	These areas are for the binding between the filesystem and the driver,
	they are 'relative' constants, that is they can only be redefined at
	startup.

*/

static	pid_t		NotifyPid;			/* Proxy for int handler to use	*/
static	pid_t		FsysPid;
static	unsigned	MaxIO				= MAX_IO,
					esdi_addr 		= CTL_ADDRESS,
					intno = CTL_INTERRUPT;

static	int			int_id;

/*
	This buffer is defined when the driver is initialed.
*/
static	int	dma_channel = DMA_CHANNEL;
static	int	dma_flags = DMA_PS2;
static	struct dma_buffer	iobuffer;	

/* 
	Stack used by fsys when it calls the driver, 
	initial stack pointer is set to be 4-byte aligned.
*/
static	char	stack[STACK_SIZE + sizeof(long)];
static	char	*init_sptr = stack + (sizeof(stack) & ~ 003);	


int	max_ndrives = 1;

static DISK	disk[MAX_NDRIVES];	/* Disk parameter tables	*/



/*
 *	Driver functions, a jump table through which fsys enters the driver code
 *
 *	Array of pointers to driver functions.
 */

pid_t __far	int_handler();


char	*fsys_name=_FSYS_NAME;

/*
 *	Driver functions, a jump table through which fsys enters the driver code
 *
 *	Array of pointers to driver functions.
 */

/*
 *  Driver entry points.
 */

pid_t __far int_handler();

int __far drvinit(pid_t driver_proxy, short tabsize, unsigned __far *cachetab);
int __far drvterm(unsigned flags, int rsvd1, int rsvd2, int rsvd3);
int __far drvopen(int disknum, int flags, int rsvd1, int rsvd2);
int __far drvclose(int disknum,int rsvd1, int rsvd2, int rsvd3);
int __far drvio(unsigned disknum, unsigned mode,
                unsigned num_blocks, unsigned rsvd,
                struct _block __far *bp);
int __far drvctl(unsigned disknum, unsigned request,
				 unsigned rsvd1, unsigned rsvd2,
				 void __far *cdata);
int __far drvctmask(unsigned disknum, unsigned request,
				unsigned rsvd1, unsigned rsvd2);

void (* (etable[_DRIVER_TBL_SIZE]))() = {
		drvinit,
		drvterm,
		drvopen,
		drvclose,
		drvio,
		drvctl,
		drvctmask,
	};



main(int argc, char **argv) 
{
int opt;

	while ((opt=getopt(argc,argv,"a:d:i:m:n:")) != EOF) {
		switch( opt ) {
		case 'a':		/*	Control address */
			esdi_addr = strtol(optarg,NULL,16);
			break;
		case 'd':		/* Disk number	*/
			dma_channel = strtol(optarg,NULL,0);
			break;
		case 'i':		/*	IRQ level */
			intno = strtol(optarg,NULL,0);
			break;
		case 'm':
			MaxIO = strtol(optarg,NULL,0);
			break;
		case 'n':
			if (strlen(optarg) < 1) {
				fputs("error, nil name not allowed\n",stderr);
				return EXIT_FAILURE;
			}
			if (strlen(optarg) > _DNAME_PREFIX) {
				fputs("error, name too long\n",stderr);
				return EXIT_FAILURE;
			}
			strcpy(driver_name,optarg);
			break;
		default:
			return EXIT_FAILURE;
			break;
		}
	}

	if ((FsysPid = qnx_name_locate(getnid(),_FSYS_NAME, 0,0)) == -1) {
		perror("name_locate");
		fputs("Cannot locate filesystem on this node\n",stderr);
		return EXIT_FAILURE;
	}
	if (fsys_mount_driver(FsysPid, max_ndrives, disk, etable,
					MaxIO, MAX_PARALLEL,init_sptr,&version_stamp) == -1) {
		perror("fsys attach");
		fputs("cannot attach to filesystem\n",stderr);
		return EXIT_FAILURE;
	}
	return EXIT_FAILURE;
}


#define	ESDI_OPCODE(x,y) ((x) | ((y) << 5))
static unsigned cmd_buf[4];		/* Place to build commands	*/
static unsigned status_buf[6];		/* Place to read status		*/




int __far drvinit(pid_t proxy, short tabsize, unsigned __far *cachetab)
{
int	rc;

	if (tabsize || cachetab)
		;	
	NotifyPid = proxy;
	
	if ((rc = alloc_dma_buf(&iobuffer, FsysPid, MaxIO * _BLOCK_SIZE)) != 0) {
		return rc;
	}
	outp(esdi_addr+CONTROL,0);	/*	no dma or int's */

	/*
		We explicitly make a far pointer to int_handler because the CS
		at runtime, when called by fsys, will not be the same as the
		constant that will be generated by the compiler at compile time.
	*/

	if ((int_id = qnx_hint_attach(intno, MK_FP(my_cs(),&int_handler),my_ds()))
		 == -1) {
		qnx_segment_free(FP_SEG(iobuffer.bufp));
		return -1;
	}
	return 0;
}


int __far drvterm(unsigned flags, int rsvd1, int rsvd2, int rsvd3)
{
	if (flags || rsvd1 || rsvd2 || rsvd3)
		;
	qnx_hint_detach(int_id);
	qnx_segment_free(FP_SEG(iobuffer.bufp));
	return( 0 );
}


int __far drvopen(int disknum, int flags, int rsvd1, int rsvd2)
{
DISK *dp;
int	status;

	if (rsvd1 || rsvd2)
		;
	if (disknum >= max_ndrives) {
		return -1;
	}

	(dp=disk+disknum)->d_busy = 1;			/* The drive is busy while we open it			*/
	dp->d_type = _HARD;		
	dp->d_sctr_base = 1;	/* controller numbers blocks starting at 1	*/
	strcpy(dp->d_prefix,driver_name);
	
	StatusPtr = dp;
	Function = GET_STATUS;		/*	actually get config. */
	DriveNo = disknum;
	cmd_buf[0] = ESDI_OPCODE(GET_DEV_CONFIG, disknum);
	cmd_buf[1] = 0;
	if (status = send_command(2, disknum)) {
		return -1;
	}
	if ((flags & 01) == 01) 
		wait_interrupt();
	return	dp->d_retval ;
}

/* A closed disk has a zero'd parameter table	*/
int __far drvclose(int disknum,int rsvd1, int rsvd2, int rsvd3)
{
	if (rsvd1 || rsvd2 || rsvd3)
		;
	memset( &disk[ disknum ], 0, sizeof( DISK ) );
	return( 0 );
}
	

/*
 *  Read n blocks (512 bytes) from the disk.
 */

int __far drvio(unsigned disknum, unsigned mode,
                unsigned nblocks, unsigned rsvd,
                struct _block __far *bp)
{
DISK	*dp;

	if (rsvd)
		;
	if (disknum >= max_ndrives || nblocks > MaxIO) {
		return -1;
	}
	(dp = disk+disknum)->d_busy = 1;
	dp->d_retval = -1;
	dp->d_nblks  = 0;

	if (mode == WRITE_BLK) {
		cache_to_buf(iobuffer.bufp,bp,nblocks);
		cmd_buf[0] = ESDI_OPCODE(WRITE_DATA,disknum);
	} else {
		cmd_buf[0] = ESDI_OPCODE(READ_DATA,disknum);
	}
	cmd_buf[1] = nblocks;
	*(long *)(cmd_buf+2) = bp->b_block -1;

	Function = mode;
	NumBlocks = nblocks;
	CachePtr = bp;
	StatusPtr = dp;
	DriveNo = disknum;

	_disable();
	outp(SYS_PORT_A, inp(SYS_PORT_A) | DISK_LIGHT);
	_enable();
	if (send_command(2+sizeof(long)/sizeof(int), disknum)) {
		return -1;
	}
	return 0;
}

/*
 *	Control the disk
 *
 *	Prospective control functions:  ???
 *		format
 *		driver_info - irq, port, statistics, errors detected, etc.
 */
int __far drvctl(unsigned disknum, unsigned request,
				 unsigned rsvd1, unsigned rsvd2,
				 void __far *cdata)
{
	if (disknum || request || rsvd1 || rsvd2 || cdata)
		;
	return -1;
}
int __far drvctmask(unsigned disknum, unsigned request,
				unsigned rsvd1, unsigned rsvd2)
{
	if (disknum || request || rsvd1 || rsvd2)
		;
	return 0;
}


static send_command(unsigned n, unsigned device)
{
register int i;
timeout_t timeout;
	
/* Just in case an old int is there. This is really not needed. */

disp_str("send_command(");
disp_hex(n);	disp_char(',');
disp_hex(device); 
disp_char(')');

disp_str("command = {");
	for (i=0; i < n; i++)
		disp_hex(cmd_buf[i]);
disp_str("} ");

	inp(esdi_addr + INT_STATUS);	
	timeout = 0;
	while(inp(esdi_addr + STATUS) & BUSY)
		if(--timeout == 0) {
disp_str("controller busy! ");
			abort_cmd(device);
			return(-1);
		}

	IntStatus = 0;
	outp(esdi_addr + CONTROL,INT_ENABLE);
	outp( esdi_addr + ATTENTION, ESDI_OPCODE(COMMAND_REQUEST,device));
	for( i = 0; i < n; ++i ) {
		timeout = 0;
		while(inp(esdi_addr + STATUS) & CMD_REG_FULL)
			if(--timeout == 0) {
				abort_cmd(device);
disp_str("send: gave up at "); disp_hex(i); 
				return -1;
			}
		outpw(esdi_addr + INTERFACE_CMD, cmd_buf[i]);
	}
	return(0);
}
    	

static get_status()
{
register unsigned i, n;
timeout_t	  timeout;

	if (inp(esdi_addr + STATUS) & STATUS_OUT) {
		status_buf[0] = inpw(esdi_addr + INTERFACE_STATUS);
		n = status_buf[0] >> 8;
		for(i = 1 ; i < n  &&  i < 6 ; ++i) {
			timeout = 0;
			while((inp(esdi_addr + STATUS) & STATUS_OUT) == 0) {
				if(--timeout == 0) {
disp_str("read_status: timed out at");
disp_hex(i);
					return -1;
				}
			}
			status_buf[i] = inpw(esdi_addr + INTERFACE_STATUS);
		}
		return 0;
	}
disp_str("read_status: status Not Ready!\n");
	return -1;
}


wait_interrupt()
{
	while (IntStatus == 0)
		sleep(1);
}



#ifdef	DEBUGGING
static	char	*attn_error[] = {
				"reserved",
				"Command Complete with Success",
				"reserved",
				"Command Complete with ECC",
				"reserved",
				"Command Complete with Retry",
				"Format Command Partially Complete (Status)",
				"Command Complete with ECC & Retry",
				"Command Complete with Warning",
				"Abort Complete",
				"Reset Complete",
				"Data Transfer Ready - No Status Block",
				"Command Complete with Failure",
				"DMA ERROR",
				"Command Block Error",
				"Attention Error",
		};

static	char	*basic_status[] = {
				"Adaptor Interrupting",
				"Transfer Request",
				"Command Interface Register Full",
				"Status Out Full",
				"Busy",
				"Work In Progress",
				"Interrupt Pending",
				"DMA enabled",
			};


#endif


static pid_t __far	int_handler()
{
int		int_id;
int		wakeup = 0;

	int_id = inp(esdi_addr + INT_STATUS) & 0x1f;
	

disp_str("interrupt(");
disp_hex(int_id);
disp_str(")\n");


	switch (Function) {
	case	GET_STATUS:
		StatusPtr->d_retval = -1;
		status_buf[1] = 0;
		get_status();
		if (int_id == EI_CC_OK) {
			/*	retrieve, and pre-calculate info */
			StatusPtr->d_num_blks = *(long *)(status_buf+2);
			StatusPtr->d_heads = 64;
			StatusPtr->d_sctr_trk = 32;
			StatusPtr->d_sctr_cyl = StatusPtr->d_heads * StatusPtr->d_sctr_trk;
			StatusPtr->d_cyls = 
					StatusPtr->d_num_blks / StatusPtr->d_sctr_cyl;
			StatusPtr->d_precomp = 0;
			StatusPtr->d_retval = 0;
		}
		StatusPtr->d_busy = 0;
		outp(esdi_addr+ATTENTION, ESDI_OPCODE(END_OF_INT,DriveNo));
		/*	someday, change this to set Notify Pid */
		IntStatus = 1;			
		break;

	case	READ_BLK:
	case	WRITE_BLK:
		switch (int_id) {
		case	EI_CC_OK:
			_disable();
			outp(SYS_PORT_A,inp(SYS_PORT_A) & ~ DISK_LIGHT);
			outp(esdi_addr+CONTROL,0);	/*	no dma or int */
			_enable();
			if (Function == READ_BLK)
				buf_to_cache(CachePtr, iobuffer.bufp, NumBlocks);
			wakeup = NotifyPid;
			StatusPtr->d_retval = get_status();
			outp(esdi_addr+ATTENTION,ESDI_OPCODE(END_OF_INT,DriveNo));
			break;

		case	EI_DTR:
			setup_dma(dma_flags | (Function == WRITE_BLK ? DMA_OUTPUT : 0),
				&iobuffer,_BLOCK_SIZE*NumBlocks, dma_channel);
			outp(esdi_addr+CONTROL,INT_ENABLE|DMA_ENABLE);
			break;
		default:
disp_str("\nUnexpected Interrupt :("); 
disp_hex(int_id);
disp_str(") ");
disp_str(attn_error[int_id & 0x0f]);
disp_char('\n');
			get_status();
			wakeup = cleanup_intr() ? NotifyPid : 0;
		}
		break;

	default:
disp_str("Unexpected Function: ");
disp_hex(Function);
disp_char('\n');
		get_status();
		cleanup_intr();
	}
	if(wakeup) {
		Function = 0;
		StatusPtr->d_nblks = (StatusPtr->d_retval ? 0 : NumBlocks);
		StatusPtr->d_busy = 0;
	}
	return wakeup;
}

#ifdef	DEBUGGING	


esdi_failed()
{
int	x;
int	i;
	x = inpw(esdi_addr+INTERFACE_STATUS);
	disp_str("IF_STAT = "); disp_hex(x); disp_char('\n');


	x = inp(esdi_addr+STATUS);
	disp_str("STATUS  = "); disp_hex(x);
	disp_str("( ");
	for (i=0; i < 8; i++) 
		if (x & (1 << i)) {
			disp_str(basic_status[i]);
			disp_str("  ");
		}
	disp_str(")\n");


	x = inp(esdi_addr+INT_STATUS);
	disp_str("INT STAT= "); disp_hex(x);
	disp_str("( "); disp_str(attn_error[x&0x0f]); disp_str(" )\n");

}



#endif

wait_intr(int pdrive)
{
timeout_t	timeout = 0;

	while ((inp(esdi_addr+ STATUS) & INT_REQUEST) == 0) {
		if (--timeout == 0) {
			abort_cmd(pdrive);
			return -1;
		}
	}
	return 0;
}

abort_cmd(int pdrive)
{
timeout_t	timeout = 0;
/*
	attempt recovery! 
	if this guy breaks, there is little we can do -- maybe pray
	or reset the controller then pray?
*/
disp_str("ABORT");
	outp(esdi_addr+ATTENTION,ESDI_OPCODE(ABORT_COMMAND,pdrive));

	while ((inp(esdi_addr+STATUS) & INT_REQUEST) == 0) {
		if (--timeout == 0)
disp_str("abort Timed Out!");
			return -1;
	}

	inp(esdi_addr+INT_STATUS);
	inpw(esdi_addr+INTERFACE_STATUS);	/* discard interface reg */
	outp(esdi_addr + ATTENTION, ESDI_OPCODE(END_OF_INT,pdrive));
	disp_str("Abort Successful?");
#ifdef	DEBUGGING
	esdi_failed();
#endif
	return 0;
}

cleanup_intr()
{
timeout_t timeout = 0;

	while(inp(esdi_addr + STATUS) & BUSY) {
		if(--timeout == 0) {
			return -1;
		}
	}
	outp( esdi_addr + ATTENTION, ESDI_OPCODE(END_OF_INT,DriveNo));
}
