#ifdef __USAGE
%C	[-a address] [-c drivespec] [-d dma channel] [-i interrupt] 
%C	[-m num blocks] [-n name]",
where:",
   -a address      base address of the controller (320)
   -c drivespec    drive spec heads,cylinders,sectors[,drive]
   -d dma channel  dma channel to use (3)
   -i interrupt    controller interrupt (14)
   -m num blocks   maximum number of blocks read/write (1)
   -n driver name  set name of driver (hd)
#endif
/*----------------------------------------------------------------------------
	QNX 4.0 hard disk driver for PS/2 MFM controllers.

	Copyright (c) Quantum Software Systems Ltd. 1989  UNPUBLISHED

History:
	Ported code from 2.15 ps2 driver -- Steve McPolin.

Unimplemented features:
	Driver termination is coded but not tested.

----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <conio.h>
#include <i86.h>
#include <sys/name.h>
#include <sys/irqinfo.h>
#include <sys/kernel.h>
#include <sys/seginfo.h>

#include <diskio.h>
#include "ps2mfm.h"

#ifndef LOCAL
#define	LOCAL	static
#endif

#define	Kludge_fsys_sync() 	Receive(NotifyPid,0,0)



#define DRIVER			FIXED_DISK_PREFIX

LOCAL	char	*fsys_name=_FSYS_NAME;
char	driver_name[_DNAME_PREFIX+1] = DRIVER;
struct	_sysmsg_version_reply	version_stamp = {
		"PS/2 MFM",
		__DATE__,
		400,
		' ',
		0
	};
	
#define	STACK_SIZE		64	/* Size of locally allocated stack			*/



DISK			disk[MAX_DRIVES];	/* Disk parameter tables				*/
BDATA			bdata;		/* Block # decomposed into heads/tracks/sectors	*/

/*-- Memory shared with interrupt handler ----------------------------------*/
/*
 * driver states.
 */

enum {
	ST_IDLE = 0,
	ST_INIT,
	ST_IOSTART,
	ST_IOXFER,
	ST_IOERR
	};

int		dstate = ST_IDLE;
unsigned	Function,		/* Function interrupt handler is doing	*/
			NumBlocks;		/* Number of blocks yet to transfer		*/
DISK		*StatusPtr;		/* Blocks actually transfered so far	*/
BLOCK __far *CachePtr;		/* Linked list of blks read/written		*/
long		FirstBlock;
int			DriveNo;
int		control_word;

pid_t __far int_handler();

int __far drvinit(pid_t proxy, short tabsize, unsigned __far *cachetab);
int __far drvterm(unsigned flags, int rsvd1, int rsvd2, int rsvd3);
int __far drvopen(int disknum, int flags, int rsvd1, int rsvd2);
int __far drvclose(int disknum,int rsvd1, int rsvd2, int rsvd3);
int __far drvio(unsigned __disknum, unsigned mode,
                unsigned num_blocks, unsigned rsvd,
                struct _block __far *bp);
int __far drvctl(unsigned disknum, unsigned request,
				 unsigned rsvd1, unsigned rsvd2,
				 void __far *cdata);
int __far drvctmask(unsigned disknum, unsigned request,
				unsigned rsvd1, unsigned rsvd2);

int (* (etable[_DRIVER_TBL_SIZE]))() = {
		drvinit,
		drvterm,
		drvopen,
		drvclose,
		drvio,
		drvctl,
		drvctmask,
	};






unsigned char	cmd_buf[16];
unsigned char	sense[16];


pid_t		 NotifyPid;			/* Proxy for int handler to use	*/
pid_t		FsysPid;			/* Pid for file system */
int			 int_id;						/* Interrupt handler id			*/

unsigned	 MaxIO				= MAX_IO,
			 ControlInterrupt	= CTL_INTERRUPT;

int	dma_channel = DMA_CHANNEL;
int	dma_flags	= DMA_PS2;
dma_t	iobuffer;



struct	ps2_mfm	ps2;



main(int argc, char **argv) 
{
int		opt; 
char	*p;
int		h, t, n, d;
DISK	*dp;
int		 ControlAddress 	= CTL_ADDRESS;
long	stack[STACK_SIZE+1];

	while ((opt = getopt(argc, argv, "a:c:d:i:m:n:")) != -1) {
		switch( opt ) {
		case 'a':		/*	Control address */
			ControlAddress = strtol(optarg,NULL,16);
			break;
		case 'c':		/*	Cylinders */
			h = strtol(optarg,&p,0);
			if (*p == ',') {
				t = strtol(p+1,&p,0);
				if (*p == ',') {
					n = strtol(p+1,&p,0);
					d = 0;
					if (*p == ',') {
						d = strtol(p+1,NULL,0);
					}
					if (d < MAX_DRIVES) {
						dp = disk + d;
						dp->d_heads = h;
						dp->d_cyls = t;
						dp->d_sctr_trk = n;
						break;
					}
				}
			}
			fputs("Invalid specification\n",stderr);
			return EXIT_FAILURE;
			break;
		case	'd':
			dma_channel = strtol(optarg,NULL,0);
			break;
		case 'i':		/*	IRQ level */
			ControlInterrupt = strtol(optarg,NULL,0);
			break;
		case 'm':
			MaxIO = strtol(optarg,NULL,0);
			if (MaxIO <= 0)
				MaxIO = 1;
			break;
		case 'n':		/*	driver name */
			if (strlen(optarg) < 1) {
				fputs("name too short\n",stderr);
				return EXIT_FAILURE;
			}
			if (strlen(optarg) >= _DNAME_PREFIX) {
				fputs("name too long\n",stderr);
				return EXIT_FAILURE;
			}
			strcpy(driver_name, optarg);
			break;
		default:
			fputs("unknown argument, ignored\n",stderr);
			break;
		}
	}

	ps2.data      = ControlAddress;
	ps2.control   = ControlAddress+2;
	ps2.status    = ControlAddress+2;
	ps2.attention = ControlAddress+4;
	ps2.intstat   = ControlAddress+4;

	if ((FsysPid = qnx_name_locate(getnid(),fsys_name,0,NULL)) == -1) {
		perror(_FSYS_NAME);
		fputs("unable to locate filesystem\n",stderr);
		return EXIT_FAILURE;
	}
	
	if (fsys_mount_driver(FsysPid, MAX_DRIVES, disk, 
			etable, MaxIO, MAX_PARALLEL,
			stack + STACK_SIZE -1, &version_stamp) == -1) {
		perror("fsys mount");
		return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
}

				
int __far drvinit(pid_t proxy, short tabsize, unsigned __far *cachetab)
{
int	rc;

	NotifyPid = proxy;
	if (tabsize || cachetab)
		;
	if ((rc=alloc_dma_buf(&iobuffer, FsysPid, MaxIO * _BLOCK_SIZE)) != 0) {
        return rc;
	}
	/*
		We explicitly make a far pointer to int_handler because the CS
		at runtime, when called by fsys, will not be the same as the
		constant that will be generated by the compiler at compile time.
	*/
	if ((int_id = qnx_hint_attach(ControlInterrupt,MK_FP(my_cs(),&int_handler),my_ds()))
				== -1) {
		qnx_segment_free(FP_SEG(iobuffer.bufp));
		return -1;
	}
	return 0;
}

/*
 *	Terminate the driver - disconnect the interrupt handler.
 */
int __far drvterm(unsigned flags, int rsvd1, int rsvd2, int rsvd3)
{
	if (flags  || rsvd1 || rsvd2 || rsvd3)
		;
	qnx_segment_free(FP_SEG(iobuffer.bufp));
	qnx_hint_detach(int_id);
	return 0;
}

				
int __far drvopen(int disknum, int flags, int rsvd1, int rsvd2)
{
DISK *dp;

	if (rsvd1 || rsvd2)
		;
	if (disknum >= MAX_DRIVES)
		return -1;

	dp= disk + disknum;

	if ( !dp->d_heads && !dp->d_cyls && !dp->d_sctr_trk ) {
		if (get_disk_parms(dp,disknum,1) == -1 ) 
			return -1;
	}

	dp->d_busy = 1;	
	dp->d_type = _HARD;
	dp->d_sctr_base = 1;
	strcpy(dp->d_prefix,driver_name);
	
	dp->d_sctr_cyl = dp->d_heads * dp->d_sctr_trk;
	dp->d_num_blks = (long) dp->d_sctr_cyl * dp->d_cyls;

	Function = GET_STATUS;
	NumBlocks = dp->d_nblks = 0;
	DriveNo  = disknum;
	StatusPtr = dp;

/*	enable interrupts on controller */

	dstate = ST_INIT;
	control_word = FIXED_BITS | INT_ENABLE;
	outp(ps2.control, control_word);

	cmd_buf[0] = RECALIBRATE;
	cmd_buf[1] = cmd_buf[2] = cmd_buf[3] = cmd_buf[4] = cmd_buf[5] = 0;

	if(send_command( COMMAND_CONTROL | (DriveNo << 2), 6 ))
		return(-1);

	if ((flags & 01) == 01)
		Kludge_fsys_sync();	
	return( dp->d_retval );
}

				
int __far drvclose(int disknum,int rsvd1, int rsvd2, int rsvd3)
{
	if (rsvd1 || rsvd2 || rsvd3)
		;
	memset( &disk[ disknum ], 0, sizeof( DISK ) );
	return 0;
}

				
int __far drvio(unsigned disknum, unsigned mode,
                unsigned nblocks, unsigned rsvd,
                struct _block __far *bp)
{
DISK	*dp;

	if (rsvd)
		;
	if (disknum >= MAX_DRIVES)
		return -1;
	
	(dp=disk+disknum)->d_busy   = 1;	/* Busy until int handler clears it	*/
	dp->d_retval = -1;
	dp->d_nblks  = nblocks;

	bdata.pdrive = disknum & 0x01;
	FirstBlock = bdata.block = bp->b_block - 1;	
	/* Adjust for sector base of 0	*/
	/* Initialize variables that the interrupt handler will use	*/

	Function	= mode;
	NumBlocks	= nblocks;
	CachePtr	= bp;
	StatusPtr	= dp;
	DriveNo		= disknum;

	map_thn( &bdata, dp );			/* Compute head / track / sector numbers*/

	if (mode == WRITE_BLK) {
		cache_to_buf(iobuffer.bufp,bp,nblocks);
		cmd_buf[0] = WRITE_DATA | AUTO_SEEK | ECC;
	} else {
		cmd_buf[0] = READ_DATA  | AUTO_SEEK | ECC;
	}
	cmd_buf[1] = (bdata.head << 4) | (bdata.track >> 8) & 0xff;
	cmd_buf[2] = bdata.track & 0xff;
	cmd_buf[3] = bdata.sector;
	cmd_buf[4] = 0x02;					/* 512 byte sector		*/
	cmd_buf[5] = nblocks;
	dstate = ST_IOSTART;
	_disable();
	outp(SYS_PORT_A, inp(SYS_PORT_A) | DISK_LIGHT);
	_enable();
	if (send_command(COMMAND_CONTROL | (DriveNo << 2), 6)) {
		dp->d_busy = 0;
		return -1;
	}
	return 0;
}




pid_t __far int_handler(void) 
{
int		int_status;
int		status;
pid_t	rval = 0;
int		i;
long	last_block;

#ifdef DEBUGGING
disp_char('!');
disp_char('0'+dstate);
#endif

	status = inp(ps2.status);
	int_status = inp(ps2.intstat);

#ifdef DEBUGGING
disp_hex(status);
disp_hex(int_status);
#endif

	switch (dstate) {
	case	ST_INIT:
		_disable();
		outp(SYS_PORT_A, inp(SYS_PORT_A) & ~DISK_LIGHT);
		_enable();
		StatusPtr->d_retval = (int_status == 0) ? 0 : -1;
		StatusPtr->d_busy = 0;
		dstate = ST_IDLE;
		rval = NotifyPid;
		break;
	case	ST_IOSTART:
		if (int_status == 0) {
			setup_dma(dma_flags | (Function == WRITE_BLK ? DMA_OUTPUT : 0),
				&iobuffer,NumBlocks * _BLOCK_SIZE,dma_channel);
			outp(ps2.control,control_word |= DMA_ENABLE);
			outp(ps2.attention, DATA_REQUEST | (DriveNo << 2));
			dstate = ST_IOXFER;
			break;
		} 
		StatusPtr->d_retval = -1;
		StatusPtr->d_busy = 0;
		dstate = ST_IDLE;
		rval = NotifyPid;
		break;
	case	ST_IOXFER:
		outp(ps2.control,control_word &= ~DMA_ENABLE);
		if (int_status == 0) {
			dstate = ST_IDLE;
			if (Function == READ_BLK) {
				buf_to_cache(CachePtr,iobuffer.bufp, NumBlocks);
			}
			StatusPtr->d_retval = 0;
			StatusPtr->d_nblks = NumBlocks;
			StatusPtr->d_busy = 0;
			_disable();
			outp(SYS_PORT_A, inp(SYS_PORT_A) & ~DISK_LIGHT);
			_enable();
			rval = NotifyPid;
			break;
		} 
		outp(ps2.attention, SENSE_SUMMARY | (DriveNo << 2));
/*
	I'm not sure why the controller won't interrupt us,....
*/
		if (!wait_status(DATA_READY,0)) {
#ifdef DEBUGGING
disp_str("not ready!");
#endif
			break;
		}

		for (i=0; i < 14; i += 2) {
			*(unsigned *)(sense+i) = inpw(ps2.data);
#ifdef DEBUGGING
disp_hex(sense[i]);
disp_hex(sense[i+1]);
#endif
		}
		last_block = StatusPtr->d_sctr_trk * (sense[4]&0x0f)
	               + StatusPtr->d_sctr_cyl * (sense[3]+((sense[4]<<3)&0x300))
				   + sense[5] - StatusPtr->d_sctr_base;
#ifdef DEBUGGING
disp_str("last_block =");
disp_lhex(last_block);
disp_str("first =");
disp_lhex(FirstBlock);
#endif
		if (Function == READ_BLK) {
			buf_to_cache(CachePtr, iobuffer.bufp, last_block - FirstBlock);
		}
		StatusPtr->d_retval = (StatusPtr->d_nblks = last_block - FirstBlock) 
								? 0 : -1;
		StatusPtr->d_busy = 0;
		rval = NotifyPid;
		dstate = ST_IDLE;
		break;
	default:
		break;
	}
	return rval;
}

int __far drvctl(unsigned disknum, unsigned request,
				 unsigned rsvd1, unsigned rsvd2,
				 void __far *cdata)
{
	if (disknum || request || rsvd1 || rsvd2 || cdata)
		;
	return -1;
}

int __far drvctmask(unsigned disknum, unsigned request,
				unsigned rsvd1, unsigned rsvd2)
{
	if (disknum || request || rsvd1 || rsvd2)
		;
	return 0;
}

wait_status(int mask, int logic)
{
int	timeout = 0;

	if (logic) {
		while (inp(ps2.status) & mask) {		
			if (--timeout == 0)
				return 0;
		}
	} else {
		while (!(inp(ps2.status) & mask)) {
			if (--timeout == 0)
				return 0;
		}
	}
	return 1;
}

send_command( cmd, n )
unsigned cmd, n;
{
int i;

	/*	wait for status.... */
	if (!wait_status(BUSY,1))
		return -1;

	/*	tell it the command. */
	outp(ps2.attention, cmd );
	for( i = 0; i < n; i+=2) {
		if (!wait_status(DATA_READY,0))
			return -1;
		outpw(ps2.data, *(unsigned *)(cmd_buf+i));
	}
	return(0);
}
