
/*
	Manifests, etc. for ps2 mfm drive.
*/

#define MAX_DRIVES		2		/* 2 physical drives				*/
#define DMA_CHANNEL		3		/* DMA channel						*/
#define	CTL_ADDRESS		0x320	/* Controller base address			*/
#define	CTL_INTERRUPT	14		/* Controller interrupt				*/
#define	MAX_IO			34		/*	default io max block	*/
#define	MAX_PARALLEL	1



struct	ps2_mfm {
		int	data,  			/*	+0,rw	*/
			control,		/*	+2,w	*/
			status,			/*	+2,r	*/
			attention,		/*	+4,w	*/
			intstat;		/*	+4,r	*/
		};


/*-- Control register bits --*/
#define DMA_ENABLE		0x01	/* Enable DMA operations					*/
#define INT_ENABLE		0x02	/* Enable INT's to the bus					*/
#define	NEVER			0x18	/* must be zero */
#define	USE_16BIT       0x24    /* 16 bits */
#define	ALLWAYS         0x40    /* allways set except during reset          */
#define	RESET			0x80	/* Reset the adapter */

#define FIXED_BITS		0x64	/* 16 bit mode with 1 reserved bit set 1	*/

/*-- Commands for Attention Register --*/

#define	COMMAND_CONTROL	0x80	/* Indicate upcoming command				*/
#define	COMMAND_SPECIFY	0x40	/* Indicate upcoming drive specification	*/
#define	SENSE_SUMMARY	0x20	/* Request sense summary block				*/
#define	DATA_REQUEST	0x10	/* Request controller to read or write data	*/
#define	ABORT_COMMAND	0x01	/* Abort current command					*/


/*-- Status bits found in status register --*/
#define USE_16_BITS		0x20
#define DATA_READY		0x10	/* Data available to be read				*/
#define DIRECTION		0x08	/* Set if controller has data for us		*/
#define BUSY			0x04
#define	INT_REQUEST		0x02
#define	XFER_ENABLE		0x01	/* Set if block of data going in / out		*/

/*-- Controller Command Codes --*/
#define READ_DATA		0x10	/* Read sector in DMA mode					*/
#define RECALIBRATE		0x80	/* Restore with fast steps					*/
#define	WRITE_DATA		0x90	/* Write sector in DMA mode					*/
#define SEEK			0xe0	/* Seek with fast steps						*/
#define FORMAT_TRACK	0xf0	/* Format track in DMA mode 				*/

/*-- Bottom 4 bits of Command Codes --*/
#define	NO_DATA			0x08	/* Don't transfer data from controller		*/
#define	AUTO_SEEK		0x04	/* Automatic seek with command				*/
#define	SKIP_CONTROL	0x02	/* Skip control data						*/
#define	ECC				0x01	/* Process with ECC, not CRC				*/

/*-- IO port and bit for flashing the hard disk light --*/
#define SYS_PORT_A	0x0092
#define DISK_LIGHT	0x40


