#ifdef __USAGE
%C	[-n devname] [-i interrupt] [-a controller address]
%C	[-m num blocks] [-n device name]
where:
   -a address       controller address (3540)
   -i interrupt     interrupt to attach (14)
   -m num blocks    number of blocks per io (34)
   -n device name   name of device (hd)
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <conio.h>
#include <i86.h>
#include <sys/name.h>
#include <sys/irqinfo.h>
#include <sys/kernel.h>
#include <sys/seginfo.h>
#include <diskio.h>

#include "ps2scsi.h"

#ifndef DEBUGGING
#define	DEBUGGING 0
#endif

#if DEBUGGING
#ifdef LOCAL
#undef LOCAL
#endif
#define	LOCAL
#endif

#ifndef LOCAL
#define	LOCAL static
#endif

struct	_sysmsg_version_reply	version_stamp = {
		"PS/2 SCSI",
		__DATE__,
		400,
		' ',
		0
	};

char driver_name[_DNAME_PREFIX] = FIXED_DISK_PREFIX;
LOCAL	char	*fsys_name=_FSYS_NAME;

pid_t __far int_handler();

int __far drvinit(pid_t proxy, short tabsize, unsigned __far *cachetab);
int __far drvterm(unsigned flags, int rsvd1, int rsvd2, int rsvd3);
int __far drvopen(int disknum, int flags, int rsvd1, int rsvd2);
int __far drvclose(int disknum,int rsvd1, int rsvd2, int rsvd3);
int __far drvio(unsigned disknum, unsigned mode,
                unsigned nblocks, unsigned rsvd,
                struct _block __far *bp);
int __far drvctl(unsigned disknum, unsigned request,
				 unsigned rsvd1, unsigned rsvd2,
				 void __far *cdata);
int __far drvctmask(unsigned disknum, unsigned request,
				unsigned rsvd1, unsigned rsvd2);

void (* (etable[_DRIVER_TBL_SIZE]))() = {
		drvinit,
		drvterm,
		drvopen,
		drvclose,
		drvio,
		drvctl,
		drvctmask,
	};


volatile	int	IoStatus;
volatile	int	DriverInit;
int			DriverState;
static	volatile	IntStatus;		/* bumped when interrupt transaction done */
static	unsigned	Function;		/* Function interrupt handler is doing	*/
static	unsigned	NumBlocks;		/* Number of blocks yet to transfer		*/
static	unsigned	DriveNo = 0;	/* Disk number */
static	unsigned	Istate = 0;		/* Interrupt Handler state (2-phase)	*/
static	DISK		*StatusPtr;		/* Blocks actually transfered so far	*/
static	BLOCK __far *CachePtr;		/* Linked list of blks read/written		*/
int			int_id;

/*
	These areas are for the binding between the filesystem and the driver,
	they are 'relative' constants, that is they can only be redefined at
	startup.

*/

static	pid_t		NotifyPid;			/* Proxy for int handler to use	*/
static	pid_t		FsysPid;			
static	unsigned	MaxIO				= MAX_IO,
					ControlAddress 		= CTL_ADDRESS,
					ControlInterrupt	= CTL_INTERRUPT;

/*
	This buffer is defined when the driver is initialed.
*/
static	struct dma_buffer	iobuffer;	
static DISK	disk[MAX_NDRIVES];	/* Disk parameter tables	*/


main(int argc, char **argv) 
{
int	 opt; 
long	stack[STACK_SIZE];

	while ((opt=getopt(argc,argv,"a:i:m:n:")) != EOF) {
		switch( opt ) {
		case 'a':		/*	Control address */
			ControlAddress = strtol(optarg,NULL,16);
			break;
		case 'i':		/*	IRQ level */
			ControlInterrupt = atoi( optarg );
			break;
		case 'm':
			MaxIO = strtol(optarg,NULL,0);
			break;
		case 'n':
			if (strlen(optarg) < 1) {
				return EXIT_FAILURE;
			}
			if (strlen(optarg) > _DRIVER_NAME_LEN) {
				fputs("error, name too long\n",stderr);
				return EXIT_FAILURE;
			}
			strcpy(driver_name,optarg);
			break;
		default:
			fputs("Unknown option ignored\n",stderr);
		}
	}

	if ((FsysPid = qnx_name_locate(getnid(),fsys_name,0,0)) == -1) {
		perror("name_locate");
		fputs("unable to locate filesystem\n",stderr);
	}
	get_cpu_speed();

	if (fsys_mount_driver(FsysPid, MAX_NDRIVES, disk, etable,
					MaxIO, MAX_PARALLEL,stack+STACK_SIZE-1,&version_stamp) == -1) {
		perror("mount");
		fputs("unable to attach to filesystem\n",stderr);
		return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
}




/*	
	the ps2 scsi spec says buffers must be on dword boundaries.
	Since doing this would cause all structures defined in include files
	to behave unpredictably, we allocate one array which guarantee's 
	enough space to hold all of the structures, on dword boundaries.
*/


/*	scsi control block */
struct _scb *scb_ptr;

/*	scsi unit capacity */
struct	_capacity	 *cap_ptr;

/*	transaction status buffer */
char *tsb_ptr;

/*	physical addresses of above */
paddr_t	scb_phys_addr,
		cap_phys_addr,
		tsb_phys_addr;

#define	MEMPOOL_SIZE	(sizeof(struct _scb) + \
						 sizeof(struct _capacity)+\
						 sizeof(short)*13 + \
						 3 * sizeof(long))

unsigned char mempool[MEMPOOL_SIZE];

#define	ALIGN(x,n)	(((unsigned)(x)+(n-1)) & ~(n-1))

static init_vars()
{
unsigned	char	*p;
paddr_t		base_addr;


	/*	force alignment */  
	base_addr = phys_addr(MK_FP(my_ds(),(unsigned)mempool));
	p = (unsigned char *)ALIGN(mempool,sizeof(long));


    scb_ptr = (struct _scb *)p;
	scb_phys_addr = base_addr + (p - mempool);

	p += sizeof(struct _scb);
	p = (unsigned char *)ALIGN(p,sizeof(long));

	cap_ptr = (struct _capacity *)p;
	cap_phys_addr = base_addr + (p - mempool);

	p += sizeof(struct _capacity);
	p = (unsigned char *)ALIGN(p, sizeof(long));

	tsb_ptr =  p;
	tsb_phys_addr = base_addr + (p - mempool);

}



void swabber(const void __far *src, void __far *dst, int nbytes)
{
const unsigned char __far *s1 = src;
unsigned char __far *s2 = dst;

	s1 += nbytes;
	while (nbytes--)
		*s2++ = *(--s1);
}
		
#define	swab_long(src,dst)	swabber((src),(dst),sizeof(long))
							

int __far drvinit(pid_t proxy, short tabsize, unsigned __far *cachetab)
{
int	rc;
	if (tabsize || cachetab)
		;
	NotifyPid = proxy;
	init_vars();
	if ((rc = alloc_dma_buf(&iobuffer, FsysPid, MaxIO * _BLOCK_SIZE)) == -1) {
		return -1;
    }

	/*	set state to zero so it doesn't wake anybody else up....
	*/

	DriverState = 0;	

	if ((int_id = qnx_hint_attach(ControlInterrupt,
				MK_FP( my_cs(), &int_handler ),	my_ds())) == -1) {
		qnx_segment_free(FP_SEG(iobuffer.bufp));
		return -1;
	}		
	io_out(ControlAddress + CONTROL, HARDWARE_RESET);
	delay(50);
	io_out(ControlAddress + CONTROL, 0);
	io_out(ControlAddress + CONTROL, INT_ENABLE);
	while (!DriverInit)
		Yield();
	io_out(ControlAddress + CONTROL, INT_ENABLE|DMA_ENABLE);
	return 0;
}

int __far drvterm(unsigned flags, int rsvd1, int rsvd2, int rsvd3)
{
	if (flags || rsvd1 || rsvd2 || rsvd3)
		;
	qnx_segment_free(FP_SEG(iobuffer.bufp));
	qnx_hint_detach(int_id);
	return 0;
}

int __far drvopen(int disknum, int flags, int rsvd1, int rsvd2)
{
DISK *dp;

register struct _scb *scb = scb_ptr;

	if (rsvd1 || rsvd2)
		;

	if (disknum >= MAX_NDRIVES) {
		return -1;
	}

	(dp=disk+disknum)->d_busy = 1;		/* The drive is busy while we open it			*/
	dp->d_sctr_base = 1;	/* controller numbers blocks starting at 1	*/

	/*
	 * We ask the controller for the size of the disk.
	 */
	scb->command= READ_CAPACITY;
	scb->enable	= 0xe200;
	scb->laddr	= 0;
	scb->buffer	= cap_phys_addr;
	scb->nbytes	= sizeof(struct _capacity);
	scb->tsb	= tsb_phys_addr;
	scb->chain	= 0;
	scb->block_cnt	= 0;
	scb->block_len	= 0;

	DriverState = 0;		/*	still 0 -- not real commands yet */
	DriverInit  = 0;
	if (send_command(START_SCB | disknum, scb_phys_addr) == 0) { 	
		while (DriverInit == 0)
			Yield();
		if (((IntStatus >> 4) & 0x0f) == 1) {
			swab_long(&(cap_ptr->nblocks), 
					 &(dp->d_num_blks));
			dp->d_type = _HARD;
			dp->d_cyls = dp->d_num_blks / (32 * 64);	
			dp->d_sctr_cyl = 32*64;
			dp->d_heads = 64;
			dp->d_sctr_trk = 32;
			dp->d_retval = 0;
			strcpy(dp->d_prefix,driver_name);
		} else {
			dp->d_type = 0;
			dp->d_retval = -1;
		}
	} else {
		dp->d_retval = -1;
		dp->d_type = 0;
	}
	dp->d_busy = 0;
	DriverState = 1;		/*	now we are for real! */
	if (!(flags & 01))
		Trigger(NotifyPid);
	return	dp->d_retval ;
}


int __far drvclose(int disknum,int rsvd1, int rsvd2, int rsvd3)
{
	if (rsvd1 || rsvd2 || rsvd3)
		;
	memset(disk+disknum,0,sizeof(*disk));
	return( 0 );
}
	


int __far drvio(unsigned disknum, unsigned mode,
                unsigned nblocks, unsigned rsvd,
                struct _block __far *bp)
{
DISK	*dp;
register struct _scb *scb = scb_ptr;

	if (rsvd)
		;
	if (disknum >= MAX_NDRIVES) {
		return -1;
	}
	if (nblocks > MaxIO) {
		return -1;
	}
	(dp = disk+disknum)->d_busy = 1;
	dp->d_retval = -1;
	dp->d_nblks  = 0;

	if (mode == WRITE_BLK) {
		cache_to_buf(iobuffer.bufp,bp,nblocks);
	}
	Function = mode;
	NumBlocks = nblocks;
	CachePtr = bp;
	StatusPtr = dp;
	Istate = 0;
	DriveNo = disknum;

	scb->command= (mode == WRITE_BLK) ? WRITE_DATA : READ_DATA;
	scb->enable	= 0xe000;
	scb->laddr	= bp->b_block-1;
	scb->buffer	= iobuffer.phys_addr;
	scb->nbytes = _BLOCK_SIZE * NumBlocks;
	scb->tsb	= tsb_phys_addr;
	scb->chain	= 0;
	scb->block_cnt	= NumBlocks;
	scb->block_len	= _BLOCK_SIZE;

	_disable();
	io_out(SYS_PORT_A, io_in(SYS_PORT_A) | DISK_LIGHT);
	_enable();
	if ((send_command( START_SCB | disknum, scb_phys_addr )) == 0) 
		return 0;

	ll_debug();
	return -1;
}


int __far drvctl(unsigned disknum, unsigned request,
				 unsigned rsvd1, unsigned rsvd2,
				 void __far *cdata)
{
	if (disknum || request || rsvd1 || rsvd2 || cdata)
		;
	return -1;
}
int __far drvctmask(unsigned disknum, unsigned request,
				unsigned rsvd1, unsigned rsvd2)
{
	if (disknum || request || rsvd1 || rsvd2)
		;
	return 0;
}



pid_t	__far	int_handler(void)
{
int	bstat;

	IntStatus = io_in(ControlAddress+INT_STATUS);
	while ((bstat = io_in(ControlAddress+STATUS)) &  BUSY)
		;

	IoStatus = bstat;
	io_out(ControlAddress+ATTENTION, (IntStatus & 0x0f) | END_OF_INT);

	if (DriverState == 0) {
		DriverInit++;
		return (pid_t)0;
	}
	_disable();
	io_out(SYS_PORT_A, io_in(SYS_PORT_A) & ~DISK_LIGHT);
	_enable();

	
	
	if ((IntStatus & 0xf0) == 0x10) {
		StatusPtr->d_retval = 0;
		StatusPtr->d_nblks = NumBlocks;
		if (Function == READ_BLK) {
			buf_to_cache(CachePtr, iobuffer.bufp, NumBlocks);
		}
	} else {
		StatusPtr->d_retval = -1;
	}
	StatusPtr->d_busy = 0;
	return NotifyPid;
}
	


send_command(const unsigned code, const long command)
{
int timeout;
int	psw;	

	timeout = 0;

	psw = disable();
	while(io_in(ControlAddress + STATUS) & BUSY) {
		restore(psw);
		if(--timeout == 0) {
			io_out(ControlAddress + CONTROL, 0);
			ll_debug();
			return -1;
		}
		psw = disable();
	}

	io_out( ControlAddress + INTERFACE_CMD1, command );
	io_out( ControlAddress + INTERFACE_CMD2, command >> 8);
	io_out( ControlAddress + INTERFACE_CMD3, command >> 16);
	io_out( ControlAddress + INTERFACE_CMD4, command >> 24);

	io_out( ControlAddress + ATTENTION, code);

	restore(psw);

	return 0;
}
