#ifdef __USAGE

%C	[-a controller] [-c drivespec] [-i interrupt] [-m maxsize]
    [-n name] [-8]
Options:
 -a controller   The controller address (default is 1f0 hex).
 -c drivespec    Specify drive parameters explicitly.
 -i interrupt    The interrupt level used by the controller (default is 14).
 -m maxsize      The maximum number of blocks that can be transferred from
                 the disk to memory in one operation (default is 34).
 -n name         The name of the block special device associated with the
                 hard drive. The default is hd e.g. /dev/hd0
 -8              Enable head select 3, even if drive has less than 7 heads.
 -7              Do not enable head select 3, under any condition.
 -s              Do not send setup information to drive.
 -r              Reset drive before starting.
 
Where:
 drivespec       is in the format heads,cylinders,sectors[,disknum]
                 If disknum is omitted, 0 is assumed.
#endif

/*----------------------------------------------------------------------------
	QNX 4.0 hard disk driver for Western Digital 1010 compatible controllers.

	Copyright (c) Quantum Software Systems Ltd. 1991  UNPUBLISHED

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <i86.h>
#include <sys/name.h>
#include <sys/kernel.h>
#include <sys/irqinfo.h>
#include <malloc.h>
#include "diskio.h"
#include "wd.h"

#ifndef DEBUGGING
#define	DEBUGGING 0
#endif

#if DEBUGGING
#include <sys/osinfo.h>
struct _osinfo _osdata;
#endif


#define	LOCAL

LOCAL	char	drv_prefix[_DNAME_PREFIX] = FIXED_DISK_PREFIX;
LOCAL	char	*fsys_name=_FSYS_NAME;
LOCAL	struct	_sysmsg_version_reply	version_stamp = {
		"WD 1003",
		__DATE__,
		400,
		' ',
		0
	};

/*
 *	Driver functions, a jump table through which fsys enters the driver code
 *
 *	Array of pointers to driver functions.
 */

/*
 *  Driver entry points.
 */

pid_t __far int_handler();

int __far drvinit(pid_t driver_proxy, short tabsize, unsigned __far *cachetab);
int __far drvterm(unsigned flags, int rsvd1, int rsvd2, int rsvd3);
int __far drvopen(int disknum, int flags, int rsvd1, int rsvd2);
int __far drvclose(int disknum,int rsvd1, int rsvd2, int rsvd3);
int __far drvio(unsigned __disknum, unsigned mode,
                unsigned num_blocks, unsigned rsvd,
                struct _block __far *bp);
int __far drvctl(unsigned disknum, unsigned request,
				 unsigned rsvd1, unsigned rsvd2,
				 void __far *cdata);
int __far drvctmask(unsigned disknum, unsigned request,
				unsigned rsvd1, unsigned rsvd2);

int	(* (etable[_DRIVER_TBL_SIZE]))() = {
		drvinit,
		drvterm,
		drvopen,
		drvclose,
		drvio,
		drvctl,
		drvctmask,
	};


 /*
 * int_id is the interrupt cooke
 */
LOCAL	DISK	disk[MAX_DRIVES];	/* Disk parameter tables	*/
LOCAL	pid_t	fsys_proxy;		/*	proxy to wake up filesystem */
LOCAL	int	int_id;			/*	interrupt cookie */


/* This info is used on a per-request basis.
 *
 * bdata is the first block address decomposed to head/track/sector.
 * first_block is the head of the cache list
 * cur_block is the current cache block
 * cur_disk is the physical disk number.
 * io_size is the number of blocks requested.
 * io_current is the relative block no currently used.
 * ecc_blockno is used if error correction is being ignored to
 *				flag the last valid block.
 */
LOCAL	BDATA	bdata;		
LOCAL	BLOCK __far *first_block;
LOCAL	BLOCK __far *cur_block;	
LOCAL	int	cur_disk;
LOCAL	int	io_size;
LOCAL	int	io_current;	
LOCAL	int	ecc_blockno;
/*	valid states */
enum {
	IDLE = 0,
	WAIT_STATUS,
	WAIT_READ,
	WAIT_WRITE,
	WAIT_ECC_ERR,
};
LOCAL	volatile int	state = IDLE;

#if DEBUGGING > 2
LOCAL	SET_STATE(int nstate)
{
	disp_char('!');
	disp_char(state+'0');
	disp_char('/');
	disp_char(nstate+'0');
	state = nstate;
}
#else
#define SET_STATE(nstate)	((state) = (nstate))
#endif



LOCAL	struct	{
		short	ecc_ignore:1,
				bit8:1,
				nobit8:1,
				nosetup:1,
				reset:1;
		} exflags;


struct	wd1010 {
	short int	data;
	short int	error,		/* read only */
				precomp;	/* write only */
	short int	sect_count;
	short int	sect_no;
	short int	cyl_low;
	short int	cyl_high;
	short int	drv_head;
	short int	status,	/* read only */
				command; /* write only */
	short int	dio_port; /* funny thing */
} wd;

LOCAL	set_wdaddr(int base_addr)
{
	wd.data = base_addr;
	wd.error = wd.precomp = base_addr+1;
	wd.sect_count = base_addr + 2;
	wd.sect_no = base_addr + 3;
	wd.cyl_low = base_addr + 4;
	wd.cyl_high = base_addr + 5;
	wd.drv_head = base_addr + 6;
	wd.status = wd.command = base_addr + 7;
	if (base_addr == PRIMARY_ADDR)
		wd.dio_port = FDC_PRIMARY;
	else if (base_addr == SECONDARY_ADDR)
		wd.dio_port = FDC_SECONDARY;
	else
		wd.dio_port = 0;
}

LOCAL	setup_drive(int dno,int isset)
{

	if (!isset) {
		if (get_disk_parms(disk+dno,dno,0) == -1) 
			return -1;
	}
	disk[dno].d_type = _HARD;
	disk[dno].d_sctr_base = 1;
	strcpy(disk[dno].d_prefix,drv_prefix);
	/* Precompute some handy numbers	*/
	disk[dno].d_sctr_cyl = disk[dno].d_heads * disk[dno].d_sctr_trk;
	disk[dno].d_num_blks = (long) disk[dno].d_sctr_cyl * disk[dno].d_cyls;

}

LOCAL	short	io_max	= MAX_IO;
LOCAL	short	intno	= CTL_INTERRUPT;



#if DEBUGGING
int	debug_sel;
#endif



main(int argc, char **argv) 
{
int	c;
pid_t	fsys_pid;
char	*_pname = basename(argv[0]);
char	*p;
int		heads, cyls, sects, disknum;
int	isset[MAX_DRIVES];
long	stack[STACK_SIZE];

	memset(isset,0,sizeof(isset));
	set_wdaddr(PRIMARY_ADDR);	/*	default */
	get_cpu_speed();	
	
	while ((c=getopt(argc, argv, "78sra:c:i:m:n:")) != EOF) {
		switch (c) {
		case '8':		/* 8 bit head numbers for some core drives	*/
			exflags.bit8 = 1;
			break;
		case '7':
			exflags.nobit8 = 1;
			break;
		case 's':
			exflags.nosetup = 1;
			break;
		case 'r':
			exflags.reset = 1;
			break;
		case 'a':		/*	Control address */
			set_wdaddr((int) strtol(optarg,NULL,16));
			break;
        case 'c':
			heads = strtol(optarg,&p,10);			
			if (p == NULL || *p != ',') {
				fprintf(stderr,"heads must be specified\n");
				return EXIT_FAILURE;
			}
			cyls = strtol(p+1, &p, 10);
			if (p == NULL || *p != ',') {
				fprintf(stderr,"cylinders must be specified\n");
				return EXIT_FAILURE;
			}
			sects = strtol(p+1, &p, 10);
			if (p != NULL && *p == ',') {
				disknum = strtol(p+1, &p, 10);
			} else {
				disknum = 0;
			}
			isset[disknum] = 1;
			disk[disknum].d_cyls = cyls;
			disk[disknum].d_heads = heads;
			disk[disknum].d_sctr_trk = sects;
			break;
		case 'i':		/*	IRQ level */
			intno = strtol(optarg,NULL,10);
			break;
		case 'n':
			strcpy(drv_prefix,optarg);
			break;
		case 'm':		/*	Max_io per read/write */
			io_max = min(255, strtol(optarg,NULL,10));
			break;
		default:
			fprintf(stderr,"unknown option %c, ignored\n",c);
		}
	}


#if	DEBUGGING
	qnx_osinfo(0, &_osdata);
	debug_sel = (_osdata.primary_monitor & 001) ? MONO_DISP : COLOUR_DISP;
	set_display(MK_FP(debug_sel,80*2*5),18,75); /* start 5 lines down */
#endif	

	for (c=0; c < MAX_DISK; c++) {
		setup_drive(c,isset[c]);
	}

	if ((fsys_pid = qnx_name_locate(getnid(), fsys_name,0,NULL)) == -1) {
		fprintf(stderr,"%s: name locate (%s) failed, %s\n", _pname, fsys_name,
						strerror(errno));
		return EXIT_FAILURE;
	}
	if (fsys_mount_driver(fsys_pid, MAX_DISK, 
			disk, etable, io_max, MAX_PARALLEL,
			stack + STACK_SIZE -1, &version_stamp) == -1) {
		perror(_pname);
		return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
}


int __far drvinit(pid_t proxy,unsigned tabsize, unsigned __far *cachetab) 
{
	if (tabsize && cachetab) 
		;
	fsys_proxy = proxy;
	int_id = qnx_hint_attach(intno,MK_FP(my_cs(),&int_handler),my_ds());
#if DEBUGGING
	disp_str(int_id == -1 ?
		"unable to attach interrupt handler\n" :
		"interrupt handler attached\n");
#endif
	return int_id == -1 ? -1 : 0;
}

int __far drvterm(unsigned flags,unsigned rsvd1, unsigned rsvd2,unsigned rsvd3)
{
	if (flags && rsvd1 && rsvd2 && rsvd3)
		;
	qnx_hint_detach(int_id);
	return 0;
}

#if DEBUGGING
long	int_sequence = 0;
long	proxy_sequence = 0;
long	io_sequence = 0;
#endif

int __far drvopen(int disknum, unsigned flags, unsigned rsvd1, unsigned rsvd2) 
{
DISK *dp = &disk[ disknum ];

	if (rsvd1 && rsvd2)
		;
		
	if (dp->d_type == 0) {	
		if ((flags & _OPEN_SYNC) == 0) {
			Trigger(fsys_proxy);
		}
#if DEBUGGING
	disp_str("drv_open: no drive ");	disp_hex(disknum);
	disp_str("\n");
#endif
		return -1;
	} 
	
	if (exflags.nosetup == 0) {
		dp->d_busy = 1;
		if (send_status() == 0 && (flags & _OPEN_SYNC)) {
			Receive(fsys_proxy,0,0);
		}
	}
	if (exflags.nobit8 == 0 && wd.dio_port) {
		if (dp->d_heads > 7)
			exflags.bit8 = 1;
		if (exflags.reset) {
			io_out(wd.dio_port, 0x04);	/* reset disk */
			usec_delay(8);				/* at least 8 usecs */	
			io_out(wd.dio_port, exflags.bit8 ? 0x08 : 0x0);
			exflags.reset = 0;		/* only do this once */
		} else {
			io_out(wd.dio_port, exflags.bit8 ? 0x08 : 0x00);
		}
	}
#if DEBUGGING
	int_sequence = proxy_sequence = io_sequence = 0;
	disp_str("drv_open: returned ");	disp_hex(disknum);
	disp_hex(dp->d_retval);
	disp_str("\n");
#endif
	return dp->d_retval;
}

int __far drvclose(int disknum,unsigned rsvd1,unsigned rsvd2,unsigned rsvd3) 
{
	if (rsvd1 && rsvd2 && rsvd3) 
		;
	memset(disk+disknum, 0, sizeof(DISK));
	return 0;
}


int __far drvio(unsigned disknum, unsigned mode, 
				unsigned nblocks, unsigned rsvd,
				BLOCK __far *bp)
{
DISK	*dp = &disk[ disknum ];
int	rval;
int	timeout=0;


	if (rsvd && nblocks)
		;
	disp_char(mode ? 'w' : 'r');

	dp->d_busy   = 1;				/* Busy until int handler clears it		*/
#if DEBUGGING	
	io_sequence++;
#endif
	bdata.pdrive = disknum & 0x01;
	bdata.block = bp->b_block - 1;			/* Adjust for sector base of 0	*/
	map_thn(&bdata, dp);			/* Compute head / track / sector numbers*/

	io_size = nblocks;
	io_current = 0;
	first_block = cur_block = bp;
	cur_disk = disknum;
	dp->d_retval = 0;
	while (--timeout && (rval=io_in(wd.status)) & (BUSY|CIP))
		;
	if (rval & (BUSY|CIP)) {
		dp->d_nblks = 0;
		dp->d_retval = _DRVR_DISK_FAILURE;
		dp->d_busy = 0;
#if DEBUGGING
		proxy_sequence++;
#endif
		Trigger(fsys_proxy);
	} else if (mode == READ_BLK) {
		io_setup(READ_SECTOR,WAIT_READ);
	} else {
		io_setup(WRITE_SECTOR,WAIT_WRITE);
		if (int_handler()) {
#if DEBUGGING
			proxy_sequence++;
#endif
			Trigger(fsys_proxy);
		}
	}
	return dp->d_retval;
}

int __far drvctl(unsigned disknum, unsigned request,
		 		  unsigned rsvd1, unsigned rsvd2, void __far *cdata)
{
int	retcode = -1;

	if (disknum && rsvd1 && rsvd2)
		;
		
	if (request == QCTL_ELEVEL) {
		exflags.ecc_ignore = *(int __far *)cdata != 0;
		*(int __far *)cdata = 0;
		retcode = 0;
	}
	return retcode;
}

int __far drvctmask(unsigned disknum, unsigned request,
				unsigned rsvd1, unsigned rsvd2)
{
	if (disknum || rsvd1 || rsvd2)
		;
	return (request == QCTL_ELEVEL) ? 
		_DRVR_CTRL_SUPPORTED | _DRVR_CTRL_PRIVILIGED 
		:0;		
}

pid_t __far int_handler()
{
int		res=0;
int		wd_status;
int		wd_error;
int		timeout=0;
int		lblock;

	while (--timeout && (wd_status=io_in(wd.status)) &BUSY) 
		;
	if (wd_status & BUSY) {	// controller is DEAD
		ll_debug();
	}
	wd_error = io_in(wd.error);
	
	switch (state) {
	case	IDLE:
		res = 0;
		break;
	case	WAIT_STATUS:
		lblock = (wd_status & ERROR) == 0;
		res = 1;
		break;
	case	WAIT_READ:
#if DEBUGGING
		if (verr(cur_block->b_ptr, 512) == 0) {
			dump_cache();
			ll_debug();
		}
#endif
		if ((wd_status & DRQ) == 0) {
			ll_debug();
		}
		inp_words(cur_block->b_ptr, 256, wd.data);
		*((unsigned *) (&cur_block)) = cur_block->b_next_off;

		/* Now check our status flags. */
		if (wd_status & ERROR) {
			lblock = io_current;
			res = 1;
			break;
		} else if (exflags.ecc_ignore && (wd_error & EMASK)) {
			ecc_blockno = io_current;
			SET_STATE(WAIT_ECC_ERR);
			if (++io_current == io_size) {
				lblock = ecc_blockno;
				res = 1;
			}
		} else if (++io_current == io_size) {
			res = 1;
			lblock = io_current;
		}
		break;
	case	WAIT_WRITE:
		if (io_current++ == io_size) {
			res = 1;
			lblock = io_current-1;
			break;
		}
#if DEBUGGING
		if (verw(cur_block->b_ptr, 512) == 0) {
			dump_cache();
			ll_debug();
		}
#endif
		if (wd_status & ERROR) {
			res = 1;
			lblock = io_current -1;
			break;
		}		
		if ((wd_status & DRQ) == 0) {
			ll_debug();
		}
		outp_words(cur_block->b_ptr, 256, wd.data);
		*((unsigned *) (&cur_block)) = cur_block->b_next_off;
		break;
	case	WAIT_ECC_ERR:	
		if ((wd_status & DRQ) == 0) {
			ll_debug();
		}
		inp_words(cur_block->b_ptr, 256, wd.data);
		*((unsigned *) (&cur_block)) = cur_block->b_next_off;
		if ((wd_status & ERROR) || ++io_current >= io_size) {
			res = 1;
			lblock = ecc_blockno;
		}
		break;
	default:
		ll_debug();		/* there is no default! */
	}
	if (res) {
#if DEBUGGING
		proxy_sequence++;
#endif
		SET_STATE(IDLE);
		disk[cur_disk].d_nblks = lblock;
		disk[cur_disk].d_retval = (lblock == 0) ?
				_DRVR_UNSPEC : _DRVR_OK;
		disk[cur_disk].d_busy = 0;
		return fsys_proxy;
	}
	return 0;
}


send_status()
{
	SET_STATE(WAIT_STATUS);
	io_out(wd.sect_count,disk[cur_disk].d_sctr_trk);
	io_out(wd.drv_head,0xa0 | (cur_disk << 4) | (disk[cur_disk].d_heads-1));
	io_out(wd.command, SET_PARMS);
	return 0;
}



int io_setup(int cmd, int nstate)	
{
	io_out(wd.precomp, disk[cur_disk].d_precomp >> 2); 
	io_out(wd.sect_no, bdata.sector);
	io_out(wd.sect_count, io_size);
	io_out(wd.cyl_low, bdata.track);
	io_out(wd.cyl_high, bdata.track >> 8);
	io_out(wd.drv_head, 0xa0 | (bdata.pdrive<<4) | (bdata.head & 0x0f));
	SET_STATE(nstate);
	io_out(wd.command,cmd);
	return 0;
}


#if DEBUGGING
dump_cache()
{
int	i;
BLOCK	__far *c;
	disp_str("!!!!cache pointer corrupt!!! reading ");
	disp_hex(io_size);
	disp_str("cache list=");
	for (i=0,c=first_block; i < io_size; i++) {
		disp_str("@");	disp_hex(i);	disp_lhex((long)c->b_ptr);
		c = MK_FP(FP_SEG(c), c->b_next_off);
	}
	ll_debug();
} 


int	wait_wd()
{
int	x;
int	timer = 0;
	while ((x=io_in(wd.status)) & BUSY) {
		if (--timer == 0)
			ll_debug();
	}
	if ((x=io_in(wd.status)) & ERROR) {
#if DEBUGGING
		disp_char('E');
		disp_hex(io_in(wd.error));
		disp_hex(x);
#else
		io_in(wd.error);
#endif
	}
	return x;
}
#endif
