#ifdef __USAGE
%C  [-a Address] [-c] [-d DMA] [-D...] [-f FsysName]
            [-i Interrupt] [-I] [-n DevName] [-N DriverName] [-m]
            [-s...] [-t UnitType] [-v...]

Options:
 -a          Address of base I/O Port (in hex) to use.
 -c          Fsys cache block mapping table size
 -d          DMA Channel (in hex) to use.
 -D          Change message level (-D -D sets to level 2)
 -f          Name Registered for local Fsys.
 -i          Interrupt Level (in hex) to use.
 -I          Use interrupts rather than busy/wait
 -n          Name to use as the block special device.
 -N          Name of driver to present to Fsys.
 -m          Maximum number (in decimal) of I/O requests from Fsys
 -s          Show selected configuration and quit (-s -s shows controller).
 -t          Type of unit:  Archive, Wangtek
 -v          Verbose, Show startup options (-v -v shows controller options).

Where:
               Address default is:     200h
   Default Fsys cache map size is:      64d
           DMA Channel default is:       1h
   Name of local Fsys defaults to: qnx/fsys
       Interrupt Level default is:       5h (if -I specified)
Name of block special defaults to:       tp
    Fsys' Driver Name defaults to:    qic02
          Max Fsys I/O default is:     128d
             Unit Type default is:  Wangtek
               Verbose default is:      OFF
#endif

/*
**	QIC02
**
**	Rob Oakley -- January 1992,
**		- re-write to remove interrupt dependancies
**		- written with level sensitive PS/2s in mind
**		- runs as a process to facilitate polling and exception conditions
**
**	$Id: qic02.c,v 1.3 1992/03/30 18:07:46 robo Exp robo $
**
**	$Log: qic02.c,v $
** Revision 1.3  1992/03/30  18:07:46  robo
** Incremented version from 'A' to 'B' and added rcs id and log variables.
**
*/
#include	<stdio.h>
#include	<stdarg.h>
#include	<stdlib.h>
#include	<setjmp.h>
#include	<signal.h>
#include	<string.h>
#include	<i86.h>
#include	<ctype.h>
#include	<errno.h>
#include	<unistd.h>

#include	<sys/disk.h>
#include	<sys/_dskdrvr.h>
#include	<sys/irqinfo.h>
#include	<sys/kernel.h>
#include	<sys/name.h>
#include	<sys/osinfo.h>
#include	<sys/proxy.h>
#include	<sys/qioctl.h>
#include	<sys/sched.h>
#include	<sys/seginfo.h>
#include	<sys/sys_msg.h>

#include	"sys/drvmount.h"
#include	"sys/dma.h"
#include	"sys/inline.h"
#include	"sys/io.h"
#include	"sys/timeout.h"

#include	"qic02.dfn"
#include	"qic02.typ"

#pragma off (check_stack)

/*
**	Simplification Macros
*/
#define TAPE_READY()		((INP( TUp->tps.status ) & TUp->tsbs.ready[ 0 ]) == TUp->tsbs.ready[ 1 ])
#define TAPE_EXCEPTION()	((INP( TUp->tps.status ) & TUp->tsbs.exception[ 0 ]) == TUp->tsbs.exception[ 1 ])
#define	TAPE_TERMINAL()		((INP( DMA_PORT + DMA_STAT ) & (1 << TapeDMA)) ? TRUE : FALSE)

#define	TAPE_DATA()			INP( TUp->tps.data )

#define	TAPE_COMMAND( c )	OUTP( TUp->tps.command, (c) )
#define	TAPE_CONTROL( c )	OUTP( TUp->tps.control, (c) )
#define	TAPE_DMA( c )		OUTP( TUp->tps.dma, (c) )

/*
**	This structure holds the unit specific ports, bits etc
*/
static TAPE_UNIT_STRUCT		 tus[] =
{	{	"Wangtek",							// Tape Unit Name
		{ 	0,0,1,1,0						// Offset of Status,Control,Command,Data & DMA
		},					
		{	{ BIT( 0 ),		0 },			// READY		(mask,value)
			{ BIT( 1 ),		0 }				// EXCEPTION	(mask,value)
		},
		{ 	0,BIT( 0 ),BIT( 1 ),BIT( 2 ),	// Control bit to clear,online,reset,request
			{ BIT( 3 ),BIT( 3 ),BIT( 4 )},	// DMA 1,2 & 3
			{ BIT( 5 ),BIT( 5 ),BIT( 5 )}	// Interrupt enable
		},
	},

	{	"Archive",							// Tape Unit Name
		{ 	1,1,0,0,2						// Port Offset of Status,Control,Command,Data & DMA
		},					
		{	{ BIT( 6 ),		0 },			// READY		(mask,value)
			{ BIT( 5 ),		0 }				// EXCEPTION	(mask,value)
		},
		{ 	0,0,BIT( 7 ),BIT( 6 ),			// Control bit to clear,online,reset,request
			{ BIT( 4 ),BIT( 4 ),BIT( 4 ) },	// DMA 1,2 & 3
			{ BIT( 5 ),BIT( 5 ),BIT( 5 ) }	// Interrupt enable
		},
	},

	{	"Tandy",							// Tape Unit Name
		{ 	0,0,1,1,0						// Offset of Status,Control,Command,Data & DMA
		},					
		{	{ BIT( 0 ),		0 },			// READY		(mask,value)
			{ BIT( 1 ),		0 }				// EXCEPTION	(mask,value)
		},
		{ 	0,BIT( 0 ),BIT( 1 ),BIT( 2 ),	// Control bit to clear,online,reset,request
			{ BIT( 3 ),BIT( 3 ),BIT( 3 )},	// DMA 1,2 & 3
			{ 0x08,0x08,0x08 }				// Interrupt enable
		},
	},

	{	"User Patch",						// Tape Unit Name
		{ 	0,0,0,0,0						// Offset of Status,Control,Command,Data & DMA
		},					
		{	{ BIT( 0 ),		0 },			// READY		(mask,value)
			{ BIT( 0 ),		0 }				// EXCEPTION	(mask,value)
		},
		{ 	0,BIT( 0 ),BIT( 0 ),BIT( 0 ),	// Control bit to clear,online,reset,request
			{ 0x00,0x00,0x00 },				// DMA 1,2 & 3
			{ 0x00,0x00,0x00 }				// Interrupt enable
		},
	},

	{ NULL }
};

static struct _sysmsg_version_reply		TapeVersionReply =
{	"qic02",	__DATE__,	VERSION,	RELEASE,	0
};

/*
**	Command Line Options (Port, Interrupt, DMA etc)
*/
static TAPE_UNIT_STRUCT			*TUp									= &tus[ 0 ];
static char						 TapeDevName[ _DRIVER_NAME_LEN + 1 ]	= "tp";
static char						 TapeFsysName[ _SHORT_NAME_MAX + 1 ]	= "qnx/fsys";
static int						 TapeAddress							= 0x200;
static int						 TapeInterrupt							= 5;
static int						 TapeDMA								= 1;
static unsigned					 TapeMaxIO								= DEFAULT_QIC02_IO;
static int						 TapeRunPolled							= TRUE;
static int						 TapeShowConfigThenQuit					= FALSE;
static int						 TapeVerbose							= FALSE;

static int						 TapeDriverInstalled;
static pid_t					 TapeAgentPid;
static pid_t					 TapeFsysPid;
static pid_t					 TapeFsysProxy;
static int						 TapeDriverProxy;
static int						 TapeIntProxy;

static sigjmp_buf				 TapeYankEnv;				// rude interruption

static struct _disk				 TapeInfo[ QIC02_LOG_UNITS ];
static int						 TapeCurrentUnit	= -1;
static unsigned	short			 TapeInterruptCounter;	// Increment on Interrupt
static unsigned	short			 TapeWriteErrors;
static unsigned	short			 TapeCumulativeWriteErrors;
static unsigned	short			 TapeLastCmd;			// read/write/io
static unsigned	short			 TapeLastError;			// Most Recent Error
static unsigned short			 TapeExceptionLevel;
static unsigned short			 TapeErrno2Return;		// errno given to client

static TAPE_CACHE_2_PHYS_STRUCT	*TapeCache2Phys;		// Selector to Physical 
static unsigned short			 TapeCache2PhysSize	= DEFAULT_QIC02_CACHE_TABLE_SIZE;
static unsigned	short			 TapeCache2PhysBase;

static TAPE_CMD_Q_STRUCT		*TapeCmdQ;				// Command Queue
static TAPE_CMD_Q_STRUCT		*TapeCmdHeadQp;			// Next Command Position
static TAPE_CMD_Q_STRUCT		*TapeCmdTailQp;			// Current Command Position

/*
**	Other Local Variables
*/
static char						*PNp				= "$Id$";	// Invocation name
static QIC02_STATUS_STRUCT		 TSS;							// Tape Status Structure
static unsigned					 IsPS2;

static QIC02_PMSG_STRUCT		 MessageBuffer[ MAX_PMSG_DEPTH ];
static QIC02_PMSG_STRUCT		*MessageBufferp		= MessageBuffer;
static QIC02_PMSG_STRUCT		*MessageBufferLogp;
static unsigned	short			 MessageNumber		= 1;
static int						 MessageLevel		= 1;		// level 1 - critical
																// level 2 - potential
																// level 3 - information


/*****************************************************************/
static int			PS2( void )
/*****************************************************************/
{	struct _osinfo		osdata;

	if ( qnx_osinfo( 0, &osdata ) == 0 )
		if ( strncmp( "PS/2", osdata.machine, 4 ) )
			return FALSE;
		else
			return TRUE;

	return -1;
}


/*****************************************************************/
static pid_t far	TapeISR()
/*****************************************************************/
{
	pid_t					ret_val = (pid_t) 0;
	static volatile int		InterruptInService;		// Glitch catcher

	if ( ! InterruptInService++ )
	{	TapeInterruptCounter++;
		ret_val = TapeIntProxy;
	}

	InterruptInService--;

	return ret_val;
}


/*****************************************************************/
static char			*ErrorString( unsigned error )
/*****************************************************************/
{
#ifdef	ERR
#undef	ERR
#endif
#define	ERR( Code,String )				 String,
static	char							*ErrorString[] =
{
#	include	"qic02err.h"
};
	return ErrorString[ error ];
}


/*****************************************************************/
static void			LogMessage( int level, unsigned int error, int line_num, char *fmt, ... )
/*****************************************************************/
{	char		*mp		= MessageBufferp->msg;
	char		*spacep	= "  ";
	va_list		 arg;


	if ( TapeDriverInstalled  &&  level > MessageLevel )
		return;

	if ( level == 0  &&  error != QIC02_ERR_LOG )
		TapeLastError = error;

	if ( TapeDriverInstalled )
	{	ltoa( (long) MessageNumber, mp, 10 );
		strcat( mp, spacep );
		ltoa( (long) line_num, &mp[ strlen( mp ) ], 10 );
		strcat( mp, spacep );
	}
	else
		*mp = '\0';

	strcat( mp, ErrorString( error ) );
	strcat( mp, spacep );

	/*
	**	Can't call printf-family -- so
	**	fake up a little string snooper that will
	**	handle all that we'll ever throw at it
	*/
	if ( fmt  &&  *fmt )
	{	va_start( arg, fmt );
		mp = &mp[ strlen( mp ) ];

		do
		{	if ( *fmt == '%' )
			{	switch ( *++fmt )
				{	case '%':	*mp++	= '%';
								*mp		= '\0';
					break;

					case 'd':	ltoa( (long) va_arg( arg, int ), mp, 10 );
					break;

					case 'u':	ltoa( (long) va_arg( arg, unsigned ), mp, 10 );
					break;

					case 'x':	ltoa( (long) va_arg( arg, int ), mp, 16 );
					break;

					case 'c':	*mp++	= va_arg( arg, unsigned char );
								*mp		= '\0';
					break;

					case 's':	strcpy( mp, va_arg( arg, char * ) );
					break;

					case 'l':
						switch ( *++fmt )
						{	case 'd':	ltoa( va_arg( arg, long ), mp, 10 );
							break;

							case 'x':	ltoa( va_arg( arg, unsigned long ), mp, 16 );
							break;

							default:
								*mp++	= '%';
								*mp++	= 'l';
								*mp++	= *fmt;
								*mp		= '\0';
						}
					break;

					default:	*mp++	= '%';
								*mp++	= *fmt;
								*mp		= '\0';
				}

				mp = &mp[ strlen( mp ) ];
				fmt++;
				continue;
			}
            else
				*mp++ = *fmt++;

		} while ( *fmt );

		*mp = '\0';
		va_end( arg );
	}


	if ( ! TapeDriverInstalled )
	{	write( fileno( stdout ), PNp, strlen( PNp ) );
		write( fileno( stdout ), ": ", 2 );
		write( fileno( stdout ), MessageBufferp->msg, strlen( MessageBufferp->msg ) );
		write( fileno( stdout ), "\n", 1 );
		*MessageBufferp->msg = '\0';
	}
	else
	{	if ( ++MessageBufferp  >=  &MessageBuffer[ MAX_PMSG_DEPTH ] )
			MessageBufferp = MessageBuffer;
		MessageNumber++;
	}
}

/*****************************************************************/
static char			*BitPrint( unsigned int bits, char *bit_string )
/*****************************************************************/
{   int				no_of_bits;

	if ( bits & 0xff00 )	no_of_bits = 16;
	else					no_of_bits = 8;

	*(bit_string += no_of_bits) = '\0';			// null terminate string

	while ( no_of_bits-- > 0 )
	{	*--bit_string = bits & 0x0001 ? '1' : '0';
		bits >>= 1;
	}

	return bit_string;
}


/*****************************************************************/
static void			TapeDumpParms( void )
/*****************************************************************/
{
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "   Tape Unit type is:  %s", TUp->tname );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "         Compiled On:  %s", TapeVersionReply.date );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "             Version:  %d.%d%c", TapeVersionReply.version / 100,
		TapeVersionReply.version - TapeVersionReply.version / 100 * 100, TapeVersionReply.letter );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "         I/O Address:  0x%x", TapeAddress );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "Cache Block Map Size:  %d", TapeCache2PhysSize );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "         DMA Channel:  %d", TapeDMA );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "       Message Level:  %d", MessageLevel );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "  Name of Local Fsys:  %s", TapeFsysName );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "     Interrupt Level:  %d", TapeInterrupt );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "       /dev/ name is:  %s", TapeDevName );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "Name to Give to Fsys:  %s", TapeVersionReply.name );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "             Max I/O:  %u", TapeMaxIO );
}


/*****************************************************************/
static void			TapeDumpConfig( void )
/*****************************************************************/
{	TAPE_UNIT_STRUCT		*tup = TUp;
	char					 s1[ 9 ],
							 s2[ 9 ],
							 s3[ 9 ];

	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "                           Unit is: %s", tup->tname );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "Status,Control,Command, Data & DMA: 0x%x,0x%x,0x%x,0x%x,0x%x", 
					tup->tps.status, tup->tps.control, tup->tps.command,
					tup->tps.data, tup->tps.dma );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "                  Ready Mask,Value: %s,%s", BitPrint( tup->tsbs.ready[0],s1 ), BitPrint( tup->tsbs.ready[1],s2 ) );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "              Exception Mask,Value: %s,%s", BitPrint( tup->tsbs.exception[0],s1 ), BitPrint( tup->tsbs.exception[1],s2 ) );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "                             Clear: %s", BitPrint( tup->tcbs.clear,s1 ) );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "                            Online: %s", BitPrint( tup->tcbs.online,s1 ) );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "                             Reset: %s", BitPrint( tup->tcbs.reset,s1 ) );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "                           Request: %s", BitPrint( tup->tcbs.request,s1 ) );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "                  DMA 1,2,3 Enable: %s,%s,%s", BitPrint( tup->tcbs.dma[0],s1 ), BitPrint( tup->tcbs.dma[1],s2 ), BitPrint( tup->tcbs.dma[2], s3 ) );
	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "                  Interrupt Enable: %s,%s,%s", BitPrint( tup->tcbs.inton[0],s1 ), BitPrint( tup->tcbs.inton[1],s2 ), BitPrint( tup->tcbs.inton[2], s3 ) );
}


/*****************************************************************/
static int			TapeWaitForReady( int true_false, unsigned usecs_max )
/*****************************************************************/
{
	unsigned	usecs;

	for ( usecs = 0;  usecs <= usecs_max;  usecs += USECS_INC )
	{	if ( TAPE_READY() == true_false )
			break;
		usec_delay( USECS_INC / 2 );
		if ( TAPE_READY() == true_false )
			break;
		usec_delay( USECS_INC / 2 );
	}

	return (usecs <= usecs_max) ? EOK : -1;
}


/*****************************************************************/
static int			TapeWaitForException( int true_false, unsigned usecs_max )
/*****************************************************************/
{
	unsigned	usecs;

	for ( usecs = 0;  usecs <= usecs_max;  usecs += USECS_INC )
	{	usec_delay( USECS_INC );
		if ( TAPE_EXCEPTION() == true_false )
			break;
	}

	return (usecs <= usecs_max) ? EOK : -1;
}


/*****************************************************************/
static int			TapeResetBW( void )
/*****************************************************************/
{	TapeLastCmd = T_RESET;

	/*
	**	This will help a user keep track of whether or
	**	not his/her tape heads are getting dirty.
	**	(Preserving the error count across resets)
	*/
	TapeCumulativeWriteErrors += TapeWriteErrors;

	TAPE_CONTROL( TUp->tcbs.reset );
	usec_delay( 25 );

	if ( TapeWaitForReady( FALSE, 5000 ) != EOK )
	{	LogMessage( LEVEL_1, QIC02_ERR_NOT_READY, __LINE__, NULL );
		return -1;
	}

	if ( TapeWaitForException( TRUE, 5000 ) != EOK )
	{	LogMessage( LEVEL_1, QIC02_ERR_NOT_EXCEPTION, __LINE__, NULL );
		return -1;
	}

	TAPE_CONTROL( TUp->tcbs.clear );

	TapeExceptionLevel	= EXC_CLEAR;
	TapeErrno2Return	= EOK;

	sleep( 5 );

	return EOK;
}


/*****************************************************************/
static int			TapeReadStatusBW( QIC02_STATUS_STRUCT *tssp )
/*****************************************************************/
{	char	 *sb = (char *) tssp,
			 *eb = (char *) tssp + sizeof *tssp;

	if ( NOT TAPE_EXCEPTION() )
		if ( TapeWaitForReady( TRUE, 50000 ) != EOK )
		{	LogMessage( LEVEL_1, QIC02_ERR_NOT_READY, __LINE__, NULL );
			return -1;
		}

	TAPE_COMMAND( TapeLastCmd = T_READ_STATUS );
	TAPE_CONTROL( TUp->tcbs.online | TUp->tcbs.request );

	if ( TapeWaitForReady( TRUE, 50000 ) != EOK )
	{	LogMessage( LEVEL_1, QIC02_ERR_NOT_READY, __LINE__, NULL );
		return -1;
	}

	TAPE_CONTROL( TUp->tcbs.online );
	usec_delay( 200 );

	do
	{	if ( TapeWaitForReady( TRUE, 50000 ) != EOK )
		{	LogMessage( LEVEL_1, QIC02_ERR_NOT_READY, __LINE__, "on byte %d", 6-(eb-sb) );
			return -1;	
		}

		*sb++ = TAPE_DATA();
	
		TAPE_CONTROL( TUp->tcbs.online | TUp->tcbs.request );

		if ( TapeWaitForReady( FALSE, 50000 ) != EOK )
		{	LogMessage( LEVEL_1, QIC02_ERR_NOT_READY, __LINE__, "on byte %d", 6-(eb-sb) );
			return -1;
		}

		TAPE_CONTROL( TUp->tcbs.online );

	}	while ( sb < eb );

	// We will keep track of errors even across resets
	TapeWriteErrors += tssp->ErrorCounth << 8 | tssp->ErrorCountl;

	if ( TapeExceptionLevel & EXC_FATAL_CLEAR_IOCNTL )
	{	TapeErrno2Return	= EOK;
		TapeExceptionLevel &= ~EXC_FATAL_CLEAR_IOCNTL;
	}


	return EOK;
}


/*****************************************************************/
static unsigned		TapeInterpretException( void )
/*****************************************************************/
{	unsigned					 ret_val		= -1;
	unsigned					 previous_cmd	= TapeLastCmd;	

	TapeExceptionLevel	= EXC_CLEAR;
	TapeErrno2Return	= EOK;

	/*
	**	Do a read status transparently (preserve last command)
	*/
	if ( TapeReadStatusBW( &TSS ) == EOK )
	{	TapeLastCmd				= previous_cmd;

		if ( TSS.Byte0 & E_ST0 )
		{	if ( TSS.Byte0 & (E_FIL | E_EOM) )
			{	if ( TSS.Byte0 & E_FIL )
				{	TapeExceptionLevel |= EXC_NON_FATAL;

					if ( (TapeExceptionLevel & EXC_FATAL) == FALSE )
						TapeErrno2Return = ESTALE;
				}
				else
				{	TapeExceptionLevel |= EXC_FATAL_CLEAR_IOCNTL;

					if ( (TapeExceptionLevel & EXC_FATAL) == FALSE )
						TapeErrno2Return = ESTALE;
				}

				if ( TapeLastCmd == T_READ  ||  TapeLastCmd == T_WRITE )
				{	if ( TapeCmdTailQp == TapeCmdHeadQp )
					{	ret_val = _DRVR_OK;
						LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "FIL/EOM - R/W Exact" );
					}
					else
					{	ret_val = _DRVR_PART_IO;

						if ( TSS.Byte0 & E_EOM )
							if ( (TapeExceptionLevel & EXC_FATAL) == FALSE )
								TapeErrno2Return = ENOSPC;

						LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "FIL/EOM - R/W Partial" );
					}
				}
				else
				{	ret_val = _DRVR_OK;
					LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "FIL/EOM - NON R/W" );
				}       	
			}

			if ( TSS.Byte0 & (E_BNL | E_UDE | E_DFF) )
			{	TapeExceptionLevel |= EXC_FATAL_CLEAR_IOCNTL;
				if ( (TapeExceptionLevel & EXC_FATAL) == FALSE )
					TapeErrno2Return	= EBADFSYS;
				LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "Block not found/Data Error" );
			}

			if ( TSS.Byte0 & E_WRP )
			{	TapeExceptionLevel |= EXC_NON_FATAL;
				if ( (TapeExceptionLevel & EXC_FATAL) == FALSE )
					TapeErrno2Return	= EROFS;
				LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "Write Protect Cartridge" );
			}

			if ( TSS.Byte0 & E_CNI )
			{	TapeExceptionLevel |= EXC_NON_FATAL;
				if ( (TapeExceptionLevel & EXC_FATAL) == FALSE )
					TapeErrno2Return	= EIO;
				LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "Cartridge not in place" );
			}
		}
		else if ( TSS.Byte1 & E_ST1 )
		{	if ( TSS.Byte1 & E_POR )
			{	TapeExceptionLevel |= EXC_FATAL_CLEAR_IOCNTL;
				if ( (TapeExceptionLevel & EXC_FATAL) == FALSE )
					TapeErrno2Return	= EL3RST;
				LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "POR" );
			}

			if ( TSS.Byte1 & E_NDT )
			{	TapeExceptionLevel |= EXC_FATAL_CLEAR_IOCNTL;
				if ( (TapeExceptionLevel & EXC_FATAL) == FALSE )
					TapeErrno2Return	= EIO;
				LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "NDT" );
			}

			if ( TSS.Byte1 & (E_ILL | E_PAR) )
			{	TapeExceptionLevel |= EXC_FATAL;
				TapeErrno2Return	= EIO;
				LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "ILL/PAR" );
			}

			if ( TSS.Byte1 & E_EOD )
			{	if ( TapeLastCmd == T_READ )
				{	TapeExceptionLevel |= EXC_NON_FATAL;

					if ( TapeCmdTailQp == TapeCmdHeadQp )
					{	ret_val = _DRVR_OK;
						LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "EOD (read exact)" );
					}
					else
					{	ret_val = _DRVR_PART_IO;

						if ( TSS.Byte0 & E_EOM )
							if ( (TapeExceptionLevel & EXC_FATAL) == FALSE )
								TapeErrno2Return = ENOSPC;

						LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "EOD (read partial)" );
					}
				}
				else
				{	TapeExceptionLevel |= EXC_FATAL_CLEAR_IOCNTL;
					ret_val = _DRVR_OK;
					LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "EOD" );
				}
			}

			if ( TSS.Byte1 & E_BOM )
			{	TapeExceptionLevel |= EXC_NON_FATAL;
				ret_val = _DRVR_OK;
				LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "BOM sought and found" );
			}
			
			if ( TSS.Byte1 & E_MBD )
			{	TapeExceptionLevel |= EXC_NON_FATAL;
				ret_val = _DRVR_OK;
				LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "Marginal Block Detected, replace tape / clean heads" );
			}
		}
		else
		{	TapeExceptionLevel |= EXC_FATAL;
			TapeErrno2Return	= EIO;
			LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "Exception but no status" );
		}
	}

	TapeCmdHeadQp = TapeCmdTailQp;

	return ret_val;
}


/*****************************************************************/
static unsigned		TapeWaitTillDone( void )
/*****************************************************************/
{	int				exception;
	unsigned		ret_val	= _DRVR_OK;


	if ( TapeRunPolled )
	{	if (	TapeLastCmd == T_READ	||
				TapeLastCmd == T_WRITE	)

		{	while ( (exception = TAPE_EXCEPTION()) == FALSE  &&  TAPE_TERMINAL() == FALSE );
			while ( (exception = TAPE_EXCEPTION()) == FALSE  &&  TAPE_READY() == FALSE );
		}
        else
			while ( (exception = TAPE_EXCEPTION()) == FALSE  &&  TAPE_READY() == FALSE );
		
		if ( exception )
			ret_val = TapeInterpretException();
	}
    else
		while ( 1 )
		{	if ( Receive( TapeIntProxy,NULL,0 ) )
			{	if ( TAPE_EXCEPTION() )
				{	while ( Creceive( TapeIntProxy,NULL,0 ) != -1 )
						;
					ret_val = TapeInterpretException();
					break;
				}
			}
			if (	TapeLastCmd == T_READ	||
					TapeLastCmd == T_WRITE	)
			{	if ( TAPE_TERMINAL()  &&  TAPE_READY() )
				{	while ( Creceive( TapeIntProxy,NULL,0 ) != -1 )
						;
					break;
				}
			}
			else
				if ( TAPE_READY() )
					break;
		}


	return ret_val;
}

/*****************************************************************/
static int			InBackGround( void )
/*****************************************************************/

{	struct sigaction	sigact_int,
						sigact_quit;

	sigaction( SIGINT, NULL, &sigact_int );
	sigaction( SIGQUIT, NULL, &sigact_quit );

	if ( sigact_int.sa_handler	== SIG_IGN	&&
		 sigact_quit.sa_handler	== SIG_IGN	)
		return TRUE;
		
	return FALSE;
}


/*****************************************************************/
static void			TapeYankSignalHandler( int signal_no )
/*****************************************************************/
{	siglongjmp( TapeYankEnv, -1 );
	signal_no = signal_no;				// silence wcc
}

/*****************************************************************/
static unsigned int	TapeFindBaseCacheSelector( unsigned no_of_sels, unsigned far *sels )
/*****************************************************************/
{	unsigned int	i;
	unsigned int	sel_smallest = (unsigned int) (-1);

	for ( i = 0;  i < no_of_sels;  i++,sels++ )
	{	if ( *sels < sel_smallest )
			sel_smallest = *sels;

		// paranoid code (make sure selectors increment by 1)
		if ( i > 0 )
		{	int		s1 = (int) (*sels		>> 3),
					s2 = (int) (*(sels-1)	>> 3);

			if ( abs( s2-s1 ) != 1 )
				return (unsigned int) (-1);
		}
	}

	return sel_smallest;
}


/*****************************************************************/
static int			TapeCommonCommand( unsigned int command )
/*****************************************************************/
{	TapeLastCmd = command;

	if ( TapeWaitForReady( TRUE, 50000 ) != EOK )
	{	LogMessage( LEVEL_1, QIC02_ERR_NOT_READY, __LINE__, NULL );
		return -1;
	}

	TAPE_CONTROL( TUp->tcbs.online );
	TAPE_COMMAND( command );
	TAPE_CONTROL( TUp->tcbs.online | TUp->tcbs.request );

	if ( TapeWaitForReady( TRUE, 50000 ) != EOK )
	{	LogMessage( LEVEL_1, QIC02_ERR_NOT_READY, __LINE__, NULL );
		return -1;
	}

	TAPE_CONTROL( TUp->tcbs.online |
					TapeRunPolled ? TUp->tcbs.clear : TUp->tcbs.inton[ TapeDMA - 1 ] );

	usec_delay( 100 );

	TapeExceptionLevel &= ~EXC_FATAL_CLEAR_IOCNTL;

	return EOK;
}


/*****************************************************************/
static int			RWCommon( int rw )
/*****************************************************************/
{	unsigned		i;
	unsigned		exception;


	// Disable DMA
	if ( IsPS2 )
		clear_dma( IsPS2,TapeDMA );
		
	TAPE_CONTROL( TUp->tcbs.online );

	if ( ! IsPS2 )
		clear_dma( IsPS2,TapeDMA );

	if ( TapeWaitForReady( TRUE, 50000 ) != EOK )
	{	LogMessage( LEVEL_1, QIC02_ERR_NOT_READY, __LINE__, NULL );
		return -1;
	}

	TAPE_COMMAND( TapeLastCmd = rw );
	TAPE_CONTROL( TUp->tcbs.online );
	TAPE_CONTROL( TUp->tcbs.online | TUp->tcbs.request );

	if ( TapeWaitForReady( TRUE, 50000 ) != EOK )
	{	LogMessage( LEVEL_1, QIC02_ERR_NOT_READY, __LINE__, NULL );
		return -1;
	}
	
	TAPE_CONTROL( TUp->tcbs.online );

	usec_delay( 100 );

	for ( i = 2000;  i > 0;  i-- )
		if ( (exception = TAPE_EXCEPTION())  ||  TapeWaitForReady( TRUE, 10000 ) == EOK )
			break;
		else if ( i == 1 )
		{	LogMessage( LEVEL_1, QIC02_ERR_NOT_READY, __LINE__, NULL );
			return -1;
		}

	if ( exception )
	{	TapeInterpretException();
		return -1;
	}
    else
		return EOK;
}


/*****************************************************************/
static int			RW( int rw, TAPE_DMA_BUFFER_STRUCT *tdbp )
/*****************************************************************/
{	if ( TapeLastCmd == rw )
	{	// Disable DMA
		if ( IsPS2 )
			clear_dma( IsPS2,TapeDMA );

		TAPE_CONTROL( TUp->tcbs.online );

		if ( ! IsPS2 )
			clear_dma( IsPS2,TapeDMA );
	}
	else
		if ( RWCommon( rw ) != EOK )
			return -1;

	setup_dma( (rw == T_WRITE ? DMA_OUTPUT : 0) | (IsPS2 ? IsPS2 : DMA_WAIT), &tdbp->buffer, tdbp->no_of_blocks * _BLOCK_SIZE, TapeDMA );

	if ( TUp->tps.control != TUp->tps.dma )
		TAPE_DMA( 0x00 );

	TAPE_CONTROL(	TUp->tcbs.online |
					TUp->tcbs.dma[ TapeDMA - 1 ] |
					(TapeRunPolled ? TUp->tcbs.clear : TUp->tcbs.inton[ TapeDMA - 1 ]) );

	if ( ! IsPS2 )
		kick_dma( IsPS2,TapeDMA );


	return EOK;
}


/*
**	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
**					Fsys Entry Points Follow					
**	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
*/
/*****************************************************************/
static int far		DriverInit( pid_t Proxy, unsigned tablesize, unsigned far *cachetabp )
/*****************************************************************/
{	struct _seginfo				 seg_info;

	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "Driver Initialize" );

	TapeFsysProxy = Proxy;

	if ( (TapeDriverProxy = qnx_proxy_attach( TapeAgentPid, NULL, 0, -1 )) == -1 )
	{	LogMessage( LEVEL_1, QIC02_ERR_PROXY, __LINE__, "(Fsys -> driver)" );
		return -1;
	}

	if ( tablesize > TapeCache2PhysSize )
	{	LogMessage( LEVEL_1, QIC02_ERR_TBL_SIZE, __LINE__, NULL );
		return -1;
	}

	if ( (TapeCache2PhysBase = TapeFindBaseCacheSelector( TapeCache2PhysSize = tablesize, cachetabp )) == (unsigned int) (-1) )
	{	LogMessage( LEVEL_1, QIC02_ERR_TBL_OFFSET, __LINE__, NULL );
		return -1;
	}

	/*
	**	Get Physical address of Fsys' cache blocks
	*/
	while ( tablesize-- )
	{	TAPE_CACHE_2_PHYS_STRUCT	*tc2pp = &TapeCache2Phys[ CACHE_2_PHYS_OFFSET( *cachetabp,TapeCache2PhysBase ) ];


		if ( qnx_segment_info( 0, TapeFsysPid, *cachetabp, &seg_info ) == -1 )
		{	LogMessage( LEVEL_1, QIC02_ERR_SEL_INFO, __LINE__, NULL );
			return -1;
		}

		tc2pp->phys_addr	= seg_info.addr;
		tc2pp->nbytes		= seg_info.nbytes;

		cachetabp++;
	}

	return EOK;
}


/*****************************************************************/
static int far		DriverTerminate( unsigned flags, unsigned reserve1, unsigned reserve2, unsigned reserve3 )
/*****************************************************************/
{
	flags		= flags;
	reserve1	= reserve1;
	reserve2	= reserve2;
	reserve3	= reserve3;

	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "Driver Terminate" );

	if ( TapeCmdQ )
	{	free( TapeCmdQ );
		TapeCmdQ = NULL;
	}

	if ( TapeCache2Phys )
	{	free( TapeCache2Phys );
		TapeCache2Phys = NULL;
	}

	return EOK;
}


/*****************************************************************/
static int far		DriverUnitInit( unsigned unit, unsigned pflag, unsigned reserve1, unsigned reserve2 )
/*****************************************************************/
{
	pflag		= pflag;
	reserve1	= reserve1;
	reserve2	= reserve2;

	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "Unit (%d) Initialize", unit );

	return EOK;
}


/*****************************************************************/
static int far		DriverUnitTerm( unsigned unit, unsigned reserve1, unsigned reserve2, unsigned reserve3 )
/*****************************************************************/
{	reserve1	= reserve1;
	reserve2	= reserve2;
	reserve3	= reserve3;

	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "Unit (%d) Terminate", unit );

	return EOK;
}


/*****************************************************************/
static int far		DriverRW( unsigned unit, unsigned mode, unsigned blockcount, unsigned reserved, struct _block far *cachep )
/*****************************************************************/
{	struct _disk			*tip		= &TapeInfo[ unit ];

	reserved	= reserved;

	LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, (mode == _READ_BLK ? "Read %u blocks" : "Write %u blocks"), blockcount );

	if ( TapeCurrentUnit == -1 )
	{	LogMessage( LEVEL_1, QIC02_ERR_NOT_OPEN, __LINE__, NULL );
		return -1;
	}

	if ( TapeCurrentUnit != unit )
	{	LogMessage( LEVEL_1, QIC02_ERR_OPEN, __LINE__, "%d", unit );
		return -1;
	}

	tip->d_busy		= TRUE;
	tip->d_nblks	= 0;
	tip->d_retval	= EOK;

	if ( TapeCmdHeadQp != TapeCmdTailQp )
	{	LogMessage( LEVEL_1, QIC02_ERR_IO_BUFF, __LINE__, NULL );
		tip->d_retval	= _DRVR_UNSPEC;
		tip->d_busy		= FALSE;
		return -1;
	}

	while ( blockcount-- )
	{	unsigned	offset = CACHE_2_PHYS_OFFSET( FP_SEG( cachep->b_ptr ), TapeCache2PhysBase );


		if ( offset >= TapeCache2PhysSize )
		{	TapeCmdHeadQp	= TapeCmdTailQp;
			LogMessage( LEVEL_1, QIC02_ERR_SELECTOR, __LINE__, "offset = %u, tablesize = %u", offset, TapeCache2PhysSize );
			tip->d_retval	= _DRVR_UNSPEC;
			tip->d_busy		= FALSE;
			return -1;
		}

		if ( TapeCmdHeadQp->np == TapeCmdTailQp )
		{	LogMessage( LEVEL_1, QIC02_ERR_IO_BUFF, __LINE__, NULL );
			tip->d_retval	= _DRVR_UNSPEC;
			tip->d_busy		= FALSE;
			return -1;
		}

		TapeCmdHeadQp = TapeCmdHeadQp->np;
		memset( &TapeCmdHeadQp->flag, 0, sizeof TapeCmdHeadQp->flag );
		TapeCmdHeadQp->tip							 = tip;
		TapeCmdHeadQp->data.dma.block_no			 = cachep->b_block;
		TapeCmdHeadQp->data.dma.no_of_blocks		 = 1;
#ifdef DEBUG
		TapeCmdHeadQp->data.dma.buffer.bufp			 = MK_FP( TapeCache2Phys[ offset ].lsel, FP_OFF( cachep->b_ptr ) );
#endif
		TapeCmdHeadQp->data.dma.buffer.phys_addr	 = TapeCache2Phys[ offset ].phys_addr + FP_OFF( cachep->b_ptr );
		cachep										 = MK_FP( FP_SEG( cachep ), cachep->b_next_off );

		if ( mode == _READ_BLK )
			TapeCmdHeadQp->flag.read	= TRUE;
		else
			TapeCmdHeadQp->flag.write	= TRUE;
	}

	// Give the agent a tap on the shoulder
	Trigger( TapeDriverProxy );

	return EOK;
}

/*
**	There are three "flavours" of IOCNTLS.
**		1) Synchronous (no proxy) agent assistance not needed
**		2) Asynchronous (proxy Fsys directly)
**		3) Asynchronous (proxy Agent, Agent proxies Fsys)
*/
/*****************************************************************/
static int far		DriverControl( unsigned unit, unsigned request, unsigned reserve1, unsigned reserve2, void far *datap )
/*****************************************************************/
{	int							 ret_val	= EOK;
	struct _disk				*tip		= &TapeInfo[ unit ];
	QIC02_MSG_STRUCT far		*requestdatap;


	reserve1 = reserve1;
	reserve2 = reserve2;

	TapeExceptionLevel			&= ~EXC_FATAL_CLEAR_IOCNTL;

	tip->d_nblks				 = 1;		// Must be 1 for successful iocntl
	tip->d_busy					 = FALSE;	// Assume async
	tip->d_retval				 = EOK;		// Operation suceeded

	requestdatap 				 = datap;	/* Point this to something we can access */
	requestdatap->Header.RetVal  = EOK;

	if ( request != QCTL_RAW_CMD )
		LogMessage( LEVEL_2, QIC02_ERR_LOG, __LINE__, "Unit (%d) IOCTL 0x%x", unit,request );

	switch( request )
	{	case QCTL_GET_LAST_ERR:
			requestdatap->Header.RetVal = TapeLastError;
			_fstrcpy( requestdatap->Body.Error.ErrorString, (char far *) ErrorString( TapeLastError ) );
		break;

		case QCTL_RESET:
			tip->d_busy				= TRUE;
			kill( TapeAgentPid, SIGUSR1 );	// driver will trigger Fsys
		break;

		case QCTL_RAW_CMD:
			if ( requestdatap->Header.Command != T_DEBUG )
				LogMessage( LEVEL_2, QIC02_ERR_LOG, __LINE__, " Unit (%d) IOCTL      RAW 0x%x", unit,requestdatap->Header.Command );

			switch ( requestdatap->Header.Command )
			{	case T_READ_STATUS:
					if ( TapeReadStatusBW( &TSS ) == EOK )
					{	TSS.ErrorCounth	= TapeCumulativeWriteErrors >>	8;
						TSS.ErrorCountl	= TapeCumulativeWriteErrors &	0x00ff;
						requestdatap->Body.Status = TSS;
					}
					else
					{	requestdatap->Header.RetVal = TapeLastError;
						_fmemset( &requestdatap->Body.Status, 0, sizeof (requestdatap->Body.Status) );
					}
				break;

				case T_LAST_ERROR:
					requestdatap->Header.RetVal = TapeLastError;
					_fstrcpy( requestdatap->Body.Error.ErrorString, (char far *) ErrorString( TapeLastError ) );
				break;

				case T_DEBUG:
					if ( MessageBufferLogp == NULL )
					{	LogMessage( LEVEL_2, QIC02_ERR_LOG, __LINE__, "Unit (%d) IOCTL (dump log)", unit );
						if ( *MessageBuffer[ MAX_PMSG_DEPTH - 1 ].msg )
							MessageBufferLogp = MessageBufferp;
						else
							MessageBufferLogp = MessageBuffer;
					}

					_fstrcpy( requestdatap->Body.Log.Msg, (char far *) MessageBufferLogp->msg );

					if ( ++MessageBufferLogp >= &MessageBuffer[ MAX_PMSG_DEPTH ] )
						MessageBufferLogp = MessageBuffer;

					if ( *MessageBufferLogp->msg == '\0'  ||  MessageBufferLogp == MessageBufferp )
					{	MessageBufferLogp = NULL;
						requestdatap->Body.Log.Again = FALSE;
					}
					else
						requestdatap->Body.Log.Again = TRUE;
				break;

				case T_VERSION:
					requestdatap->Body.Version = TapeVersionReply;
				break;

				case T_RESET:
					tip->d_busy = TRUE;
					kill( TapeAgentPid, SIGUSR1 );
				break;

				default:
					if ( TapeCmdHeadQp->np == TapeCmdTailQp )
					{	LogMessage( LEVEL_1, QIC02_ERR_IO_BUFF, __LINE__, NULL );
						tip->d_retval				= -1;
						tip->d_error				= EIO;
						tip->d_busy					= FALSE;
						tip->d_nblks				= 0;
						requestdatap->Header.RetVal	= TapeLastError;

						Trigger( TapeFsysProxy );
					}
                    else
					{	TapeCmdHeadQp								= TapeCmdHeadQp->np;
						memset( &TapeCmdHeadQp->flag, 0, sizeof TapeCmdHeadQp->flag );
						TapeCmdHeadQp->tip							= tip;
						TapeCmdHeadQp->tip->d_nblks					= 0;
						TapeCmdHeadQp->tip->d_busy					= TRUE;
						TapeCmdHeadQp->flag.common_command			= TRUE;
						TapeCmdHeadQp->data.command					= requestdatap->Header.Command;
	
						Trigger( TapeDriverProxy );
					}
				break;
			}
		break;

		default:
			ret_val			= -1;
			tip->d_error	= EIO;

	}

	return ret_val;
}


/*****************************************************************/
static int far		DriverControlMask( unsigned unit, unsigned request, void far *datap )
/*****************************************************************/
{	QIC02_MSG_STRUCT far	*requestdatap	= datap;
	int						 iflags			= _DRVR_CTRL_SUPPORTED | _DRVR_CTRL_ASYNCH;


	if ( request != QCTL_RAW_CMD  )
		LogMessage( LEVEL_2, QIC02_ERR_LOG, __LINE__, "Unit (%d) IOCTL Mask 0x%x", unit,request );

	switch( request )
	{	case QCTL_GET_LAST_ERR:
			iflags &= ~_DRVR_CTRL_ASYNCH;
		break;

		case QCTL_RAW_CMD:
			switch ( requestdatap->Header.Command )
			{	case T_READ_STATUS:
				case T_LAST_ERROR:
				case T_VERSION:
					LogMessage( LEVEL_2, QIC02_ERR_LOG, __LINE__, "Unit (%d) IOCTL Mask RAW 0x%x (no proxy)", unit,requestdatap->Header.Command );
				case T_DEBUG:
					iflags &= ~_DRVR_CTRL_ASYNCH;
				break;

				case T_WRITE_FM:
					LogMessage( LEVEL_2, QIC02_ERR_LOG, __LINE__, "Unit (%d) IOCTL Mask RAW (FIL) (proxy)", unit );
				break;

				default:
					LogMessage( LEVEL_2, QIC02_ERR_LOG, __LINE__, "Unit (%d) IOCTL Mask RAW 0x%x (proxy)", unit,requestdatap->Header.Command );
					iflags |=  _DRVR_CTRL_DESTRUCTIVE;
			}
		break;

		default:
			LogMessage( LEVEL_1, QIC02_ERR_U_MSG, __LINE__, "IOCTL = 0x%x", request );
			iflags  = 0;
	}

	return iflags;
}

/*****************************************************************/
static int far		DriverUnitOpen( unsigned unit, unsigned reserved1, unsigned reserved2, unsigned reserved3 )
/*****************************************************************/
{	reserved1	= reserved1;
	reserved2	= reserved2;
	reserved3	= reserved3;

	if ( TapeCmdTailQp != TapeCmdHeadQp  ||  TapeInfo[ unit ].d_busy )
	{	LogMessage( LEVEL_1, QIC02_ERR_OPEN, __LINE__, "Unit (%d) -- BUSY (last operation pending)", unit );
		return -1;
	}

	if ( TapeCurrentUnit != -1  &&  TapeCurrentUnit != unit )
	{	LogMessage( LEVEL_1, QIC02_ERR_OPEN, __LINE__, "Unit (%d) -- BUSY (already open)", unit );
		return -1;
	}

	LogMessage( LEVEL_2, QIC02_ERR_LOG, __LINE__, "Unit (%d) Open", unit );

	TapeCurrentUnit = unit;


	return EOK;
}

/*****************************************************************/
static int far		DriverUnitClose( unsigned unit, unsigned reserved1, unsigned reserved2, unsigned reserved3 )
/*****************************************************************/
{	unsigned					 ret_val = EOK;

	static QIC02_MSG_STRUCT		 ScratchMsgStruct;

	reserved1 = reserved1;
	reserved2 = reserved2;
	reserved3 = reserved3;

	if ( TapeCurrentUnit != -1  &&  TapeCurrentUnit != unit )
	{	LogMessage( LEVEL_1, QIC02_ERR_LOG, __LINE__, "Unit (%d) Close - NOT OPEN", unit );
		ret_val = -1;
	}
	else
	{	if ( TapeLastCmd == T_WRITE )
		{	memset( &ScratchMsgStruct, 0, sizeof ScratchMsgStruct );
			ScratchMsgStruct.Header.Command = T_WRITE_FM;
			DriverControl( unit, QCTL_RAW_CMD, (unsigned) 0, (unsigned) 0, (void far *) &ScratchMsgStruct );
		}

		if ( unit & 1 )
		{	memset( &ScratchMsgStruct, 0, sizeof ScratchMsgStruct );
			ScratchMsgStruct.Header.Command = T_BOT;
			DriverControl( unit, QCTL_RAW_CMD, (unsigned) 0, (unsigned) 0, (void far *) &ScratchMsgStruct );
		}
    
		TapeLastCmd			 = 0;
		TapeExceptionLevel	&= ~EXC_NON_FATAL;
		TapeCurrentUnit		 = -1;

		LogMessage( LEVEL_2, QIC02_ERR_LOG, __LINE__, "Unit (%d) Close", unit );
	}


	return ret_val;
}

/*
*****************************************************************
**
**	Driver Entry Points
**
*****************************************************************
*/
static int			(*DriverEntryList[ _DRIVER_TBL_SIZE ])() =
{	DriverInit,
	DriverTerminate,
	DriverUnitInit,
	DriverUnitTerm,
	DriverRW,
	DriverControl,
	DriverControlMask,
	NULL,
	DriverUnitOpen,
	DriverUnitClose
};


/*****************************************************************/
int					main( int argc, char *argv[] )
/*****************************************************************/
{	int				  i;
	char			 *correct_unit = "Was correct unit selected?";
	long			  stack[ QIC02_STACK_SIZE ];

	extern char		*optarg;
	extern int		 optind,
					 opterr,
					 optopt;


	PNp		= argv[0];				// program name pointer
	opterr	= 0;					// disable getopt printing of errors

	get_cpu_speed();				// used in timing loops

	while ( (i = getopt( argc, argv, "a:c:d:Df:i:Im:n:N:st:v" )) != -1 )
		switch ( i )
		{	case 'a':	TapeAddress			= strtol( optarg,NULL,16 );								break;
			case 'c':	TapeCache2PhysSize	= strtol( optarg,NULL,16 );								break;
			case 'd':	TapeDMA				= strtol( optarg,NULL,16 );								break;
			case 'D':	MessageLevel++;																break;
			case 'f':	strncpy( TapeFsysName, optarg, sizeof TapeFsysName );						break;
			case 'i':	TapeInterrupt		= strtol( optarg,NULL,16 );				// fall through and turn off poll
			case 'I':	TapeRunPolled		= FALSE;												break;
			case 'm':	TapeMaxIO			= strtol( optarg,NULL,10 );								break;
			case 'n':	strncpy( TapeDevName, optarg, sizeof TapeDevName );							break;
			case 'N':	strncpy( TapeVersionReply.name, optarg, sizeof TapeVersionReply.name - 1 ); break;
			case 's':	TapeShowConfigThenQuit++;													break;
			case 't':
				while ( *TUp->tname  &&  tolower( *optarg ) != tolower( *TUp->tname ) )
					TUp++;
				if ( ! *TUp->tname )
				{	LogMessage( LEVEL_1, QIC02_ERR_BAD_UNIT, __LINE__, "`%s'", optarg );
					exit( 1 );
				}
			break;
			case 'v':	TapeVerbose++;																break;

			default:
				LogMessage( LEVEL_1, QIC02_ERR_BAD_OPTION, __LINE__, "`%c'", optopt );
				exit( 1 );
		}

	// Convert offsets into addresses
	TUp->tps.status		+= TapeAddress;
	TUp->tps.control	+= TapeAddress;
	TUp->tps.command	+= TapeAddress;
	TUp->tps.data		+= TapeAddress;
	TUp->tps.dma		+= TapeAddress;

	switch ( TapeVerbose  |  TapeShowConfigThenQuit )
	{	case 0:							break;
		case 1:		TapeDumpParms();	break;
		default:	TapeDumpConfig();	break;
	}

	if ( TapeShowConfigThenQuit )
		exit( 0 );

	if ( getuid() != 0  &&  geteuid() != 0 )
	{	LogMessage( LEVEL_1, QIC02_ERR_EPERM, __LINE__, NULL );
		exit( 1 );
	}

	if ( INP( TUp->tps.status ) == 0xff )	
	{	LogMessage( LEVEL_1, QIC02_ERR_NOT_FOUND, __LINE__, "at 0x%x", TapeAddress );
		exit( 1 );
	}

	/*
	**	See if we're running on a PS/2
	*/
	if ( PS2() )
		IsPS2 |= DMA_PS2;	// Use the setup_dma() flag for the truth value

	/*
	**	Reset and ReadStatus Run busy wait.  Do both of these and
	**	see if the control lines behave as we expect.
	*/
	if ( TapeResetBW()  != EOK )
	{	LogMessage( LEVEL_1, QIC02_ERR_CANT_RESET, __LINE__, correct_unit );
		exit( 1 );
	}

	if ( TapeReadStatusBW( &TSS ) != EOK )
	{	LogMessage( LEVEL_1, QIC02_ERR_CANT_STATUS, __LINE__, correct_unit );
		exit( 1 );
	}

	/*
	**	Ok, poll mode seems to work.  Install the interrupt handler
	**	for the tape and perform a T_SELECT and see if an interrupt
	**	comes in (interrupt should come in for either ready *or*
	**	exception).
	*/
	if ( ! TapeRunPolled )
	{	int		tapehint;

		if ( (tapehint = qnx_hint_attach( TapeInterrupt, MK_FP( my_cs(), TapeISR ), my_ds() )) == -1 )
		{	LogMessage( LEVEL_1, QIC02_ERR_TAPE_ATT, __LINE__, NULL );
			exit( 1 );
		}

		TapeInterruptCounter = 0;
		TapeCommonCommand( T_SELECT );
		usec_delay( 32000 );					// Wait for command to complete
		if ( TapeInterruptCounter != 1 )
		{	qnx_hint_detach( tapehint );		// Let go of the interrupt
			TapeRunPolled++;					// run busy/wait
		}
		if ( TapeReadStatusBW( &TSS ) != EOK )	// Clear possible exception
		{	LogMessage( LEVEL_1, QIC02_ERR_CANT_STATUS, __LINE__, correct_unit );
			exit( 1 );
		}
	}

	if ( TapeVerbose )
		LogMessage( LEVEL_3, QIC02_ERR_LOG, __LINE__, "Driver will run %s", TapeRunPolled ? "busy/wait" : "with Interrupts" );

	if ( ! TapeRunPolled )
		if ( (TapeIntProxy = qnx_proxy_attach( 0, NULL, 0, -1 )) == -1 )
		{	LogMessage( LEVEL_1, QIC02_ERR_PROXY, __LINE__, "(tape unit)" );
			exit( 1 );
		}

	if ( (TapeFsysPid = qnx_name_locate( 0,TapeFsysName,0,0 )) == -1)
	{	LogMessage( LEVEL_1, QIC02_ERR_LOST_FSYS, __LINE__, NULL );
		exit( 1 );
	}

	// calloc Command Queue
	if ( (TapeCmdQ = TapeCmdHeadQp = calloc( i = TapeMaxIO + 1, sizeof *TapeCmdQ )) == NULL )
	{	LogMessage( LEVEL_1, QIC02_ERR_ENOMEM, __LINE__, NULL );
		exit( 1 );
	}

	// use link list (eliminates having to worry about wrap to front)
	while ( --i )
	{	TapeCmdHeadQp->np	= TapeCmdHeadQp + 1;
		TapeCmdHeadQp++;
	}
	TapeCmdHeadQp->np		= TapeCmdQ;

	TapeCmdHeadQp = TapeCmdTailQp = TapeCmdQ;

	// calloc Selector to Physical
	if ( (TapeCache2Phys = calloc( TapeCache2PhysSize, sizeof *TapeCache2Phys )) == NULL )
	{	LogMessage( LEVEL_1, QIC02_ERR_ENOMEM, __LINE__, NULL );
		return -1;
	}

	for ( i = 0;  i < QIC02_LOG_UNITS;  i++ )
	{	TapeInfo[ i ].d_type		= _TAPE;
		strcpy( TapeInfo[ i ].d_prefix, TapeDevName );
		if ( i & 1 )
			TapeInfo[ i ].d_prefix[ strlen( TapeInfo[ i ].d_prefix ) ] = 'r';
	}
	
	signal( SIGHUP, SIG_IGN );

#if 0
	// Check to see if we're in the background
	if ( ! InBackGround() )
	{	signal( SIGINT, SIG_IGN );
		signal( SIGQUIT, SIG_IGN );
	
		if ( fork() )
			exit( 0 );
	}
#endif

	TapeDriverInstalled = TRUE;
	TapeAgentPid		= getpid();

	if ( fsys_mount_driver(	TapeFsysPid,
							QIC02_LOG_UNITS,
							TapeInfo,
							DriverEntryList,
							TapeMaxIO,
							MAX_QIC02_PARALLEL,
							stack + QIC02_STACK_SIZE - 1,
							&TapeVersionReply,
							_DRVR_DMA ) == -1 )
	{	LogMessage( LEVEL_1, QIC02_ERR_MOUNT, __LINE__, NULL );
		write( fileno( stderr ), (MessageBufferp-2)->msg, strlen( (MessageBufferp-2)->msg ) );
		write( fileno( stderr ), (MessageBufferp-1)->msg, strlen( (MessageBufferp-1)->msg ) );
		exit( 1 );
	}

#ifdef DEBUG
	/*
	** This code was orignally added for debugging purposes.  It allows me
	** to see what data was written and read at process time (wvideo)
	** (These selectors map onto Fsys's cache space).
	*/
	for ( i = 0; i < TapeCache2PhysSize;  i++ )
		if ( (TapeCache2Phys[ i ].lsel = qnx_segment_overlay( TapeCache2Phys[ i ].phys_addr,TapeCache2Phys[ i ].nbytes)) == -1 )
			LogMessage( LEVEL_1, QIC02_ERR_SEL_INFO, __LINE__, NULL );
#endif

	if ( TapeRunPolled )
		LogMessage( LEVEL_2, QIC02_ERR_LOG, __LINE__, "Driver running busy/wait" );
	else
		LogMessage( LEVEL_2, QIC02_ERR_LOG, __LINE__, "Driver running with interrupts" );

	/*
	** From this point forward we become an agent
	*/
	fclose( stdin );
	fclose( stdout );
 	fclose( stderr );

	// Get out of the way unless we've been started otherwise
	if ( getprio( 0 ) == 10 )
		setprio( 0,9 );

	signal( SIGUSR1, TapeYankSignalHandler );	// dead man's switch

	if ( sigsetjmp( TapeYankEnv, TRUE ) != 0 )
	{	LogMessage( LEVEL_1, QIC02_ERR_USER_RESET, __LINE__, NULL );

		TapeResetBW();
		TapeReadStatusBW( &TSS );

		TapeLastError			= EOK;
		TapeLastCmd				= 0;
		TapeExceptionLevel		= EXC_CLEAR;
		TapeInterruptCounter	= 0;

		TapeCmdHeadQp = TapeCmdTailQp = TapeCmdQ;

		while ( Creceive( 0,NULL,0 ) != -1 )
			;
		for ( i = 0;  i < QIC02_LOG_UNITS;  i++ )
		{	if ( TapeInfo[ i ].d_busy == TRUE )
			{	TapeInfo[ i ].d_busy	= FALSE;
				TapeInfo[ i ].d_nblks	= 0;
				Trigger( TapeFsysProxy );
			}
		}
	}

	while ( 1 )
	{	if ( Receive( TapeDriverProxy, NULL, 0 ) == -1 )
			continue;

        // Hi-hoe, Hi-hoe, its off to work I go...

		while ( TapeCmdTailQp != TapeCmdHeadQp )
		{	(TapeCmdTailQp = TapeCmdTailQp->np)->tip->d_retval = _DRVR_ERROR;

			if ( TapeCmdTailQp->flag.common_command )
			{	if ( TapeCommonCommand( TapeCmdTailQp->data.command ) == EOK )
					switch ( TapeCmdTailQp->tip->d_retval = TapeWaitTillDone() )
					{	case _DRVR_OK:
							TapeCmdTailQp->tip->d_nblks = 1;
						break;

						default:
							TapeCmdTailQp->tip->d_retval	= (unsigned) -1;
							TapeCmdTailQp->tip->d_error		= TapeErrno2Return;
					}
				else
				{	TapeCmdTailQp->tip->d_retval	= (unsigned) -1;
					TapeCmdTailQp->tip->d_error		= EIO;
				}
			}
			else	// Read/Write
			{	if ( TapeExceptionLevel )
				{	TapeCmdTailQp = TapeCmdHeadQp;

					if ( TapeExceptionLevel & EXC_NON_FATAL )
					{	TapeCmdTailQp->tip->d_retval = _DRVR_EOT;
						TapeExceptionLevel &= ~EXC_NON_FATAL;
					}

					if ( TapeExceptionLevel & EXC_FATAL_CLEAR_IOCNTL )
						TapeCmdTailQp->tip->d_retval = _DRVR_EOT;

					if ( TapeExceptionLevel & EXC_FATAL )
						TapeCmdTailQp->tip->d_retval = _DRVR_ERROR;
					
					TapeLastCmd = 0;
				}
				else
				{	if ( RW( TapeCmdTailQp->flag.write ? T_WRITE : T_READ, &TapeCmdTailQp->data.dma ) == EOK )
					{	switch ( TapeCmdTailQp->tip->d_retval = TapeWaitTillDone() )
						{	case _DRVR_OK:
							case _DRVR_PART_IO:
								TapeCmdTailQp->tip->d_nblks += TapeCmdTailQp->data.dma.no_of_blocks; 
							break;
						}
					}
					else
					{	if ( TapeCmdTailQp == TapeCmdHeadQp )
							TapeCmdTailQp->tip->d_retval	= _DRVR_ERROR;
						else
						{	TapeCmdTailQp->tip->d_retval	= _DRVR_PART_IO;
						}
					}
				}
			}
		}
                	
		LogMessage( LEVEL_2, QIC02_ERR_LOG, __LINE__, "%u block(s) I/O'd", (unsigned) TapeCmdTailQp->tip->d_nblks );

		TapeCmdTailQp->tip->d_busy	= FALSE;
		Trigger( TapeFsysProxy );
	}	

	return EOK;
}


/* End of File */
