#ifdef __USAGE

%C [-b Block Count] [-r] [-n Read Count] [-N Tape Device] [-s IO Size]

Options:
  -b            Blocks to write (rounded up to next IO Size)
  -r            Read Only (don't perform write pass)
  -n            Number of read iterations
  -N            Name of the tape device
  -s            Size (in blocks)
  -v            No verification of data (read/write as fast as possible)
  -w            Write Only (don't perform read (on this pass))

Where:
  Block Count   Number of Blocks to read/write (default is 10000 +- IO Size)
  Read Count    The number of read tests (iterations) to perform (default is 1)
  Tape Device   Is /dev/tp0 (default)
  IO Size       The size of buffer when reading and writing (max 63 blocks)

#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<fcntl.h>
#include	<unistd.h>
#include	<errno.h>

#include	<sys/types.h>
#include	<sys/stat.h>
#include	<sys/qioctl.h>


#define		TRUE			1
#define		FALSE			0
#define		NOT				!


typedef struct
{	unsigned long	BlockNumber;
	unsigned char	Filler[ _BLOCK_SIZE - sizeof (unsigned long) ];

}	BLOCK;

QIC02_MSG_STRUCT	 qic02ms;
int					 fd_tape;
char				*ProgramNamep;
char				 TapeName[ 64 ] = "/dev/tp0";


/***********************************************************/
void				Rewind( void )
/***********************************************************/
{
	fprintf( stdout, "Rewinding..." );
	fflush( stdout );
	errno = EOK;
	qic02ms.Header.Command = T_BOT;
	if ( qnx_ioctl( fd_tape, QCTL_RAW_CMD, &qic02ms, sizeof (QIC02_HEADER_STRUCT),
					&qic02ms, sizeof(QIC02_MSG_STRUCT) ) == -1 ||
			qic02ms.Header.RetVal != EOK )
	{	fprintf( stderr, "Unable to Rewind (\"%s\")\n", strerror( errno ) );
		exit( -1 );
	}
	fprintf( stdout, "\n" );
	fflush( stdout );
}


/***********************************************************/
void				OpenTape( int flag )
/***********************************************************/
{
	errno = EOK;
	if ( (fd_tape = open( TapeName,flag )) == -1 )
	{	fprintf( stderr, "%s: Unable to open tape %s (%s)\n", ProgramNamep, TapeName, strerror( errno ) );
		exit( -1 );
	}
}


/***********************************************************/
					main( int argc, char *argv[] )
/***********************************************************/
{
	register BLOCK		*blocktmpp;
	register char		*blockp;
	register int		 i;

	int		 			 io_size			= 63;
	unsigned long		 block_count		= 10000L,
						 block_current;
	unsigned int		 read_only			= FALSE;
	unsigned int		 write_only			= FALSE;
	unsigned int		 verify				= TRUE;
	unsigned int		 read_count			= 1;
	int					 block_size;


	ProgramNamep = argv[0];

	while ( (i = getopt( argc, argv, "b:rn:N:s:vw" )) != -1 )
		switch ( i )
		{	case 'b':	block_count = strtol( optarg,NULL,10 );				break;
			case 'r':	read_only++;	write_only = FALSE;					break;
			case 'n':	read_count  = strtol( optarg,NULL,10 );				break;
			case 'N':	strncpy( TapeName,optarg,sizeof TapeName - 1 );		break;
			case 's':	io_size		= strtol( optarg,NULL,10 ); 			break;
			case 'v':	verify		= FALSE;								break;
			case 'w':	write_only++;	read_only = FALSE;					break;
		}

	if ( io_size > 63 )
	{	fprintf( stderr, "%s: Block size limited to 63\n", ProgramNamep );
		io_size = 63;
	}

	if ( block_count % io_size )
		block_count = (block_count / io_size + 1) * io_size;

	if ( (blockp = malloc( io_size * sizeof (BLOCK) )) == NULL )
	{	fprintf( stderr, "%s: Unable to allocate %d blocks\n", ProgramNamep, io_size );
		exit( -1 );
	}
				
	if ( read_only )
		fprintf( stdout, "%s: Read only test", ProgramNamep );
	else if ( write_only )
		fprintf( stdout, "%s: Write Only test", ProgramNamep );
	else
		fprintf( stdout, "%s: Will perform write pass then a read", ProgramNamep );

	fprintf( stdout, " of %lu blocks (%d blocks per io)\n", block_count, io_size );
	fprintf( stdout, "%s: Will perform %u read pass(es)\n", ProgramNamep, read_count );
	fprintf( stdout, "%s: Read verification of data will %sbe done\n", ProgramNamep, verify ? "" : "NOT " );

	if ( NOT read_only )
	{	OpenTape( O_WRONLY );
		Rewind();

		for ( i = 0;  i < _BLOCK_SIZE;  i++ )
			blockp[ i ] = i;

		blocktmpp = (BLOCK *) blockp;

		for ( i = 0;  i < io_size;  i++ )
			memcpy( blocktmpp++, blockp, sizeof *blocktmpp );
		
		for ( block_current = 1L;  block_current <= block_count;  block_current += io_size )
		{	for ( blocktmpp = (BLOCK *) blockp, i = 0;  i < io_size;  blocktmpp++, i++ )
				blocktmpp->BlockNumber = block_current + i;

			fprintf( stdout, "\r%s: Writing %lu/%lu...(%lu) ", ProgramNamep, block_current + io_size - 1, block_count, (block_current + io_size - 1) * sizeof (BLOCK) );
			fflush( stdout );

			if ( write( fd_tape, blockp, io_size * sizeof (BLOCK) ) == -1 )
			{	fprintf( stderr, "%s: Unable to write to tape\n", ProgramNamep );
				exit( -1 );
			}
		}

		if ( close( fd_tape ) != EOK )
		{	fprintf( stderr, "\nERROR unable to close" );
			exit( 1 );
		}
	}

	if ( NOT write_only )
	{	while ( read_count-- )
		{	OpenTape( O_WRONLY );
			Rewind();
			if ( close( fd_tape ) != EOK )
			{	fprintf( stderr, "\nERROR unable to close" );
				exit( 1 );
			}

			OpenTape( O_RDONLY );
			block_current = 1L;

			for ( block_current = 1L; block_current <= block_count;  block_current += io_size )
			{	fprintf( stdout, "\r%s: Read %lu/%lu (%lu)...", ProgramNamep, block_current + io_size - 1, block_count, (block_current +io_size - 1) * sizeof (BLOCK) );
				fflush( stdout );

				if ( (block_size = read( fd_tape, blockp, io_size * sizeof (BLOCK) )) == io_size * sizeof (BLOCK) )
				{	if ( verify )
						for ( blocktmpp = (BLOCK *) blockp, i = 0;  i < io_size;  blocktmpp++, i++ )
						{	/*
							**	Verify that the block read contains the block number
							**	written in the first long word
							*/
							if ( blocktmpp->BlockNumber != block_current + i )
							{	fprintf( stderr, "\nERROR - Expected Block %lu, Read Block %lu\n", block_current + i, *(long *)(blockp + sizeof (BLOCK) * i) );
								exit( -1 );
							}

							// Check Middle
							if (	*((char *)blocktmpp+0x100) != 0x00 ||
									*((char *)blocktmpp+0x101) != 0x01 ||
									*((char *)blocktmpp+0x102) != 0x02 ||
									*((char *)blocktmpp+0x103) != 0x03		)
							{	fprintf( stderr, "\nERROR - (Middle) Block %lu contains unexpected data\n", block_current + i );
								exit( -1 );
							}
						

							// Check End
							if (	*((char *)blocktmpp+0x1fc) != 0xfc ||
									*((char *)blocktmpp+0x1fd) != 0xfd ||
									*((char *)blocktmpp+0x1fe) != 0xfe ||
									*((char *)blocktmpp+0x1ff) != 0xff		)
							{	fprintf( stderr, "\nERROR - (End) Block %lu contains unexpected data\n", block_current + i );
								exit( -1 );
							}
						}
				}
				else
				{	fprintf( stderr, "%s: ERROR, read returned returned %d bytes, wanted %d (%s)\n", ProgramNamep, block_size, io_size * sizeof (BLOCK), strerror( errno ) );
					exit( -1 );
				}
			}	

			if ( close( fd_tape ) != EOK )
			{	fprintf( stderr, "\nERROR unable to close" );
				exit( 1 );
			}
		}
	}
}


/* End of File */
