
/*
 * dma.h: manifests and structures
 *  Copyright Quantum Software Systems Ltd. 1984 UNPUBLISHED
 */
#ifndef	_qnx_sys_dma_h_included
#define	_qnx_sys_dma_h_included

#ifndef _qnx_sys_dktypes_h_included
#include <qnx/sys/dktypes.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * structure for dma io, maintains logical-physical mapping
 */
struct	dma_buffer {
	caddr_t bufp;			/*	logical address 	*/
	paddr_t phys_addr;		/*	physical address 	*/
};

typedef	struct dma_buffer dma_t;

/*
 * flags for dma io routines
 * DMA_OUTPUT = input or output.
 * PS/2 dma   = uses micro-channel dma
 * DMA_WORD   = 16 bit vs 8 bit dma
 * DMA_WAIT   = for setup_dma, set parameters, but do not arm the controller
 *              (controller may be armed with kick_dma)
 */
#define	DMA_OUTPUT	001
/*	ps2 / at	*/
#define	DMA_PS2		002
/*	word / byte	*/
#define	DMA_WORD	004
/* don't arm dma controller on setup */
#define	DMA_WAIT	010

extern int alloc_dma_buf(dma_t *dbuf, pid_t process, ulong nbytes);
extern int setup_dma(uint dma_flags, dma_t *dbuf, uint iosize, int channel);
extern int kick_dma(uint dma_flags, int channel);
extern int clear_dma(uint dma_flags, int channel);

extern paddr_t phys_addr(caddr_t ptr);

#ifdef __cplusplus
}
#endif

#endif
