

#ifndef _qnx_sys_inline_h_included
#define	_qnx_sys_inline_h_included

#ifdef __cplusplus
extern "C" {
#endif

/*

	some handy pragma's for low-level work:	

void	ll_debug(void);
	-	issues an int f3h which puts you in to the low level debugger.

void	delay_8X(int count)
	-	delays approximately 8 clocks * count.
		useful for 'little' timing loops

unsigned	pswget(void);
	-	returns the current value of the flags register.

void	restore(unsigned psw)
	-	stores psw into flags register.

unsigned	disable(void);
	-	disables interrupts, and returns previous value of flags.

unsigned	enable(void);
	-	enables interrupts, and returns previous value of flags.

unsigned	str_forward(void);
	-	sets direction flag to forward and returns previous value of flags.

unsigned	str_reverse(void);
	-	sets direction flag to reverse and returns previous value of flags.

int segsize(ushort_t sel);
    -   return the size of the segment

int sldt(void);
    -   return descriptor for LDT.
    
unsigned	my_cs(void)
	-	returns current value of 'cs' register.

unsigned	my_ds(void)
	-	returns current value of 'ds' register.

void	io_safe(void);
	-	does 2 "jmp ." to avoid trashing io device.

long	swapw(int val)
	-	swaps byte ordering of a word

long	swap(long lval)
	-	swaps byte ordering of a long


memcpy8(void far *dst, const void far *src, int nbytes);
memcpy16(void far *dst, const void far *src, int nbytes);
memcpy32(void far *dst, const void far *src, int nbytes);

	-	memcopies by 8,16 and 32 bits resp.
	-	note nbytes is BYTES and is expected to be an even
		multiple of the copy size (ie 16 or 32).

*/
#ifndef __TYPES_H_INCLUDED
#include <sys/types.h>
#endif

ushort_t  inb(int port);
#pragma aux inb = \
          0x32 0xe4          /* xor ah,ah */\
          0xec               /* in  al,dx */\
          parm nomemory [dx] \
          modify exact nomemory [ax];

ushort_t  inw(int port);
#pragma aux inw = \
          0xed              /* in ax,dx   */\
          parm nomemory [dx] \
          modify exact nomemory [ax];

void outb(int port, ushort_t val);
#pragma aux outb = \
          0xee             /*  out dx,al   */\
          parm nomemory [dx] [ax] \
          modify exact nomemory [];

void outw(int port, ushort_t val);
#pragma aux outw = \
          0xef             /*  out dx,ax    */\
          parm nomemory [dx] [ax] \
          modify exact nomemory [];

void	delay_8X(unsigned count);
#pragma aux delay_8X =	 		\
		0xe2 0xfe	/*	loop -2	*/		\
		parm nomemory [cx]	\
		modify exact nomemory [cx];



void	ll_debug(void);
#pragma	aux	ll_debug = \
		0xcd 0xf3	/*	int 0f3h	*/	\
		parm []         		\
		modify exact [];

void	io_safe(void);
#pragma	aux	io_safe = \
		0xeb 0x00	/*	jmp .	*/	\
		0xeb 0x00	/*	jmp .	*/	\
		parm []         		\
		modify exact [];

unsigned pswget(void);
#pragma aux pswget = \
		0x9c	/*	pushf	*/ \
		0x58	/*	pop ax  */ \
		parm nomemory	[]		\
		modify exact nomemory [ AX ];

unsigned disable(void);
#pragma aux disable = \
		0x9c	/*	pushf	*/ \
		0xfa	/*	cli		*/ \
		0x58	/*	pop ax  */ \
		parm []		\
		modify exact [ AX ];

unsigned enable(void);
#pragma	aux	enable = \
		0x9c	/*	pushf	*/ \
		0xfb	/*	sti		*/ \
		0x58	/*	pop ax  */ \
		parm []		\
		modify exact [ AX ];
		

unsigned str_forward(void);
#pragma	aux str_forward = \
		0x9c	/*	pushf	*/ \
		0xfc	/*	cld		*/ \
		0x58	/*	pop ax  */ \
		parm nomemory	[]		\
		modify exact nomemory [ AX ];

unsigned str_reverse(void);
#pragma	aux cforward = \
		0x9c	/*	pushf	*/ \
		0xfd	/*	cld		*/ \
		0x58	/*	pop ax  */ \
		parm nomemory	[]		\
		modify exact nomemory [ AX ];

void restore(unsigned psw);
#pragma	aux	restore = \
		0x50	/*	push	ax	*/ \
		0x9d	/*	popf		*/ \
		parm [ AX ]                	\
		modify exact [];

unsigned short my_cs( void );
#pragma aux my_cs =				\
		0x8c 0xc8		/* mov ax,cs	*/			\
		parm nomemory [] value [ AX ]		\
		modify exact nomemory [ AX ];

unsigned short my_ds( void );
#pragma aux my_ds =				\
		0x8c 0xd8		/* mov ax,ds	*/			\
		parm nomemory [] value [ AX ]		\
		modify exact nomemory [ AX ];
		
unsigned short my_ss( void );
#pragma aux my_ss =	"mov ax,ss" \
		parm nomemory [] value [ AX ]		\
		modify exact nomemory [ AX ];
		
unsigned short my_es(void);
#pragma aux my_es = \
		0x8c 0xc0     /* mov ax,es */\
		parm nomemory [] value [ax] \
		modify exact nomemory [ax];
	
#if __WATCOMC__ >= 900
caddr_t insw(caddr_t buffer, unsigned count, unsigned port);
#pragma aux insw = ".286p" "cld" "rep insw"\
	parm [es di] [cx] [dx] \
	modify [cx di] \
	value [es di];

caddr_t outsw(caddr_t buffer, unsigned count, unsigned port);
#pragma aux outsw = ".286p" "push ds" "cld" "mov ds,di" "rep outsw" "pop ds" \
	parm [es di] [cx] [dx] \
	modify [cx di ax] \
	value [di si];
	
#endif	

void inp_words( char __far *buffer, unsigned count, unsigned port );
#pragma aux inp_words =	\
		0xfc	/* cld	*/	\
		0xf2 0x6d	/* rep insw ; Read them in */	\
		parm nomemory [ ES DI ] [ CX ] [ DX ]		\
		modify exact nomemory [ CX DI ];

void outp_words( char __far *buffer, unsigned count, unsigned port );
#pragma aux outp_words =\
		0xfc	/* cld			*/	\
		0x1e	/* push ds		*/	\
		0x8e 0xdf /*	mov ds,di	; Change		*/	\
		0xf2 0x6f /*	rep outsw	; Write them out	*/	\
		0x1f	/*	pop ds		*/	\
		parm nomemory [ SI DI ] [ CX ] [ DX ]		\
		modify exact nomemory [ CX SI ];


/* most thanks to aboyd... */
 /* dx = selector	ax = offset 	cx = length in bytes */
int verr(void __far *ptr, int len);
#pragma aux verr =		\
		0x0f 0x00 0xe2		/* verr dx		  */ \
		0x75 0x13		/* jnz verr_bad   */ \
		0xe3 0x0c		/* jcxz verr_ok   */ \
 		0x49	/* dec cx		  */ \
		0x03 0xc8		/* add cx,ax	  */ \
		0x72 0x0c		/* jc verr_bad    */ \
 		0x0f 0x03 0xC2		/* lsl ax,dx	  */ \
 		0x3b 0xc1		/* cmp ax,cx	  */ \
 		0x72 0x05		/* jb verr_bad    */ \
		/* verr_ok:       */ \
 		0xb8 0x01 0x00		/* mov ax,#1      */ \
		0xeb 0x02		/* jmp verr_done  */ \
		/* verr_bad:      */ \
 		0x33 0xc0		/* xor ax,ax      */ \
		/* verr_done:     */ \
		parm	nomemory [dx ax] [cx] \
		modify exact nomemory [ax cx];

int segsize(ushort_t sel);
#pragma aux segsize = \
        0x0f 0x03 0xc0      /* lsl ax,ax      */\
        parm nomemory [ax] \
        modify exact nomemory [ax];


int sldt(void);
#pragma aux sldt = \
	0x0f 0x00 0xc0          /* sldt ax         */\
	parm nomemory [] \
	modify exact nomemory [ax];
        
 /* dx = selector	ax = offset 	cx = length in bytes */
int verw(void __far *ptr, int len);
#pragma aux verw =		\
		0x0f 0x00 0xea		/* verw dx		  */ \
		0x75 0x13		/* jnz verr_bad   */ \
		0xe3 0x0c		/* jcxz verr_ok   */ \
 		0x49	/* dec cx		  */ \
		0x03 0xc8		/* add cx,ax	  */ \
		0x72 0x0c		/* jc verr_bad    */ \
 		0x0f 0x03 0xC2		/* lsl ax,dx	  */ \
 		0x3b 0xc1		/* cmp ax,cx	  */ \
 		0x72 0x05		/* jb verr_bad    */ \
		/* verr_ok:       */ \
 		0xb8 0x01 0x00		/* mov ax,#1      */ \
		0xeb 0x02		/* jmp verr_done  */ \
		/* verr_bad:      */ \
 		0x33 0xc0		/* xor ax,ax      */ \
		/* verr_done:     */ \
		parm	nomemory [dx ax] [cx] \
		modify exact nomemory [ax cx];

unsigned swapw(unsigned);
#pragma aux swapw =		\
		0x86 0xe0 /* xchg ah,al */	\
		parm nomemory [ax]		\
		modify exact nomemory [ax];

long   lswap(long);
#pragma aux lswap parm [dx ax] =\
    0x86 0xf0 /* xchg dh,al */\
    0x86 0xd4 /* xchg dl,ah */\



memcpy8(void far *a, const void far *b, int nbytes);
#pragma aux memcpy8 =\
	0x1e		/*	push ds	*/	\
	0x8e 0xda	/*  mov ds,dx */ \
	0xf3 0xa4	/*	rep movsb */ \
	0x1f		/*	pop ds */ \
	parm nomemory [es di] [dx si] [cx] \
	modify exact [si di cx];
	
memcpy16(void far *a, const void far *b, int nbytes);
#pragma aux memcpy16 =	\
	0x1e		/*	push ds	*/	\
	0x8e 0xda	/*  mov ds,dx */ \
	0xc1 0xe9 0x01		/* shr cx,1 */ \
	0xf3 0xa5	/* rep movsw */		\
	0x1f		/*	pop ds */ \
	parm nomemory [es di] [dx si] [cx] \
	modify exact [si di cx];
	

memcpy32(void far *a, const void far *b, int nbytes);
#pragma aux memcpy32 =	\
	0x1e		/*	push ds	*/	\
	0x8e 0xda	/*  mov ds,dx */ \
	0xc1 0xe9 0x02		/* shr cx,2 */  \
	0x66 0xf3 0xa5		/* 32: rep movsw */  \
	0x1f		/*	pop ds */ \
	parm nomemory [es di] [dx si] [cx] \
	modify exact [si di cx];

#ifdef __cplusplus
}
#endif

#endif
